/**** $Id: nexraget.c,v 1.22 2022/02/04 16:11:14 tyoksas Tst $ ****/
/*
** Name:
**      nexraget : ADDE data server for NEXRAD Level III products.
**
** Interface:
**      int
**      main(int argc, char *argv[])
**
** Input:
**      argc : argument count
**      argv : argument vector
**
** Input and Output:
**      none
**
** Output:
**      none
**
** Return values:
**        0 : success
**      < 0 : failure, enumerated by text message returned to client
**
** Remarks:
**
** Categories:
**      ADDE server
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include "mcidas.h"
#include "mcidasp.h"
#include "m0arg.h"
#include "nexrutil.h"

int errno;

/* Program specific debug mode */
#define DEBUG 0

int
main( int argc, char *argv[] )
{

    const char *adde;                       /* AUX= arg; YES-> update aradir */
    const char *configname;                 /* name of directory to search   */
    const char *dset_descriptor;            /* dataset descriptor name       */
    const char *dum;                        /* dummy for arg fetchers        */
    const char *img_place;                  /* image placement key (EC..EU)  */
    const char *unit;                       /* requested output unit         */
    const char *cid;                        /* radar station ID              */

    char        data[MAX_LINE_LEN]={0};     /* maximum line length           */

    char        dbg[MAX_ERR_LEN];           /* textual debug message         */
    char        err[MAX_ERR_LEN];           /* err string from SelectImages  */

    char       *p;                          /* generic character pointer     */

    double      p_lat;                      /* center lat of returned image  */
    double      p_lon;                      /* center lon of returned image  */
    float       p_elem;                     /* center elem of returned image */
    float       p_line;                     /* center line of returned image */

    int         cards[MAX_NUM_CARDS*MAX_CARD_LEN]; /* mcidas comment cards   */

    int         aradir[IMG_DIR_LEN];        /* McIDAS area directory         */
    int         auxblk[AUX_BLK_LEN];        /* McIDAS auxiliary block        */
    int         len_auxblk=0;               /* length of mcidas aux block    */
    int         calcod[CAL_COD_LEN];        /* McIDAS calibration codicil    */
    int         len_calcod=0;               /* length of mcidas cal codicil  */
    int         navcod[NAV_COD_LEN];        /* McIDAS navigation codicil     */
    int         len_navcod=4*NAV_COD_LEN;   /* Nav codicil length in bytes   */

    int         a_elem;                     /* requested area element        */
    int         a_line;                     /* requested area line           */
    int         band;                       /* band number to read           */
    int         base_res;                   /* base resolution               */
    int         begele;                     /* beg element of ret image      */
    int         bday, eday;                 /* beginning/ending day          */
    int         bpos, epos;                 /* beginning/ending position     */
    int         bss,  ess;                  /* beginning/ending ss           */
    int         btim, etim;                 /* beginning/ending time         */
    int         bytes_per_line=0;           /* number of bytes/image line    */
    int         bytes_per_pix=0;            /* number of bytes/pixel         */
    int         bytes_to_send=0;            /* number of data bytes to send  */
    int         bytes_to_zero=0;            /* number of bytes to zero       */
    int         pixels_to_send=0;           /* num of image pixels to send   */
    int         curline;                    /* loop variable for lines       */
    int         ccyy;                       /* year                          */
    int         data_bytes;                 /* number of bytes in read       */
    int         elem_res;                   /* element resolution            */
    int         i;                          /* loop variable                 */
    int         jjj;                        /* Julian day                    */
    int         line_res;                   /* line resolution               */
    int         nelems;                     /* number of elements to send    */
    int         nj, nk;                     /* numbers of lines and elements */
    int         nlines;                     /* number of lines to send       */
    int         nval;                       /* number of elements to calib.  */
    int         orig_nelems;                /* unalterd # of lines to send   */
    int         orig_nlines;                /* unalterd # of lines to send   */
    int         rc;                         /* function return status        */
    int         spac;                       /* byte size of one pixel        */
    int         total_bytes=0;              /* number of bytes to send       */
    int         ul_elem;                    /* upper left elem of ret image  */
    int         ul_line;                    /* upper left line of ret image  */
    int         zeros_on_left;              /* number of zero bytes on left  */

    CONFIG      cfginfo = {0};              /* configuration file info       */
    CRITERIA    request = {0};              /* set of tests for image        */
    FILELIST   *satisfy = NULL;             /* pointer to a FILELIST struct  */
    READPARM    read    = {0};              /* read parameters               */

    const char server[] = {"nexraget"};     /* server name                   */
    servacct controlBlock = {0};            /* server control block struct   */
    char     *ctrlstr = NULL;               /* server control string         */


    /*  ADDED RCD 20151015 -- standard error for invalid band */
    char       *cALL;                       /* for BAND=ALL */
    int         nband;                      /* number of bands */
    int         bband;                      /* beginning band */
    int         eband;                      /* ending band */
    /*  ADDED RCD 20151015 -- standard error for invalid band */


    /*
    ** Initialize local server  This should do the following:
    **
    **    1) initialize the comm_block
    **    2) retrieve the client request string
    **    3) initialize the McIDAS environment
    **    4) parse the client request for future Mccmd***() calls
    **    5) activate tracing if requested
    */

    rc = M0InitLocalServer( server, argc, argv, &controlBlock, &ctrlstr );

    if ( rc < 0 ) {
      M0sxdone( &controlBlock );
      exit( 1 );
    }

    /*
    ** Setup tracing if directed to by request
    */

#if DEBUG
    (void) M0SetTraceFlag( 1 );
#else
    rc = M0IsTraceSet( ctrlstr );
#endif

    M0sxtrce( "nexraget:: starting up: test level = 1" );

    /*
    ** Get the dataset group name
    */

    rc = Mcargstr( 0, " ", 1, " ", &dset_descriptor );

    (void) sprintf( dbg, "nexraget:: VERSION 1.13" );
    M0sxtrce( dbg );

    (void) sprintf( dbg, "nexraget:: dataset group name: %s", dset_descriptor );
    M0sxtrce( dbg );

    /*
    ** Get the configuration file name that was specified in the INFO= field.
    */

    rc = Mcargquo( 0, &configname ) - 10;

    if ( rc > 0 ) {
      (void) sprintf( dbg, "nexraget:: configuration file name: %s", configname );
      M0sxtrce( dbg );
    }

    rc = GetConfigInfo( dset_descriptor, configname, &cfginfo, err );

    if ( rc == FAILURE ) {
      (void) sprintf( controlBlock.errormsg, err );
      controlBlock.returncode = errno;
      M0sxdone( &controlBlock );
      exit(1);
    }

    (void) sprintf( dbg, "nexraget:: data directory: %s", cfginfo.dataloc );
    M0sxtrce( dbg );

    (void) sprintf( dbg, "nexraget:: file mask: %s", cfginfo.filemask );
    M0sxtrce( dbg );

    (void) sprintf( dbg, "nexraget:: ip mask: %s", cfginfo.ipmask );
    M0sxtrce( dbg );

    /*
    ** See if the requesting machine has been enabled to get data from
    ** this server.  If not, return an error and exit.
    */

    rc = AllowedAccess( cfginfo, &controlBlock );

    if ( rc == FAILURE ) {
      (void) sprintf( controlBlock.errormsg, "Dataset access denied by server" );
      controlBlock.returncode = errno;
      M0sxdone( &controlBlock );
      exit(1);
    }

    /*
    ** Dataset position
    */

    rc = Mcargint(0, " ", 2,    0, 0, -1, &bpos, &dum);

    (void) sprintf( dbg, "nexraget:: dset pos = %d", bpos );
    M0sxtrce( dbg );

    /*
    ** Set the ending position equal to the beginning position.
    ** The underlying function, SelectNexrImages, allows a range to be
    ** specified, but the current ADDE protocol does not support this.
    */

    epos = bpos;

    /*
    ** Image placement key ( EU, EC, ED,
    **                       AU, AC, AD,
    **                       IU, IC, ID )
    */

    rc = Mcargstr( 0, " ", 3, "IU", &img_place );

    /*
    ** Resolution
    **
    ** NOTE: In McIDAS versions after 7.0, there is no LMAG or EMAG keys
    ** passed in the protocol.  The parsing for these keys is left in
    ** to match the code in 'agetserv.fp'.
    */

    rc = Mcargint( 0,     " ", 6,        1, 0, -1, &base_res, &dum );

    rc = Mcargint( 0, "LMA.G", 1, base_res, 0, -1, &line_res, &dum );
    rc = Mcargint( 0, "EMA.G", 1, base_res, 0, -1, &elem_res, &dum );

    line_res = (line_res == 0) ? 1 : abs(line_res);
    elem_res = (elem_res == 0) ? 1 : abs(elem_res);

    (void) sprintf( dbg, "nexraget:: line_res = %d elem_res = %d", line_res, elem_res );
    M0sxtrce( dbg );

    /*
    ** Number of lines and elements to read
    */

    rc = Mcargint( 0, " ", 7, DEF_NUM_LINES, 0, -1, &nlines, &dum );
    rc = Mcargint( 0, " ", 8, DEF_NUM_ELEMS, 0, -1, &nelems, &dum );

    orig_nlines = nlines;
    orig_nelems = nelems;

    /*
    ** Band number to read
    */

    rc = Mcargint( 0, "BAN.D", 1, 1, 0, -1, &band, &dum );

    /*
    ** Beginning and ending McIDAS Band number
       ONLY IF user enters a band range 
    */
    
    bband = 1;
    eband = MAX_NUM_BAND;
    nband = Mcargnum( 0, "BAN.D" );
    if( nband > 0 ) {
      rc = Mcargstr( 0, "BAN.D", 1, " ", (const char **)&cALL);
      if( strncmp( cALL, "ALL",3 ) == 0 ) {
        bband = 1;
        eband = MAX_NUM_BAND;
      } else {
        rc = Mcargint( 0, "BAN.D", 1,    -1,     1, MAX_NUM_BAND, &bband, &dum );
        if( rc < 0 ) {
          (void)strcpy(controlBlock.errormsg, "No images satisfy the selection criteria");
          controlBlock.returncode = -30;
          M0sxdone(&controlBlock);
          return(0);
        }
        rc = Mcargint( 0, "BAN.D", 2, bband, bband, MAX_NUM_BAND, &eband, &dum );
        if( rc < 0 ) {
          (void)strcpy(controlBlock.errormsg, "No images satisfy the selection criteria");
          controlBlock.returncode = -30;
          M0sxdone(&controlBlock);
          return(0);
        }
      }
    }
    (void) sprintf( dbg, "nexradir:: bband = %d eband = %d", bband, eband );
    M0sxtrce( dbg );


    /*
    ** Requested output unit
    */

    rc = Mcargstr( 0, "UNI.T", 1, "RAW ", &unit );

    (void) sprintf( dbg, "nexraget:: requested unit = %s", unit );
    M0sxtrce( dbg );

    /*
    ** AUX option [YES|NO] (whether or not to send information cards).  For
    ** ADDE this should always be YES.
    */

    rc = Mcargstr( 0, "AUX", 1, "NO", &adde );

    (void) sprintf( dbg, "nexraget:: update AREA directory with unit, scale = %s", adde );
    M0sxtrce( dbg );

    /*
    ** Size of one pixel (spac).  Have to overide directive client
    ** size of a pixel for number of different types.
    */

    rc = Mcargint( 0, "SPA.C", 1, 1, -1, 0, &spac, &dum );
    if ( strcmp (unit, "BRIT") == 0 ) spac = 1;
    if ( strcmp (unit, "RAW" ) == 0 ) spac = 1;
    if ( strcmp (unit, "VIP" ) == 0 ) spac = 1;
    if ( strcmp (unit, "TOPS") == 0 ) spac = 1;
    if ( strcmp (unit, "H2O" ) == 0 ) spac = 4;    /* <<<<< UPC mod 20100826 >>>>> */
    if ( strcmp (unit, "ECHO") == 0 ) spac = 4;    /* values can be negative */
    if ( strcmp (unit, "RAIN") == 0 ) spac = 4;    /* values are scaled ints */
    if ( strcmp (unit, "VEL" ) == 0 ) spac = 4;    /* values can be negative */
    if ( strcmp (unit, "SRMV") == 0 ) spac = 4;    /* values can be negative */

    /*
    ** Beginning and ending data times
    */

    rc = Mcargihr( 0, "TIM.E", 1,   -1, -1, 240000, &btim, &dum );
    rc = Mcargihr( 0, "TIM.E", 2, btim, -1, 240000, &etim, &dum );

    (void) sprintf( dbg, "nexraget:: btim = %d etim = %d", btim, etim );
    M0sxtrce( dbg );

    /*
    ** Beginning and ending day in ccyyddd format
    */

    rc = Mcargiyd( 0, "DAY", 1,   -1, 0, -1, &bday, &dum );
    rc = Mcargiyd( 0, "DAY", 2, bday, 0, -1, &eday, &dum );

    (void) sprintf( dbg, "nexraget:: bday = %d eday = %d", bday, eday );
    M0sxtrce( dbg );

    /*
    ** Beginning and ending McIDAS ss number
    */

    rc = Mcargint( 0, "SS", 1,  -1, 0, -1, &bss, &dum );
    rc = Mcargint( 0, "SS", 2, bss, 0, -1, &ess, &dum );

    (void) sprintf( dbg, "nexraget:: bss = %d ess = %d", bss, ess );
    M0sxtrce( dbg );

    /*
    ** NEXRAD Level III product station ID
    */

    rc = Mcargstr( 0, "ID", 1, "*", &cid );

    if ( rc == 0 ) {
      (void) sprintf( controlBlock.errormsg,
    "Use ID= to specify NEXRAD station ID; ID=LIST to list available stations");
      controlBlock.returncode = ERR_STAT_NOIMG;
      M0sxdone( &controlBlock );
      exit(1);

    } else if ( rc < 0 ) {
      (void) sprintf( controlBlock.errormsg,
                      "Error parsing NEXRAD station ID" );
      controlBlock.returncode = ERR_STAT_NOIMG;
      M0sxdone( &controlBlock );
      exit(1);
    }

    if ( strcmp(cid, "LIST") == 0 ) {
      (void) sprintf( controlBlock.errormsg,
      "ID=LIST valid for IMGLIST only; use ID= to specify NEXRAD station ID" );
      controlBlock.returncode = ERR_STAT_NOIMG;
      M0sxdone( &controlBlock );
      exit(1);
    }

    (void) sprintf( dbg, "nexraget:: cid = %s", cid );
    M0sxtrce( dbg );

    /*
    ** Check search directory for replacables.  If found replace with values
    ** passed in server invocation
    */

    rc = ReplaceToken( cfginfo.dataloc, "\\ID", (char *) cid );
    rc = ReplaceToken( cfginfo.dataloc, "\\TYPE", (char *) dset_descriptor );

    (void) sprintf( dbg, "nexraget:: data directory mask: %s", cfginfo.dataloc );
    M0sxtrce( dbg );

    /*
    ** Setup the CRITERIA structure
    */

    request.bpos   = bpos;
    request.epos   = epos;
    request.begday = bday;
    request.endday = eday;
    request.begtim = btim;
    request.endtim = etim;
    request.begss  = bss;
    request.endss  = ess;
    (void) strcpy( request.ptype, dset_descriptor );
    (void) strcpy( request.id, cid );
    (void) strcpy( request.filemask, cfginfo.filemask );
    (void) strcpy( request.dirmask, cfginfo.dataloc );

    /*
    ** Get a list of qualified files.  Exit on selection error.
    */

    rc = SelectNexrImages( &request, &satisfy, err );

#if DEBUG
    (void) sprintf( dbg, "nexraget:: SelectNexrImages: rc: %d", rc );
    M0sxtrce( dbg );
#endif

    if ( rc == FAILURE ) {
      (void) strcpy( controlBlock.errormsg, err );
      controlBlock.returncode = ERR_STAT_SELECT;
      M0sxdone( &controlBlock );
      exit(1);
    }

    /*
    ** No images were found.  Set return message and error code and exit.
    */

    if ( satisfy == NULL ) {
      (void) strcpy( controlBlock.errormsg,
                     "No images satisfy the selection criteria" );
      controlBlock.returncode = ERR_STAT_NOIMG;
      M0sxdone( &controlBlock );
      exit(1);
    }

    /*
    ** Loop through the NEXRAD Level III product files that met the
    ** search criteria.
    */

    while ( satisfy != NULL ) {

      /*
      ** Read the AREA directory and navigation codicil for this image
      */

      rc = GetNexrDirs( satisfy, aradir, navcod, auxblk );

      if ( rc == FAILURE ) {
        (void) sprintf( dbg, "nexraget:: error reading area directory: %s  ERR: %s",
                        satisfy->name, err );
        M0sxtrce( dbg );

        satisfy = satisfy->next;
        continue;
      }

      /*
      ** Set AUX block length
      */

      len_auxblk = auxblk[1];

      /*
      ** Fill read specifications with the size of the read buffer,
      ** the min and max number of lines/elements, the size of a
      ** source pixel and the number of bands in the image
      */

      read.bufsiz   = READ_BUFFER_SIZE;
      read.elem_res = elem_res;
      read.line_res = line_res;
      read.maxlin   = aradir[8];
      read.maxele   = aradir[9];
      read.src_len  = sizeof(unsigned char);
      read.numband  = aradir[13];

      /*
      ** Re-adjust the number of lines and elements if needed. a number
      ** of 99999 from the client means that the entire image was requested
      */

#if DEBUG
      (void) sprintf( dbg, "nexraget:: from request:: nlines: %d  nelems: %d",
                      nlines, nelems );
      M0sxtrce( dbg );
#endif

      if ( nlines == 99999 ) {              /* 99999 indicates all lines     */
        nlines = aradir[8];
        orig_nlines = nlines;
      }

      if ( nelems == 99999 ) {              /* 99999 indicates all elements  */
        nelems = aradir[9];
        orig_nelems = nelems;
      }

#if DEBUG
      (void) sprintf( dbg, "nexraget:: from AREA:: nlines: %d  nelems: %d",
                      nlines, nelems );
      M0sxtrce( dbg );
#endif

      /*
      ** Pick up coordinate parameters
      */

      switch ( img_place[0] ) {

        case 'E':                           /* Earth coordinates             */

          rc = Mcargdll( 0, " ", 4, (double) 0,
                            (double) 0, (double) -1, &p_lat, &dum);

          rc = Mcargdll( 0, " ", 5, (double) 0,
                            (double) 0, (double) -1, &p_lon, &dum);

          /*
          ** Convert lat/lon to line/elem
          */

          rc = NavEarthToImg( satisfy, navcod, (float)p_lat, (float)p_lon,
                                 &p_line, &p_elem, err);

          if ( rc == FAILURE ) {
            (void) strcpy( controlBlock.errormsg, err );
            controlBlock.returncode = ERR_STAT_NONAV;
            M0sxdone( &controlBlock );
            exit(1);
          }

          break;

        case 'A':                           /* AREA (file) coordinates       */

          rc = Mcargint( 0, " ", 4, 0, 0, -1, &a_line, &dum );
          rc = Mcargint( 0, " ", 5, 0, 0, -1, &a_elem, &dum );

          p_line = aradir[5] + aradir[11] * a_line;
          p_elem = aradir[6] + aradir[12] * a_elem;

          break;

        default:                            /* Image coordinates             */

          rc = Mcargint( 0, " ", 4, 1, 0, -1, &a_line, &dum );
          rc = Mcargint( 0, " ", 5, 1, 0, -1, &a_elem, &dum );

          p_line = (float) a_line;
          p_elem = (float) a_elem;

      }

#if DEBUG
      (void) sprintf( dbg, "nexraget:: place: %c%c place coord = %f %f",
                      img_place[0], img_place[1], p_line, p_elem );
      M0sxtrce( dbg );
#endif

      /*
      ** Calculate line/elem of upper left corner line/elem
      */

      switch ( img_place[1] ) {

        case 'C':                           /* center location               */

            nj = (nlines - 1) / 2;
            nk = (nelems - 1) / 2;
            break;

        case 'D':                           /* lower right location          */

            nj = p_line - 1;
            nk = p_elem - 1;
            break;

        default:                            /* upper left location           */

            nj = 0;
            nk = 0;

      }

      p_line  = p_line - line_res * nj * aradir[11];
      p_elem  = p_elem - elem_res * nk * aradir[12];

      ul_line = (int) ((p_line - aradir[5]) / aradir[11]);
      ul_elem = (int) ((p_elem - aradir[6]) / aradir[12]);

      (void) sprintf( dbg, "nexraget:: start coord = %d %d", ul_line, ul_elem );
      M0sxtrce( dbg );

      /*
      ** Initialize read struct with starting line and element
      */

      read.ul_line = ul_line;
      read.ul_elem = ul_elem;

      /*
      ** Check the bounds of the returned image to make sure it will
      ** contain data if not, return an error
      */

      rc = UPCCheckImgBounds( &read, orig_nlines, orig_nelems );

      if ( rc == FAILURE ) {
        (void) strcpy( controlBlock.errormsg,
                       "The portion of the image requested does not exist" );
        controlBlock.returncode = ERR_STAT_BLANK_IMG;
        M0sxdone( &controlBlock );
        exit(1);
      }

      /*
      ** Adjust output area directory entries
      */

      rc = Mccydtoiyd(aradir[3], &aradir[3]);    /* Julian date [YYYDDD]     */
      rc = Mccydtoiyd(aradir[45], &aradir[45]);  /* Julian date [YYYDDD]     */

      aradir[5] = aradir[5]+ul_line*aradir[11];  /* upper left image line    */
      aradir[6] = aradir[6]+ul_elem*aradir[12];  /* upper left image element */

      aradir[8] = nlines;                        /* number of lines          */
      aradir[9] = 4 * ((nelems + 3) / 4);        /* number of elements       */
      nelems    = aradir[9];

      if ( spac == 0 ) spac = aradir[10];

      aradir[10] = spac;                         /* # bytes per element      */

      aradir[11] = aradir[11] * line_res;        /* line resolution          */
      aradir[12] = aradir[12] * elem_res;        /* element resolution       */

      /*
      ** Fill read specifications with the size of a destination pixel
      */

      read.des_len = aradir[10];

      /*
      ** Adjust AREA directory entries for unit and scale if AUX= YES
      */

      aradir[52] = lit_( unit, 4 );

      if ( strncmp( adde, "YES", 3 ) == 0 ) {

        rc = GetUnitsAndScale( (char *) &aradir[52],
                               (char *) &aradir[57],
                                        &aradir[58] );
        (void) sprintf( dbg, "nexraget:: unit: %s  scale: %d", unit, aradir[58] );
        M0sxtrce( dbg );

      }

      /*
      ** Read McIDAS calibration codicil
      */

      rc = GetNexrCalCod( satisfy, calcod, &len_calcod );

      if ( rc == FAILURE ) {
        (void) strcpy( controlBlock.errormsg,
                       "Unable to initialize calibration" );
        controlBlock.returncode = ERR_STAT_NOCAL;
        M0sxdone( &controlBlock );
        exit(1);
      }

      /*
      ** Compute total byte transfer to client
      */

      bytes_per_line = aradir[9]  *         /* # elems in a line             */
                       spac       *         /* # bytes per element           */
                       aradir[13] +         /* # spectral bands              */
                       aradir[14];          /* length of line prefix         */

      total_bytes    = MAX_CARD_LEN *       /* # bytes in                    */
                       aradir[63]   +       /*            comment cards      */
                       aradir[33]   +       /* # bytes to start of data      */
                       aradir[8]    *       /* # lines in image              */
                       bytes_per_line;      /* # bytes in a line             */

      /*
      ** Send total bytes in transaction
      */

      (void) sprintf( dbg, "nexraget:: sending %d bytes", total_bytes );
      M0sxtrce( dbg );

      controlBlock.reply_length += total_bytes;

      M0swbyt4( &total_bytes, 1 );
      M0sxsend( 4, &total_bytes );

      /*
      ** Send the synthesized AREA directory
      */

      (void) sprintf( dbg, "nexraget:: sending directory: nlines: %d  nelems: %d",
                      aradir[8], aradir[9] );
      M0sxtrce( dbg );


      /* 
      ** ADDE dataset position:  negated to signal apps that this is a named
      ** dataset 
      */

      aradir[0] = -satisfy->pos;

#if DEBUG
      (void) sprintf( dbg, "nexraget:: named dataset position: %d", aradir[0] );
      M0sxtrce( dbg );
#endif

      /*
      ** Entry 19 is part of the band map. It's value must be 0 for NEXRAD.
      */ 
      aradir[19] = 0;

#if DEBUG
      for ( i = 0; i < 64; i++ ) {
        (void) sprintf( dbg, "aradir[%02d] = %d", i, aradir[i] );
        M0sxtrce( dbg );
      }
#endif

      (void) UPCSendDir( aradir );

      /*
      ** Send the navigation codicil
      */

#if DEBUG
      (void) sprintf( dbg, "nexraget:: sending navigation codicil, nbytes: %d",
                      len_navcod );
      M0sxtrce( dbg );
#endif

      (void) UPCSendCod( navcod, len_navcod );

      /*
      ** Send the calibration codicil if necessary
      */

      if ( len_calcod ) {

#if DEBUG
        (void) sprintf( dbg, "nexraget:: sending calib. codicil, nbytes: %d",
                        len_calcod );
        M0sxtrce( dbg );
#endif

        (void) UPCSendCod( calcod, len_calcod );

      }

      /*
      ** Send the aux block if necessary
      */

      if ( len_auxblk ) {

#if DEBUG
        (void) sprintf( dbg, "nexraget:: sending aux block, nbytes: %d",
                        len_auxblk );
        M0sxtrce( dbg );
#endif

        (void) UPCSendAux( auxblk, len_auxblk );

      }

      /*
      ** Read and send the image one line at a time
      */

      data_bytes  = orig_nelems * read.src_len * read.numband + aradir[14];
      data_bytes *= read.bufsiz;

#if DEBUG
      (void) sprintf( dbg, "nexraget:: data_bytes: %d", data_bytes );
      M0sxtrce( dbg );
#endif

      /*
      ** Total elements needed
      */

      curline       = ul_line;
      begele        = ul_elem;
      bytes_to_send = bytes_per_line;
      bytes_per_pix = read.des_len * read.numband;

      for ( i = 1; i <= orig_nlines; i++ ) {

        /*
        ** If (maxline < curline < 1) fill buffer with zeros and send them
        */

        if ( (curline < 0) || (curline > read.maxlin) ) {

          bytes_to_zero = aradir[9] * read.des_len * read.numband;
          (void) UPCSendZeros( data, bytes_to_zero );
          curline += line_res;
          bytes_to_zero = 0;
          continue;

        }

        /*
        ** If the image requested does not contain data on the left side,
        ** send the appropriate number of zeros and adjust the starting
        ** element
        */

        zeros_on_left = 0;
        nelems = read.maxele;

        if ( ul_elem < 0 ) {

          bytes_to_zero = abs(ul_elem/elem_res) * read.des_len * read.numband;
          zeros_on_left = bytes_to_zero;
          (void) UPCSendZeros( data, bytes_to_zero );

          begele = 0;

          bytes_to_send = bytes_per_line - bytes_to_zero;
          bytes_to_zero = 0;

        }

        /*
        ** If the image requested does not contain data on the right side,
        ** adjust the number of elements. The appropriate number of zero filler
        ** will be send later (after the read)
        */

        if ( ul_elem + (aradir[9] * elem_res) > read.maxele ) {

          bytes_to_zero = (ul_elem/elem_res + aradir[9] -
                          (read.maxele/elem_res)) * read.des_len *
                          read.numband;

          if ( ul_elem > 0 ) nelems = read.maxele - ul_elem/elem_res;

          bytes_to_send = bytes_per_line - bytes_to_zero - zeros_on_left;

#if DEBUG
          (void) sprintf( dbg, "nexraget:: send %d bytes", bytes_to_send );
          M0sxtrce( dbg );

          (void) sprintf( dbg, "nexraget:: zero %d bytes", bytes_to_zero );
          M0sxtrce( dbg );
#endif
        }

        /*
        ** Fill read specifications with beginning line/element and
        ** the number of lines and elements to read
        */

        read.beglin = curline;
        read.begele = begele;
        read.numele = nelems;

#if DEBUG
        (void) sprintf( dbg,
                        "nexraget:: curline = %d, begele = %d, nelems = %d",
                        curline, begele, nelems );
        M0sxtrce( dbg );
#endif

        /*
        ** Read a line of data
        */

        rc = GetNexrLine( satisfy, &read, band, data, err );

        if ( rc == FAILURE ) {
          (void) strcpy( controlBlock.errormsg, err );
          controlBlock.returncode = ERR_STAT_BADLINE;
          M0sxdone( &controlBlock );
          exit(1);
        }

        /*
        ** Calibrate the data
        */

        rc = CalibrateNexrImgData( auxblk, (unsigned char *)data, read.numele, unit );

        /*
        ** Reduce the resolution
        */

#if DEBUG
        (void) sprintf( dbg, "nexraget:: #bytes_to_send: %d  elem_res: %d  bytes/pix: %d",
                        bytes_to_send, elem_res, bytes_per_pix );
        M0sxtrce( dbg );
#endif

        pixels_to_send = bytes_to_send / bytes_per_pix;
        (void) m0resred_( data, &elem_res, &pixels_to_send, &bytes_per_pix );

        /*
        ** Send line of image data
        */

        (void) UPCSendLine( data, bytes_to_send );

        /*
        ** If we still have bytes_to_zero (set because the image is padded
        ** on the right, then we send those
        */

        if ( bytes_to_zero != 0 ) {
          (void) UPCSendZeros( data, bytes_to_zero );
          bytes_to_zero = 0;
        }

        /*
        ** Increment the line pointer
        */

        curline += line_res;
      }

      /*
      ** Send the comment cards and increment to the next list element
      */

      if (aradir[63] != 0) (void) UPCSendCards( aradir, cards );

      satisfy = satisfy->next;

    }

    /*
    ** Bind off the transaction
    */

    M0sxtrce( "nexraget:: binding off transaction" );

    M0sxdone( &controlBlock );

}

/******************************** GetImgUnits ********************************/

int
GetImgUnits( int *aradir, int *auxblk, int *cards, int *num_cards, char *err )

/*
** Name:       GetImgUnits
**
** Purpose:    Build the required ADDE comment cards containing the valid
**             calibration units for this image
**
** Parameters: 
**             aradir    - McIDAS area directory for name
**             cards     - McIDAS comment cards returned as integers
**             num_cards - number of comment cards
**             err       - error string from ReadImgDir
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
** History:    20021127 - Added passing of auxblk
*/

{

    char dbg[256];

    char one_card[MAX_CARD_LEN-1];          /* a single comment card */

    int band;                               /* McIDAS band number */
    int rc;                                 /* function return for UPCAddCard */

    unsigned int bandmask=0;                /* 32-bit map for bands present */
    unsigned int bandmap=0;                 /* 32-bit map for bands present */

#if DEBUG
    M0sxtrce( "in GetImgUnits" );
#endif

    /*
    ** Set bandmap from AREA directory
    */

    bandmap = (unsigned int) aradir[18];

    (void) sprintf( dbg, "GetImgUnits:: bandmask = %d", bandmask );
    M0sxtrce( dbg );

    /*
    ** Provide valid units for each band.  All bands have RAW and BRIT.
    */

    for ( band = 1; band <= MAX_NUM_BAND; band++ ) {

      bandmask = 1 << (band - 1);
      if ( (bandmask & bandmap) == 0 ) continue;

      /*
      ** RAW
      */

      (void) sprintf( one_card, VALID_UNIT_CARD, band,
                      "RAW ", "RAW");
      (*num_cards)++;

      rc = UPCAddCard( one_card, MAX_CARD_LEN, cards, *num_cards );

      if ( rc == 0 ) {
        (void) strcpy( err, "Unable to allocate memory" );
        return FAILURE;
      }

      /*
      ** BRIGHTNESS
      */

      (void) sprintf( one_card, VALID_UNIT_CARD, band,
                      "BRIT", "BRIGHTNESS");
      (*num_cards)++;

      rc = UPCAddCard( one_card, MAX_CARD_LEN, cards, *num_cards );

      if ( rc == 0 ) {
        (void) strcpy( err, "Unable to allocate memory" );
        return FAILURE;
      }

      switch ( bandmask ) {

        case 1:
        case 2:
        case 4:

          /*
          ** Reflectivity
          */

          (void) sprintf( one_card, VALID_UNIT_CARD, band,
                    "ECHO", "REFLECTIVITY");
          (*num_cards)++;

          rc = UPCAddCard( one_card, MAX_CARD_LEN, cards, *num_cards );

          if ( rc == 0 ) {
            (void) strcpy( err, "Unable to allocate memory" );
            return FAILURE;
          }

          /*
          ** VIP level
          */

          (void) sprintf( one_card, VALID_UNIT_CARD, band,
                          "VIP ", "VIP LEVEL");
          (*num_cards)++;

          rc = UPCAddCard( one_card, MAX_CARD_LEN, cards, *num_cards );

          if ( rc == 0 ) {
            (void) strcpy( err, "Unable to allocate memory" );
            return FAILURE;
          }

          break;

        case 8:

          /*
          ** Echo tops
          */

          (void) sprintf( one_card, VALID_UNIT_CARD, band,
                    "TOPS", "ECHO TOPS");
          (*num_cards)++;

          rc = UPCAddCard( one_card, MAX_CARD_LEN, cards, *num_cards );

          if ( rc == 0 ) {
            (void) strcpy( err, "Unable to allocate memory" );
            return FAILURE;
          }

          break;

        case 16:
        case 32:
        case 64:

          /*
          ** Precipitation
          */

          (void) sprintf( one_card, VALID_UNIT_CARD, band,
                    "RAIN", "PRECIPITATION");
          (*num_cards)++;

          rc = UPCAddCard( one_card, MAX_CARD_LEN, cards, *num_cards );

          if ( rc == 0 ) {
            (void) strcpy( err, "Unable to allocate memory" );
            return FAILURE;
          }

          break;

        case 128:

          /*
          ** Vertically Integrated Water
          */

          (void) sprintf( one_card, VALID_UNIT_CARD, band,
                    "H2O ", "VERTICAL H2O");
          (*num_cards)++;

          rc = UPCAddCard( one_card, MAX_CARD_LEN, cards, *num_cards );

          if ( rc == 0 ) {
            (void) strcpy( err, "Unable to allocate memory" );
            return FAILURE;
          }

          break;

        case 256:

          /*
          ** Radial Velocity
          */

          (void) sprintf( one_card, VALID_UNIT_CARD, band,
                    "VEL ", "RADIAL VELOCITY");
          (*num_cards)++;

          rc = UPCAddCard( one_card, MAX_CARD_LEN, cards, *num_cards );

          if ( rc == 0 ) {
            (void) strcpy( err, "Unable to allocate memory" );
            return FAILURE;
          }

          break;

        case 512:

          /*
          ** Storm Relative Radial Velocity
          */

          (void) sprintf( one_card, VALID_UNIT_CARD, band,
                    "SRMV", "RADIAL VELOCITY");
          (*num_cards)++;

          rc = UPCAddCard( one_card, MAX_CARD_LEN, cards, *num_cards );

          if ( rc == 0 ) {
            (void) strcpy( err, "Unable to allocate memory" );
            return FAILURE;
          }

          break;

        case 1024:

          /*
          ** Digital Precipitation Array
          */

          (void) sprintf( one_card, VALID_UNIT_CARD, band,
                    "DBA ", "DIGITAL PRECIP");
          (*num_cards)++;

          rc = UPCAddCard( one_card, MAX_CARD_LEN, cards, *num_cards );

          if ( rc == 0 ) {
            (void) strcpy( err, "Unable to allocate memory" );
            return FAILURE;
          }

          break;

        case 2048:

          /*
          ** Differential reflectivity
          */

          (void) sprintf( one_card, VALID_UNIT_CARD, band,
                    "DZD ", "DIFF REFLECT");
          (*num_cards)++;

          rc = UPCAddCard( one_card, MAX_CARD_LEN, cards, *num_cards );

          if ( rc == 0 ) {
            (void) strcpy( err, "Unable to allocate memory" );
            return FAILURE;
          }

          break;

        case 4096:

          /*
          ** Instantaneous precipitation rate
          */

          (void) sprintf( one_card, VALID_UNIT_CARD, band,
                    "DPR ", "PRECIP RATE");
          (*num_cards)++;

          rc = UPCAddCard( one_card, MAX_CARD_LEN, cards, *num_cards );

          if ( rc == 0 ) {
            (void) strcpy( err, "Unable to allocate memory" );
            return FAILURE;
          }

          break;

        case 8192:

          /*
          ** Correlation coefficient
          */

          (void) sprintf( one_card, VALID_UNIT_CARD, band,
                    "DCC ", "CORR COEF");
          (*num_cards)++;

          rc = UPCAddCard( one_card, MAX_CARD_LEN, cards, *num_cards );

          if ( rc == 0 ) {
            (void) strcpy( err, "Unable to allocate memory" );
            return FAILURE;
          }

          break;

        case 16384:

          /*
          ** Specific differential phase
          */

          (void) sprintf( one_card, VALID_UNIT_CARD, band,
                    "DKD ", "DIFF PHASE");
          (*num_cards)++;

          rc = UPCAddCard( one_card, MAX_CARD_LEN, cards, *num_cards );

          if ( rc == 0 ) {
            (void) strcpy( err, "Unable to allocate memory" );
            return FAILURE;
          }

          break;

        case 32768:

          /*
          ** Hydrometeor classification
          */

          (void) sprintf( one_card, VALID_UNIT_CARD, band,
                    "HC  ", "HYDRO CLASS");
          (*num_cards)++;

          rc = UPCAddCard( one_card, MAX_CARD_LEN, cards, *num_cards );

          if ( rc == 0 ) {
            (void) strcpy( err, "Unable to allocate memory" );
            return FAILURE;
          }

          break;

        case 65536:

          /*
          ** Digital Accumulation Array
          */

          (void) sprintf( one_card, VALID_UNIT_CARD, band,
                    "DAA ", "DIG ACCUM ARRAY");
          (*num_cards)++;

          rc = UPCAddCard( one_card, MAX_CARD_LEN, cards, *num_cards );

          if ( rc == 0 ) {
            (void) strcpy( err, "Unable to allocate memory" );
            return FAILURE;
          }

          break;

        case 262144:                        /* <<<<< UPC add 20150619 >>>>> */

          /*
          ** Base Spectrum Width
          */

          (void) sprintf( one_card, VALID_UNIT_CARD, band,
                    "BSW", "BASE SPECTRUM WIDTH");
          (*num_cards)++;

          rc = UPCAddCard( one_card, MAX_CARD_LEN, cards, *num_cards );

          if ( rc == 0 ) {
            (void) strcpy( err, "Unable to allocate memory" );
            return FAILURE;
          }

          break;

      }

    }

    return SUCCESS;

}
