/**** $Id: nexrutil.c,v 1.20 2022/04/05 21:55:50 tyoksas Tst $ ****/
#include <sys/types.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <math.h>
#include <time.h>
#include <sys/stat.h>
#include "mcidas.h"
#include "nexrutil.h"
#include "stations.h"
#include "zlib.h"
#include "zutil.h"

#define DEBUG_CALIB 0
#define DEBUG_CALL  0
#define DEBUG_CODLU 0
#define DEBUG_DIRS  0
#define DEBUG_DVDLN 0
#define DEBUG_FLT16 0
#define DEBUG_FLT32 0
#define DEBUG_GTBUF 0
#define DEBUG_INFO  0
#define DEBUG_ISBZP 0
#define DEBUG_ISENC 0
#define DEBUG_ISZLB 0
#define DEBUG_LFMRS 0
#define DEBUG_LINE  0
#define DEBUG_MCCAL 0
#define DEBUG_MSGHD 0
#define DEBUG_PRDES 0
#define DEBUG_RADIM 0
#define DEBUG_RADRS 0
#define DEBUG_RASIM 0
#define DEBUG_SELIM 0
#define DEBUG_STNID 0
#define DEBUG_TEST  0
#define DEBUG_TIME  0
#define DEBUG_ZLIBD 0

/* <<<<< UPC add 20081204 - add needed function prototypes >>>>> */
Fint lit_( const char *, FsLen );
float GetFloat16( unsigned short );
float GetFloat32( unsigned short, unsigned short );

static char dbg[MAX_ERR_LEN];               /* debug message container       */

Msghead                  header;            /* product message header        */
Proddesc                 product;           /* product description           */
Prodinfo                 prod_info;         /* product metadata              */
product_symbology_header ps_head;           /* product symbology header      */

radial_packet_header     radp_head={0};     /* packet header for radial data */
raster_packet_header     rasp_head={0};     /* packet header for raster data */
lfmras_packet_header     lfm_head={0};      /* packet header for LFM raster data */


/*************************** CalibrateNexrImgData ****************************/

int
CalibrateNexrImgData( int *auxblk, unsigned char *data, int npixel, const char *unit )

/*
** Name:    CalibrateNexrImgData
**
** Purpose: Convert raw NEXRAD Level III product data to requested unit
**          in place
**
** Parameters:
**          aradir - synthesized McIDAS AREA directory
**          auxblk - synthesized INFO supplementary directory
**          data   - buffer of data to convert
**          npixel - number of bytes to convert
**          unit   - unit to convert to
**
** Returns:
**          SUCCESS == 1
**          FAILURE == 0
**
*/

{

    int *cal;                               /* pointer to calibration array  */

    int  i, j;                              /* generic counters              */

    int  ival;                              /* calibrated value              */
    int  mode;                              /* radar operation mode          */
    int  nlevel;                            /* # aux block data levels       */
    int  offset;                            /* offset for aux block data lev */
    int  iscale;                            /* scale for aux block data lev  */
    int  dvip[]={ 0, 30, 40, 45, 50, 55 };  /* DVIP break points             */
    int *levels=NULL;                       /* data levels from aux block    */

    char cunit[5];

    float  dbz;                             /* echo strength [dbZ]           */
    float ratio;                            /* data value bin widths         */
    float scale;                            /* scale for aux block data lev  */

#if DEBUG_CALIB || DEBUG_CALL
    M0sxtrce( "In CalibrateNexrImgData" );
#endif

    /*
    ** Extract values from AUX block
    */

    mode   = auxblk[24];
    nlevel = auxblk[40];
    offset = auxblk[41];
    iscale = auxblk[42];
    scale  = iscale;
    ratio  = 256.0 / nlevel;

#if DEBUG_CALIB
    (void) sprintf( dbg, "CalibrateNexrImgData:: input unit is %s", unit );
    M0sxtrce( dbg );
#endif

    (void) strncpy( cunit, unit, 4 );
    *(cunit+4) = 0;

#if DEBUG_CALIB
    (void) sprintf( dbg,
    "CalibrateNexrImgData:: mode: %d unit: %s  npixel: %d  nlevel: %d  scale: %d  offset: %d",
    mode, cunit, npixel, nlevel, iscale, offset );
    M0sxtrce( dbg );
#endif

    /*
    ** Allocate levels array
    */

    if ( nlevel <= 256 ) {
      levels = (int *) malloc ( nlevel * sizeof( int ) );
      if ( levels == (int *) NULL ) {
        M0sxtrce( "Leave CalibrateNexrImgData:: ERROR levels malloc" );
        return FAILURE;
      }

      for ( i = 0; i < nlevel; i++ ) {
        levels[i] = auxblk[43+i];
#if DEBUG_CALIB
        (void) sprintf( dbg, "CalibrateNexrImgData:: levels[%d]: %d", i, levels[i] );
        M0sxtrce( dbg );
#endif
      }
    }

    /*
    ** Allocate calibration array
    */

    cal = (int *) malloc( npixel * sizeof(int) );

    if ( cal == (int *) NULL ) {
      (void) memset( data, 0, npixel );
      M0sxtrce( "CalibrateNexrImgData:: Error allocating calibration array" );
      free( levels );
      M0sxtrce( "Leave CalibrateNexrImgData:: ERROR cal array malloc" );
      return ERR_STAT_MALLOC;
    }

    /*
    ** Convert raw image data values to 'unit' in place
    */

    if ( strncmp( unit, "BRIT", 4) == 0 ) {           /* Brightness           */

      for ( i = 0; i < npixel; i++ ) {
        data[i] *= ratio;
      }

    } else if ( strncmp( unit, "ECHO", 4 ) == 0 ) {   /* Reflectivity         */

      for ( i = 0; i < npixel; i++ ) {
#if DEBUG_CALIB
        (void) sprintf( dbg, "CalibrateNexrImgData:: data[%d]: %d", i, data[i] );
        M0sxtrce( dbg );
#endif
        ival = levels[data[i]];
        /* <<<<< UPC mod 20100521 - return floating point ECHO
                                    corresponding mod in servutil.c >>>>> */ 
        if ( ival > -9997 )
          cal[i] = 10 * (ival / scale + offset);
        else
          cal[i] = 10 * ival;
      }
      (void) M0swbyt4( cal, npixel );
      (void) memcpy( data, cal, npixel*sizeof(int) );

    } else if ( strncmp( unit, "VIP", 3 ) == 0 ) {    /* VIP level            */

      for ( i = 0; i < npixel; i++ ) {
        dbz = levels[data[i]] / scale + offset;
        for ( j = 0; j <= 5; j++ ) {
          if ( dbz > dvip[j] ) data[i] = j + 1;
        }
      }

    } else if ( strncmp( unit, "TOPS", 4 ) == 0 ) {   /* Echo tops            */

      for ( i = 0; i < npixel; i++ ) {
        ival = levels[0x7F&data[i]];
        if ( ival > -9996 )
          data[i] = ival / iscale + offset;
        else
          data[i] = ival;
      }

    } else if ( strncmp( unit, "RAIN", 4 ) == 0 ) {   /* Precip amount        */

      for ( i = 0; i < npixel; i++ ) {
        ival = levels[data[i]];
        if ( ival > -9996 )
          cal[i] = 100 * (ival / scale + offset);
        else
          cal[i] = 100 * ival;
      }
      (void) M0swbyt4( cal, npixel );
      (void) memcpy( data, cal, npixel*sizeof(int) );

    } else if ( strncmp( unit, "H2O", 3 ) == 0 ) {    /* Vertical H2O       */

      /* <<<<< UPC mod 20100825 - return floating point VIL/DVL
               associated changes made in servuitil.c (for 100 scale factor)
               and nexraget.cp (spac changed from 1 to 4) >>>>> */
      for ( i = 0; i < npixel; i++ ) {
        ival = levels[data[i]];
        if ( ival > -9996 )
          cal[i] = 100 * (ival / scale + offset);
        else
          cal[i] = 100 * ival;
      }
      (void) M0swbyt4( cal, npixel );
      (void) memcpy( data, cal, npixel*sizeof(int) );

    } else if ( strncmp( unit, "VEL", 3 ) == 0 ) {    /* Radial velocity      */

      for ( i = 0; i < npixel; i++ ) {
        ival = levels[data[i]];
        if ( ival > -9996 )
          cal[i] = ival / iscale + offset;
        else
          cal[i] = ival;
      }
      (void) M0swbyt4( cal, npixel );
      (void) memcpy( data, cal, npixel*sizeof(int) );

    } else if ( strncmp( unit, "SRMV", 4 ) == 0 ) {   /* Strm rel rad vel     */

      for ( i = 0; i < npixel; i++ ) {
        ival = levels[data[i]];
        if ( ival > -9996 )
          cal[i] = ival / iscale + offset;
        else
          cal[i] = ival;
      }
      (void) M0swbyt4( cal, npixel );
      (void) memcpy( data, cal, npixel*sizeof(int) );

    } else if ( strncmp( unit, "DBA", 3 ) == 0 ) {    /* DBA ??              */

      /* <<<<< UPC add 20100830 - add missing cal section for dBA >>>>> */
      for ( i = 0; i < npixel; i++ ) {
        ival = levels[data[i]];
        if ( ival > -9996 )
          cal[i] = 10 * (ival / scale + offset);
        else
          cal[i] = 10 * ival;
      }
      (void) M0swbyt4( cal, npixel );
      (void) memcpy( data, cal, npixel*sizeof(int) );

    } else if ( strncmp( unit, "DZD", 3 ) == 0 ) {    /* 159/DZD: Diff reflectivity */

      for ( i = 0; i < npixel; i++ ) {
        ival = levels[data[i]];
        if ( ival > -9990 )
          cal[i] = 10 * (ival / scale + offset);
        else
          cal[i] = 10 * ival;
      }
      (void) M0swbyt4( cal, npixel );
      (void) memcpy( data, cal, npixel*sizeof(int) );

    } else if ( strncmp( unit, "DCC", 3 ) == 0 ) {   /* 161/DCC: Cor Coefficient */

      for ( i = 0; i < npixel; i++ ) {
        ival = levels[data[i]];
        if ( ival > -9996 )
          cal[i] = 100 * (ival / scale + offset);
        else
          cal[i] = 100 * ival;
      }
      (void) M0swbyt4( cal, npixel );
      (void) memcpy( data, cal, npixel*sizeof(int) );

    } else if ( strncmp( unit, "DKD", 3 ) == 0 ) {   /* 163/DKD: Spec Diff Phase */

      for ( i = 0; i < npixel; i++ ) {
        ival = levels[data[i]];
        if ( ival > -9996 )
          cal[i] = 10 * (ival / scale + offset);
        else
          cal[i] = 10 * ival;
      }
      (void) M0swbyt4( cal, npixel );
      (void) memcpy( data, cal, npixel*sizeof(int) );

    } else if ( strncmp( unit, "HC", 2 ) == 0 ) {    /* 165/DHC: Hydrometr Class */

      for ( i = 0; i < npixel; i++ ) {
        ival = levels[data[i]];
        cal[i] = ival/10;
      }
      (void) M0swbyt4( cal, npixel );
      (void) memcpy( data, cal, npixel*sizeof(int) );

    } else if ( strncmp( unit, "HHC", 3 ) == 0 ) {   /* 177/HHC: Dig Hydrometr Class */

      for ( i = 0; i < npixel; i++ ) {
        ival = levels[data[i]];
        cal[i] = ival/10;
      }
      (void) M0swbyt4( cal, npixel );
      (void) memcpy( data, cal, npixel*sizeof(int) );

    } else if ( strncmp( unit, "DAA", 3 ) == 0 ) {  /* 170/DAA: Dig Accum Array */

      for ( i = 0; i < npixel; i++ ) {
        ival = levels[data[i]];
        if ( ival > -9999 )
          cal[i] =  100 * (ival / scale + offset);
        else
          cal[i] =  100 *  ival;
      }
      (void) M0swbyt4( cal, npixel );
      (void) memcpy( data, cal, npixel*sizeof(int) );

    } else if ( strncmp( unit, "DPR", 3 ) == 0 ) {   /* 176/DPR: Dig Inst Precip Rate */
                                                     /* !!!!! Not yet implemented !!!!! */

      for ( i = 0; i < npixel; i++ ) {
        if ( ival > -9996 )
          cal[i] = 10 * (data[i] / scale + offset);
        else
          cal[i] = 10 * data[i];
      }
      (void) M0swbyt4( cal, npixel );
      (void) memcpy( data, cal, npixel*sizeof(int) );

    } else if ( strncmp( unit, "BSW", 3 ) == 0 ) {   /* Base Spectrun Width  */
                                                     /* <<<<< UPC add 20150619 >>>>> */

      for ( i = 0; i < npixel; i++ ) {
        ival = levels[data[i]];
        if ( ival > -9996 )
          cal[i] = ival / iscale + offset;
        else
          cal[i] = ival;
      }
      (void) M0swbyt4( cal, npixel );
      (void) memcpy( data, cal, npixel*sizeof(int) );

    }

    free( levels );
    free( cal );

    /* Done */
#if DEBUG_CALIB
    M0sxtrce(( "Leave CalibrateNexrImgData:: SUCCESS" );
#endif
    return SUCCESS;

}

/******************************* GetNexrCalCod ********************************/

int
GetNexrCalCod( FILELIST *cur, int *calcod, int *len )

/*
** Name:       GetNexrCalCod
**
** Purpose:    Return a McIDAS calibration codicil for NEXRAD Level III products
**
** Parameters:
**             cur     - FILELIST structure containing current image information
**                       name  - full qualified name of the file
**                       pos   - position to add to list
**                       doff  - byte offset to beginning of image data
**                       size  - file size in bytes
**                       ctime - file creation time [sec since 1970]
**                       id    - NEXRAD Level III station ID
**                       time  - file data time [sec since 1970]
**             calcod  - output McIDAS cal codicil
**             len     - byte length of output McIDAS cal codicil
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
*/

{

#if DEBUG_MCCAL || DEBUG_CALL
    M0sxtrce( "In GetNexrCalCod" );
#endif

    /*
    ** NEXRAD Level III products have no calibration codicil
    */

    *len = 0;

#if DEBUG_MCCAL || DEBUG_CALL
    M0sxtrce( "Leave GetNexrCalCod:: SUCCESS" );
#endif

    return SUCCESS;

}


/******************************* GetNexrDirs *********************************/

int
GetNexrDirs( FILELIST *cur, int *aradir, int *navcod, int *auxblk )

/*
** Name:       GetNexrDirs
**
** Purpose:    Read a NEXRAD Level III product and create corresponding
**             McIDAS area directory
**
** Parameters:
**             cur     - FILELIST structure containing current image information
**                       name  - full qualified name of the file
**                       pos   - position to add to list
**                       hoff  - byte offset to beginning of header information
**                       doff  - byte offset to beginning of image data
**                       size  - file size in bytes
**                       type  - file type
**                       id    - NEXRAD Level III station ID
**                       time  - file data time [sec since 1970]
**             aradir  - McIDAS area directory for image
**             navcod  - McIDAS navigation codicil for image
**             auxblk  - McIDAS auxiliary block for image
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
** NOTES:
**             The indices for aradir in this function are zero-based.
**
*/

{

    char           cmemo[33];               /* memo field                    */
    char          *cmode[]={"Maintenance", "Clear Air", "Precip Mode"};
    char           ctilt[9];                /* radar tilt or layer           */
    char           ctitle[41];              /* AUX block memo field          */
    char           cunit[9];                /* product data unit             */

    int            auxsiz;                  /* AUX block size in words       */
    int            hedsiz;                  /* NEXRAD header size            */

    int            i;                       /* loop variable                 */
    int            isign;                   /* sign: +-1                     */
    int            jsign;                   /* sign: +-1                     */
    int            radial=0;                /* radial/raster product flag    */
    int            rc;                      /* function return status        */

    float          lscale;                  /* linear scale 16-bit float     */
    float          lofset;                  /* linear offset 16-bit float    */
    float          gscale;                  /* log scale 16-bit float        */
    float          gofset;                  /* log offset 16-bit float       */
    float          dscale;                  /* default scale float           */

    int            update_rate[]={10, 6, 5};/* clear air/precip/storm update */

    double         dlat;                    /* radar latitude                */
    double         dlon;                    /* radar longitude               */

    Fint4          ival;                    /* holder for data value         */

    time_t         datetime;                /* holds time in seconds         */
    struct tm     *settime;                 /* time structure pointer        */

    Fint2          divider;                 /* layer divider flag            */
    Fint4          layer_length;            /* length of data layer in bytes */
    DivLen         divlen;
    Mcuint2        rcode;                   /* radial/raster indicator       */
    float          scale;                   /* data/raster resolution ratio  */

    FILE          *fd;                      /* file descriptor for 'fread'   */
    unsigned int   destLen;                 /* malloced length of dest       */
    unsigned char *dest=NULL;               /* pointer to allocated memory   */

    BUFREAD       *bp=NULL;                 /* buffer descriptor             */

#if DEBUG_DIRS || DEBUG_CALL
    M0sxtrce( "In GetNexrDirs" );
#endif

    /*
    ** Initialize AREA directory, NAVigation codicil, and AUX block arrays
    */

    for ( i = 0; i < IMG_DIR_LEN; i++ ) {
      aradir[i] = 0;
    }

    for ( i = 0; i < NAV_COD_LEN; i++ ) {
      navcod[i] = 0;
    }

    for ( i = 0; i < AUX_BLK_LEN; i++ ) {
      auxblk[i] = 0;
    }


    /*************************************************************************/
    /*                          AREA directory                               */
    /*************************************************************************/

    /*
    ** Fill synthesized AREA directory with invariant values
    */

    aradir[ 0] =      0;                    /* always 0                     */
    aradir[ 1] =      4;                    /* always 4                     */
    aradir[ 2] =      7;                    /* McIDAS ss number for radar   */
    aradir[ 5] =      1;                    /* upper left line              */
    aradir[ 6] =      1;                    /* upper left element           */
    aradir[10] =      1;                    /* bytes per pixel              */
    aradir[11] =      1;                    /* line resolution              */
    aradir[12] =      1;                    /* element resolution           */
    aradir[13] =      1;                    /* number of bands              */
    aradir[14] =      0;                    /* length of line prefix        */
    aradir[15] =   1234;                    /* fake project number          */
    aradir[20] = lit_( "    ", 4 );         /* blank out radar ID           */

    /*
    ** Byte locations of navigation, calibration, auxiliary, and data blocks
    */

    aradir[34] = IMG_DIR_LEN * 4;           /* byte offset to nav block     */
    aradir[59] = aradir[34] + 4*NAV_COD_LEN;/* byte offset to aux block     */
    aradir[62] = aradir[59];                /* byte offset to cal block     */

    /*
    ** Source and calibration type
    */

    aradir[51] = lit_( "NEXR", 4 );         /* source type                  */
    aradir[52] = lit_( "RAW ", 4 );         /* calibration type             */

    /*
    ** Read in the contents of the NEXRAD Level III product
    */

#if DEBUG_DIRS
    (void) sprintf( dbg, "GetNexrDirs:: call ReadNexrInfo: NEXR_FILE_READ = %d",
                    NEXR_FILE_READ );
    M0sxtrce( dbg );
#endif

    bp = ReadNexrInfo( cur, (int) NEXR_FILE_READ );
    if ( bp == (BUFREAD *) NULL ) {
      (void) sprintf( dbg, "GetNexrDirs:: ERROR processing NEXRAD file: %s",
                      cur->name );
      M0sxtrce( dbg );
#if DEBUG_DIRS
      M0sxtrce( "Leave GetNexrDirs FAILURE" );
#endif
      return FAILURE;
    }

    bp->_ptr += cur->hoff;
    bp->_cnt -= cur->hoff;
    hedsiz    = cur->hoff;

#if DEBUG_DIRS
    (void) sprintf( dbg, "GetNexrDirs:: cur->hoff: %d", cur->hoff );
    M0sxtrce( dbg );
#endif

    /*
    ** Get product message header
    **
    ** Returns:
    **
    **        Bytes Element          Meaning        Range
    **          2   header.code    - code        -131 - -16,  0 - 211
    **          2   header.date    - Julian day     1 - 32767
    **          4   header.time    - time [s]       0 - 86399
    **          4   header.length  - length        18 - 1329270
    **          2   header.source  - source ID      0 - 999
    **          2   header.dest    - dest ID        0 - 999
    **          2   header.blocks  - # blocks       1 - 51  (includes this one)
    **         ----------------------------------------------------------------
    **         18
    */

    rc = read_msghead( bp, &header );
    if ( rc == FAILURE ) {
#if DEBUG_DIRS
      M0sxtrce( "Leave GetNexrDirs:: ERROR reading mesghead" );
#endif
      return FAILURE;
    }

    hedsiz += 18;

#if DEBUG_DIRS
    (void) sprintf( dbg,
        "GetNexrDirs:: header code %d, date %d, time %d, length %d, blocks %d",
        header.code, header.date, header.time, header.length, header.blocks );
    M0sxtrce( dbg );
#endif

    /*
    ** Read product description
    **
    ** Returns:
    **
    **        Bytes Element              Meaning        Range
    ** 10       2   product.divider     - divider         -1
    ** 11-12    4   product.latitude    - latitude
    ** 13-14    4   product.longitude   - longitude
    ** 15       2   product.height      - height of radar
    ** 16       2   product.code        - product code
    ** 17       2   product.mode        - radar operational mode
    ** 18       2   product.coverage    - radar coverage pattern
    ** 19       2   product.seqnum      - sequence #
    ** 20       2   product.vs_num      - vol scan #
    ** 21       2   product.vs_date     - vol scan date
    ** 22-23    4   product.vs_time     - vol scan Time
    ** 24       2   product.prod_date   - generation date
    ** 25-26    4   product.prod_time   - generation time
    ** 27       2   product.p1          - product dependent
    ** 28       2   product.p2          - product dependent
    ** 29       2   product.elev_num    - elevation number
    ** 30       2   product.p3          - product dependent
    ** 31-46 16*2   product.threshold   - threshold values
    ** 47       2   product.p4          - product dependent
    ** 48       2   product.p5          - product dependent
    ** 49       2   product.p6          - product dependent
    ** 50       2   product.p7          - product dependent
    ** 51       2   product.p8          - product dependent
    ** 52       2   product.p9          - product dependent
    ** 53       2   product.p10         - product dependent
    ** 54       2   product.num_maps    - Version, Spot Blank
    ** 55-56    4   product.off_symbol  - offset to symbology block
    ** 57-58    4   product.off_graphic - offset to graphic alphanumeric block
    ** 59-60    4   product.off_tabular - offset to tabular alphanumeric block
    **        ----------------------------------------------------------------
    **        102
    */

    rc = read_proddesc( bp, &product );

    if ( rc == FAILURE || product.divider != -1 ) {
#if DEBUG_DIRS
      M0sxtrce( "Leave GetNexrDirs:: ERROR reading proddesc" );
#endif
      return FAILURE;
    }

    hedsiz += 102;

#if DEBUG_DIRS
    for ( i=0; i<16; i++ ) {
      (void) sprintf( dbg, "GetNexrDirs:: product.threshold[%d] = %d",
                      i, product.threshold[i] );
      M0sxtrce( dbg );
    }
#endif

    /*
    ** Set product-dependent information
    */

    if ( !strcmp( cur->id, "???" ) )
      rc = StnIdFromLatLon( product.latitude, product.longitude, cur->id );

#if DEBUG_DIRS
    (void) sprintf( dbg,
    "GetNexrDirs:: ID=%3s LATLON=%6.3f %7.3f ULATLON=%6.3f %7.3f ELE= UELE= ST= CO=US NAME='TDWR' DATATYPE= NEXRAD",
                    cur->id, product.latitude/1000., -product.longitude/1000.,
                             product.latitude/1000., -product.longitude/1000. );
    M0sxtrce( dbg );
#endif

#if DEBUG_DIRS
    (void) sprintf( dbg, "GetNexrDirs:: NEXRAD id: %s", cur->id );
    M0sxtrce( dbg );

    (void) sprintf( dbg, "GetNexrDirs:: product.code = %d", product.code );
    M0sxtrce( dbg );
#endif

    (void) memcpy( &aradir[20], cur->id, strlen(cur->id) );

    code_lookup( product.code, &prod_info );

    prod_info.elevation  = product.p3;   /* default mod in specific products */
    prod_info.max        = product.p4;   /*                "                 */
    prod_info.scale      = 1.0;          /* default min/max scale factor     */

    switch ( product.code ) {

      case 16:             /* Base Reflectivity .54x1 NmixDeg 124   8 Radial */
      case 19:             /* Base Reflectivity .54x1 NmixDeg 124  16 Radial */

        prod_info.max        = product.p4;   /* Maximum reflectivity         */
        aradir[18]           = 1;

        (void) sprintf( cmemo, "Base Reflct %4.2f DEG %s",
                        prod_info.elevation/10., cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+12, 8 );
        (void) strcpy( ctitle, "BREF: Base Reflectivity" );
        (void) strcpy( cunit, "DBZ" );

        break;

      case 17:             /* Base Reflectivity 1.1x1 NmixDeg 248   8 Radial */
      case 18:             /* Base Reflectivity 2.2x1 NmixDeg 248   8 Radial */
      case 20:             /* Base Reflectivity 1.1x1 NmixDeg 248   8 Radial */
      case 21:             /* Base Reflectivity 2.2x1 NmixDeg 248   8 Radial */

        prod_info.max        = product.p4;   /* Maximum reflectivity         */
        aradir[18]           = 1;

        (void) sprintf( cmemo, "BaseRef 248 %4.2f DEG %s",
                        prod_info.elevation/10., cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+12, 8 );
        (void) strcpy( ctitle, "BREF: 248 nm Base Reflectivity" );
        (void) strcpy( cunit, "DBZ" );

        break;

      case 32:     /* Digital Hybrid Reflect     .54x1 NmixDeg 124 256 Radial */

        prod_info.max        = product.p4;   /* Maximum reflectivity         */
        prod_info.start_time = (product.p5 - 1) * 86400 + product.p6 * 60;
        aradir[18]           = 1;

        (void) sprintf( cmemo, "DigHyb Refl %4.2f DEG %s",
                        prod_info.elevation/10., cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+12, 8 );
        (void) strcpy( ctitle, "BREF: Digital Hybrid Reflect." );
        (void) strcpy( cunit, "DBZ" );

        break;

      case 35:           /* Comp Reflectivity .54x.54 NmixNmi 124   8 Raster */
      case 36:           /* Comp Reflectivity 2.2x2.2 NmixNmi 248   8 Raster */
      case 37:           /* Comp Reflectivity .54x.54 NmixNmi 124  16 Raster */
      case 38:           /* Comp Reflectivity 2.2x2.2 NmixNmi 248  16 Raster */

        prod_info.elevation  = -1;
        prod_info.max        = product.p4;   /* Maximum reflectivity         */
        aradir[18]           = 2;

        (void) sprintf( cmemo, "Composite Reflect %s", cmode[product.mode] );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "CREF Composite Reflectivity" );
        (void) strcpy( cunit, "DBZ" );

        break;

      case 94:     /* Base Reflectivity DR       .54x1 NmixDeg 248 256 Radial */

        prod_info.max        = product.p4;   /* Maximum reflectivity         */
        aradir[18]           = 1;

        (void) sprintf( cmemo, "BaseRef DR  %4.2f DEG %s",
                        prod_info.elevation/10., cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+12, 8 );
        (void) strcpy( ctitle, "BREF: 94/DR Base Reflectivity" );
        (void) strcpy( cunit, "DBZ" );

        break;

      case 153:    /* Base Reflectivity .13nm X .5deg 256LVL 153/SDR */

        prod_info.max        = product.p4;   /* Maximum reflectivity         */
        aradir[18]           = 1;

        (void) sprintf( cmemo, "BaseRef SupRes  %4.2f DEG %s",
                        prod_info.elevation/10., cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+12, 8 );
        (void) strcpy( ctitle, "BREF: 153/SupRes Base Reflectivity" );
        (void) strcpy( cunit, "DBZ" );

        break;

      case 180:    /* TDWR Base Reflectivity TRx .08x1 NmixDeg  48 256 Radial */

        prod_info.max        = product.p4;
        aradir[18]           = 1;

        (void) sprintf( cmemo, "BaseRef TRx %4.2f DEG %s",
                        prod_info.elevation/10., cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+12, 8);
        (void) strcpy( ctitle, "BREF: TDWR Base Reflectivity" );
        (void) strcpy( cunit, "DBZ" );

        break;

      case 181:   /* TDWR Base Reflectivity TRx .08x1 NmixDeg  48  16 Radial */

        prod_info.max        = product.p4;   /* Maximum reflectivity         */
        aradir[18]           = 1;

        (void) sprintf( cmemo, "BaseRef TRx %4.2f DEG %s",
                        prod_info.elevation/10., cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+12, 8 );
        (void) strcpy( ctitle, "BREF: TDWR Base Reflectivity" );
        (void) strcpy( cunit, "DBZ" );

        break;

      case 186:   /* TDWR Base Reflectivity TZL .16x1 NmixDeg 149 256 Radial */

        prod_info.max        = product.p4;   /* Maximum reflectivity         */
        aradir[18]           = 1;

        (void) sprintf( cmemo, "BaseRef TZL %4.2f DEG %s",
                        prod_info.elevation/10., cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+12, 8 );
        (void) strcpy( ctitle, "BREF: 149 nm Base Reflectivity" );
        (void) strcpy( cunit, "DBZ" );

        break;

      case 187:   /* TDWR Base Reflectivity TZx .16x2 NmixDeg 149  16 Radial */

        prod_info.max        = product.p4;   /* Maximum reflectivity         */
        aradir[18]           = 1;

        (void) sprintf( cmemo, "BaseRef TZL %4.2f DEG %s",
                        prod_info.elevation/10., cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+12, 8 );
        (void) strcpy( ctitle, "BREF: 149 nm Base Reflectivity" );
        (void) strcpy( cunit, "DBZ" );

        break;

      case 63:     /* Layer Comp Reflectivity 2.2x2.2 NmixNmi 124   8 Raster */
      case 64:     /* Layer Comp Reflectivity 2.2x2.2 NmixNmi 124   8 Raster */
      case 65:     /* Layer Comp Reflectivity 2.2x2.2 NmixNmi 124   8 Raster */
      case 66:     /* Layer Comp Reflectivity 2.2x2.2 NmixNmi 124   8 Raster */
      case 67:     /* Layer Comp Reflectivity 2.2x2.2 NmixNmi 124   8 Raster */
      case 89:     /* Layer Comp Reflectivity 2.2x2.2 NmixNmi 124   8 Raster */
      case 90:     /* Layer Comp Reflectivity 2.2x2.2 NmixNmi 124   8 Raster */

        prod_info.max        = product.p4;   /* Maximum reflectivity         */
        prod_info.elevation  = product.p5;
        prod_info.top        = product.p6;   /* Top of layer                 */
        aradir[18]           = 4;

        (void) sprintf( cmemo, "Layer Reflct %2d - %2d %s",
                        prod_info.elevation, prod_info.top,
                        cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+13, 6 );
        (void) strcpy( ctitle, "LREF: Layer Composite Reflectivity" );
        (void) strcpy( cunit, "DBZ" );

        break;

      case 41:                   /* Echo Tops 2.2x2.2 NmixNmi 124  16 Raster */

        prod_info.elevation  = -1;
        prod_info.max        = product.p4;   /* Maximum echo top             */
        aradir[18]           = 8;

        (void) sprintf( cmemo, "Echo Tops [K FT] %s", cmode[product.mode] );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "TOPS: Echo Tops" );
        (void) strcpy( cunit, "K FT" );

        break;

      case 135:    /* Enhanced Echo Tops      .54x1   NmixDeg 186 199 Radial */

        prod_info.elevation  = -1;
        prod_info.max        = product.p4;   /* Maximum echo top             */
        aradir[18]           = 8;

        (void) sprintf( cmemo, "Echo Tops [K FT] %s", cmode[product.mode] );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "TOPS: Echo Tops" );
        (void) strcpy( cunit, "K FT" );

        break;

      case 78:     /* Surface Rainfall 1-hr   1.1x1   NmixDeg 124  16 Radial */

        prod_info.elevation  = -1;
        prod_info.scale      = 10;
        prod_info.max        = product.p4;   /* Maximum rainfall             */
        prod_info.end_time   = (product.p7 - 1) * 86400 + product.p8 * 60;
        aradir[18]           = 16;

        (void) sprintf( cmemo, "1-hr Rainfall [IN] %s", cmode[product.mode] );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "PRE1: Surface 1-hour Rainfall Total" );
        (void) strcpy( cunit, "IN" );

        break;

      case 79:     /* Surface Rainfall 3-hr   1.1x1   NmixDeg 124  16 Radial */

        prod_info.elevation  = -1;
        prod_info.scale      = 10;
        prod_info.max        = product.p4;   /* Maximum rainfall             */
        prod_info.end_time   = (product.p7 - 1) * 86400 + product.p8 * 60;
        aradir[18]           = 32;

        (void) sprintf( cmemo, "3-hr Rainfall [IN] %s", cmode[product.mode] );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "PRE3: Surface 3-hour Rainfall Total" );
        (void) strcpy( cunit, "IN" );

        break;

      case 80:     /* Surface Rainfall total  1.1x1   NmixDeg 124  16 Radial */

        prod_info.scale      = 10;
        prod_info.elevation  = -1;
        prod_info.max        = product.p4;   /* Maximum rainfall             */
        prod_info.start_time = (product.p5 - 1) * 86400 + product.p6 * 60;
        prod_info.end_time   = (product.p7 - 1) * 86400 + product.p8 * 60;
        aradir[18]           = 64;

        (void) sprintf( cmemo, "Strm Tot Rain [IN] %s", cmode[product.mode] );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "PRET: Surface Storm Total Rainfall" );
        (void) strcpy( cunit, "IN" );

        break;

      case 138:    /* Digital Storm Tot Prec  1.1x1   NmixDeg 124 256 Radial */

        prod_info.elevation  = -1;
        prod_info.max        = product.p4;   /* Maximum rainfall             */
        prod_info.start_time = (product.p1 - 1) * 86400 + product.p2 * 60;
        prod_info.end_time   = (product.p5 - 1) * 86400 + product.p6 * 60;
        aradir[18]           = 64;

        (void) sprintf( cmemo, "Dig Storm Tot Precip %s", cmode[product.mode] );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "DSP: Digital Storm Total Precipitation" );
        (void) strcpy( cunit, "IN" );

        break;

      case 57:     /* Vertical Liquid H2O     2.2x2.2 NmixNmi 124  16 Raster */

        prod_info.elevation  = -1;
        prod_info.max        = product.p4;   /* Maximum VIP                  */
        aradir[18]           = 128;

        (void) sprintf( cmemo, "Vert Int Lq H2O [mm] %s", cmode[product.mode] );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "VIL: Vertically-integrated Liquid Water" );
        (void) strcpy( cunit, "kg/m^2" );

        break;

      case 134:    /* HiRes Vert Liquid H2O   .54x1   NmixDeg 248 256 Radial */

        prod_info.elevation  = -1;
        prod_info.max        = product.p4;   /* Maximum digital VIL          */
        aradir[18]           = 128;

        (void) sprintf( cmemo, "Vert Int Lq H2O [mm] %s", cmode[product.mode] );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "VIL: Vertically-integrated Liquid Water" );
        (void) strcpy( cunit, "kg/m^2" );

        break;

      case 22:     /* Base Velocity           .13x1   NmixDeg  32   8 Radial */
      case 23:     /* Base Velocity           .27x1   NmixDeg  62   8 Radial */
      case 24:     /* Base Velocity           .54x1   NmixDeg 124   8 Radial */
      case 25:     /* Base Velocity           .13x1   NmixDeg  32  16 Radial */
      case 26:     /* Base Velocity           .27x1   NmixDeg  62  16 Radial */
      case 27:     /* Base Velocity           .54x1   NmixDeg 124  16 Radial */

        prod_info.min        = product.p4;   /* Maximum negative velocity    */
        prod_info.max        = product.p5;   /* Maximum positive velocity    */
        aradir[18]           = 256;

        (void) sprintf( cmemo, "Rad Vel %4.2f DEG %s",
                        prod_info.elevation/10., cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+8, 8);
        (void) strcpy( ctitle, "VEL: Radial Velocity" );
        (void) strcpy( cunit, "KT" );

        break;

      case 55:     /* StrmRel Mean Radial Vel .27x1   NmixDeg  27  16 Radial */
      case 56:     /* StrmRel Mean Radial Vel .54x1   NmixDeg 124  16 Radial */

        prod_info.min        = product.p4;   /* Maximum negative velocity    */
        prod_info.max        = product.p5;   /* Maximum positive velocity    */
        aradir[18]           = 512;

        (void) sprintf( cmemo, "StrmRelMnVl %4.2f DEG %s",
                        prod_info.elevation/10., cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+12, 8);
        (void) strcpy( ctitle, "SRMV: Storm Relative Mean Velocity" );
        (void) strcpy( cunit, "KT" );

        break;

      case 99:     /* Base Velocity DV        .13x1   NmixDeg 124 256 Radial */

        prod_info.min        = product.p4;   /* Maximum negative velocity    */
        prod_info.max        = product.p5;   /* Maximum positive velocity    */
        aradir[18]           = 256;

        (void) sprintf( cmemo, "Rad Vel DV  %4.2f DEG %s",
                        prod_info.elevation/10., cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+12, 8 );
        (void) strcpy( ctitle, "VEL: 99/DV Base Velocity" );
        (void) strcpy( cunit, "KT" );

        break;

      case 154:    /* Base Velocity .13nm X .5deg 256LVL 154/SDV 0.5DEG      */

        prod_info.min        = product.p4;   /* Maximum negative velocity    */
        prod_info.max        = product.p5;   /* Maximum positive velocity    */
        aradir[18]           = 256;

        (void) sprintf( cmemo, "Rad Vel SupRes  %4.2f DEG %s",
                        prod_info.elevation/10., cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+12, 8 );
        (void) strcpy( ctitle, "VEL: 154/SupRes Base Velocity" );
        (void) strcpy( cunit, "KT" );

        break;

      case 182:    /* TDWR Base Velocity      .08x1   NmixDeg  48 256 Radial */

        prod_info.min        = product.p4;
        prod_info.max        = product.p5;
        aradir[18]           = 256;

        (void) sprintf( cmemo, "Rad Vel %4.2f DEG %s",
                        prod_info.elevation/10., cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+8, 8);
        (void) strcpy( ctitle, "VEL: Radial Velocity" );
        (void) strcpy( cunit, "KT" );

        break;

      case 183:    /* TDWR Base Velocity      .08x1   NmixDeg  48  16 Radial */

        prod_info.min        = product.p4;
        prod_info.max        = product.p5;
        aradir[18]           = 256;

        (void) sprintf( cmemo, "Rad Vel %4.2f DEG %s",
                        prod_info.elevation/10., cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+8, 8);
        (void) strcpy( ctitle, "VEL: Radial Velocity" );
        (void) strcpy( cunit, "KT" );

        break;

      case 81:     /* Digital Precip Array  1/40thLFM NmixDeg 124 256 Radial */

        prod_info.elevation  = -1;
        prod_info.max        = product.p4;   /* Maximum rainfall accum       */
        prod_info.end_time   = (product.p7 - 1) * 86400 + product.p8 * 60;
        aradir[18]           = 1024;

        (void) sprintf( cmemo, "Digital Precip Array %s", cmode[product.mode] );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "DBA: Digital Precip Array" );
        (void) strcpy( cunit, "DBA" );

        break;

      case 155:   /* SuperRes BaseSpectrumWidth TSx .08x1 NmixDeg  162 256 Radial */

        prod_info.max        = product.p4;   /* Maximum spectrum width       */
        aradir[18]           = 262144;

        (void) sprintf( cmemo, "BaseSW TS0 %4.2f DEG %s",
                        prod_info.elevation/10., cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+12, 8 );
        (void) strcpy( ctitle, "BSW: 48 nm Base Spectrum Width" );
        (void) strcpy( cunit, "KT" );

        break;

      case 185:   /* TDWR Base SpectrumWidth TSx .08x1 NmixDeg  48   8 Radial */

        prod_info.max        = product.p4;   /* Maximum spectrum width       */
        prod_info.min        = product.p5;   /* Maximum spectrum width       */
#if 0
        prod_info.max        = product.p5;   /* Maximum spectrum width       */
#endif

        aradir[18]           = 262144;

        (void) sprintf( cmemo, "BaseSW TS0 %4.2f DEG %s",
                        prod_info.elevation/10., cmode[product.mode] );
        (void) strncpy( ctilt, cmemo+12, 8 );
        (void) strcpy( ctitle, "BSW: 48 nm Base Spectrum Width" );
        (void) strcpy( cunit, "KT" );

        break;

      /*
      ** Dual Polarization Products
      */

      case 159:   /* Differential Reflectivity NxX .13x1 NmixDeg 162 256 radial */

        prod_info.scale      = 100;
        prod_info.elevation  = product.p3;    /* Radar elevation * 10           */
        /* <<<<< UPC mod 20120514 - hardwire min/max values: -7.9 - 7.9   >>>>> */
#if 0
        sprintf( dbg, "GetNexrDirs:: prod_info.min max: %d/0x%02X %d/0x%02X",
                 product.p4, product.p4, product.p5, product.p5 );
        M0sxtrce( dbg );
        prod_info.min        = 10*product.p4; /* Min diff reflect * 10 (sign??) */
        prod_info.max        = 10*product.p5; /* Max diff reflect * 10 (sign??) */
#endif
        prod_info.min = -prod_info.scale * 7.9;
        prod_info.max =  prod_info.scale * 7.9;
        aradir[18]           = 2048;

        (void) sprintf( cmemo, "DiffReflect DZD [dB] %4.2f DEG",
                        prod_info.elevation/10. );
        (void) strncpy( ctilt, cmemo+21, 8 );
        (void) strcpy( ctitle, "DZD: 159/NxX Diff Reflectivity" );
        (void) strcpy( cunit, "dB" );

        break;

      case 161:   /* Digital Correlation Coef. NxC .13x1 NmixDeg 162 256 radial */

        prod_info.scale      = 300; 
        prod_info.elevation  = product.p3;
        /* <<<<< UPC mod 20120514 - hardwire min/max values: 0.2 - 1.05   >>>>> */
#if 0
        sprintf( dbg, "GetNexrDirs:: prod_info.min max: %d/0x%02X %d/0x%02X",
                 product.p4, product.p4, product.p5, product.p5 );
        M0sxtrce( dbg );
        prod_info.min        = 1 *product.p4;  /* Min corr coeff * 100 (sign??) */
        prod_info.max        = 1 *product.p5;  /* Max corr coeff * 100 (sign??) */
#endif
        prod_info.min = prod_info.scale * 0.20;
        prod_info.max = prod_info.scale * 1.05;
        aradir[18]           = 8192;

        (void) sprintf( cmemo, "CorCoefficient DCC %4.2f DEG",
                        prod_info.elevation/10. );
        (void) strncpy( ctilt, cmemo+19, 8 );
        (void) strcpy( ctitle, "DCC: 161/NxC Cor Coefficient" );
        (void) strcpy( cunit, "CC" );

        break;

      case 163:   /* Digital Specific Diff. Phase NxK .13x1 NmixDeg 162 256 radial */

        prod_info.elevation  = product.p3;
        prod_info.scale      = 20;
        /* <<<<< UPC mod 20120514 - hardwire min/max values: 0.2 - 1.05   >>>>> */
#if 0
        sprintf( dbg, "GetNexrDirs:: prod_info.min max: %d/0x%02X %d/0x%02X",
                 product.p4, product.p4, product.p5, product.p5 );
        M0sxtrce( dbg );
        prod_info.min = prod_info.scale * -2.0;
        prod_info.max = prod_info.scale * 10.6;
#endif
        prod_info.min        = product.p4;  /* Minimum specific diff phase      */
        prod_info.max        = product.p5;  /* Maximum specific diff phase      */
        aradir[18]           = 16384;

        (void) sprintf( cmemo, "SpDiffPhas DKD [Deg/km] %4.2f DEG",
                        prod_info.elevation/10. );
        (void) strncpy( ctilt, cmemo+24, 8 );
        (void) strcpy( ctitle, "DKD: 163/NxK: Specific Diff Phase" );
        (void) strcpy( cunit, "PHAS" );

        break;

      case 165:   /* Digital Hydrometeor Class. NxH .13x1 NmixDeg 162 256 radial */

        prod_info.scale      = 10;
        prod_info.elevation  = product.p3;
        aradir[18]           = 32768;

        (void) sprintf( cmemo, "HydroClass DHC %4.2f DEG",
                        prod_info.elevation/10. );
        (void) strncpy( ctilt, cmemo+15, 8 );
        (void) strcpy( ctitle, "DHC: 165/NxH Digital Hydro Class" );
        (void) strcpy( cunit, "CLAS" );

        break;

      case 169:   /* One Hour Accumulation OHA 1.1x1 NmixDeg 162 16 radial */

        prod_info.scale      = 10;
        prod_info.elevation  = -1;
        prod_info.max        = product.p4;   /* Maximum accumulation            */
        prod_info.end_time   = (product.p5 - 1) * 86400 + product.p6 * 60;
        aradir[18]           = 16;

        (void) sprintf( cmemo, "1-hr Accum OHA [IN] %s", cmode[product.mode] );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "OHA: 169/OHA 1-hr Accumulation" );
        (void) strcpy( cunit, "IN" );

        break;

      case 170:   /* Digital Accmulation Array DAA .13x1 NmixDeg 162 256 radial */

        prod_info.scale      = 100;
        prod_info.elevation  = -1;
        prod_info.max        = product.p4;   /* Maximum accumulation            */
        prod_info.end_time   = (product.p5 - 1) * 86400 + product.p6 * 60;
        aradir[18]           = 65536;

        (void) sprintf( cmemo, "Digital Accum Array DAA [IN]" );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "DAA: 170/DAA Digital Accum Array" );
        (void) strcpy( cunit, "IN" );

        break;

      case 171:   /* Storm Total Accumulation PTA 1.1x1 NmixDeg 162 16 radial */

        prod_info.scale      = 10;
        prod_info.elevation  = -1;
        prod_info.max        = product.p4;   /* Maximum accumulation            */
        prod_info.start_time = (product.p1 - 1) * 86400 + product.p2 * 60;
        prod_info.end_time   = (product.p5 - 1) * 86400 + product.p6 * 60;
        aradir[18]           = 64;

        (void) sprintf( cmemo, "StmTotAccum STA [IN] %s", cmode[product.mode] );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "STA: 171/STA Strm Tot Accum." );
        (void) strcpy( cunit, "IN" );

        break;

      case 172: /* Digital Storm Total Accum. DTA .13x1 NmixDeg 162 256 radial */

        prod_info.scale      = 100;
        prod_info.elevation  = -1;
        prod_info.max        = product.p4;   /* Maximum accumulation            */
        prod_info.start_time = (product.p1 - 1) * 86400 + product.p2 * 60;
        prod_info.end_time   = (product.p5 - 1) * 86400 + product.p6 * 60;
        aradir[18]           = 64;

        (void) sprintf( cmemo, "Digital StormTotalAccum DSA [IN]" );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "DSA: 172/DTA Dig Strm Tot Accum" );
        (void) strcpy( cunit, "IN" );

        break;

      case 173:    /* 3/24-hour Accumulation   .13x1   NmixDeg 124  256 Radial */

        prod_info.scale      = 100;
        prod_info.elevation  = -1;
        prod_info.end_time   = (product.p5 - 1) * 86400 + product.p6 * 60;
        prod_info.start_time = (product.p1 - 1) * 86400 + product.p2 * 60;
        prod_info.max        = product.p4;   /* Maximum accumulation            */
        aradir[18]           = 32;

        (void) sprintf( cmemo, "3-hr Accumulation DUA [IN]" );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "DUA: 173/DUA 3-hr Total Accum" );
        (void) strcpy( cunit, "IN" );

        break;

      case 174:     /* Digital 1-hour Dif. Accum.  .13x1   NmixDeg 124  256 Radial */

        prod_info.scale      = 100;
        prod_info.elevation  = -1;
        prod_info.min        = product.p7;   /* Minimum accumulation difference */
        prod_info.max        = product.p4;   /* Maximum accumulation difference */
        prod_info.end_time   = (product.p5 - 1) * 86400 + product.p6 * 60;
        aradir[18]           = 16;

        (void) sprintf( cmemo, "Dig 1-hr Diff Accum DOD [IN]" );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "DOD: 174/DOD Dig 1-hr Diff Accum" );
        (void) strcpy( cunit, "IN" );

        break;

      case 175:   /* Dig. Storm Total Dif. Accum.  .13x1   NmixDeg 124  256 Radial */

        prod_info.scale      = 100;
        prod_info.elevation  = -1;
        prod_info.max        = product.p4;   /* Maximum accumulation difference */
        prod_info.min        = product.p7;   /* Minimum accumulation difference */
        prod_info.start_time = (product.p1 - 1) * 86400 + product.p2 * 60;
        prod_info.end_time   = (product.p5 - 1) * 86400 + product.p6 * 60;
        aradir[18]           = 64;

        (void) sprintf( cmemo, "Dig StormTot Diff Accum DSD [IN]" );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "DSD: 175/DSD DigStrmTotDiff Accum" );
        (void) strcpy( cunit, "IN" );

        break;

      case 176:     /* Dig. Instant. Precip. Rate  .13x1   NmixDeg 124  256 Radial */
                    /* !!!!! Not yet implemented !!!!! */

        prod_info.elevation  = -1;
        prod_info.max        = product.p4;   /* Max inst precip rate            */
        prod_info.start_time = (product.p1 - 1) * 86400 + product.p2 * 60;
        aradir[18]           = 4096;

        (void) sprintf( cmemo, "Dig Inst Precip Rate DPR [IN/HR]" );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "DPR: 176/DPR Dig Inst Precip Rate" );
        (void) strcpy( cunit, "IN/HR" );

        break;

      case 177:     /* Hybrid Hydrometeor Class.  .13x1   NmixDeg 124  256 Radial */

        prod_info.scale      = 10;
        prod_info.elevation  = -1;
        prod_info.top        = product.p6;   /* Highest elevation used          */
        aradir[18]           = 32768;

        (void) sprintf( cmemo, "Hybrid Hydrometeor Class HHC" );
        (void) strcpy( ctilt, "-99" );
        (void) strcpy( ctitle, "HHC: 177/HHC Hyb Scan Hydro Class" );
        (void) strcpy( cunit, "CLAS" );

        break;

      default:

        (void) sprintf( cmemo, "123456789012345678901234567890123" );
        (void) strcpy( ctitle, "123456789012345678901234567890123" );
        (void) sprintf( dbg, "Leave GetNexrDirs:: ERROR unknown product code %d",
                        product.code );
        M0sxtrce( dbg );
        return FAILURE;

    }

    (void) memset( &aradir[24], ' ', 32 );
    (void) memcpy( &aradir[24], cmemo, strlen( cmemo ) );

#if DEBUG_DIRS
    (void) sprintf( dbg, "GetNexrDirs:: raster res: %f  data res: %f",
                    prod_info.raster_res, prod_info.data_res );
    M0sxtrce( dbg );
#endif

    if ( prod_info.raster_res < 0.0001 )
      prod_info.raster_res = prod_info.data_res;

    /*
    ** Get product symbology header (needed to get image shape)
    */

#if DEBUG_DIRS
    (void) sprintf( dbg,
              "GetNexrDirs:: off_symbol: %d, off_graphic: %d, off_tabular: %d",
               product.off_symbol, product.off_graphic, product.off_tabular );
    M0sxtrce( dbg );
#endif

    if ( product.off_symbol != 0 ) {

      if ( !(rc=read_dividlen( bp, &ps_head )) || (ps_head.divider != -1) ) {
        (void) sprintf( dbg, "GetNexrDirs:: ERROR in product symbology header, rc=%d, ps_head.divider = %0hX",
                        rc, ps_head.divider );
        M0sxtrce( dbg );
        return FAILURE;
      }
#if DEBUG_DIRS
      (void) sprintf( dbg, "GetNexrDirs:: rc: %d ps_head.divider: %0hX",
                      rc, ps_head.divider);
      M0sxtrce( dbg );
#endif

      hedsiz += 10;

#if DEBUG_DIRS
      (void) sprintf( dbg, "GetNexrDirs:: number of ps layers: %d",
                      ps_head.layers );
      M0sxtrce( dbg );
#endif

      if ( ! (getbuf( bp, (char *)&divlen.divider, 1, 2 )  &&
              getbuf( bp, (char *)&divlen.length,  1, 4 )) ||
              divlen.divider != -1 ) {
        M0sxtrce( "Leave GetNexrDirs:: ERROR reading length divider" );
        return FAILURE;
      }
      hedsiz += 6;

      if ( !getbuf( bp, (char *) &rcode, 1, 2 ) ) {
        M0sxtrce( "Leave GetNexrDirs:: ERROR reading radial/raster indicator" );
        return FAILURE;
      }
      hedsiz += 2;

      scale = prod_info.data_res / prod_info.raster_res;

#if DEBUG_DIRS
      (void) sprintf( dbg, "GetNexrDirs:: rcode: 0x%x  scale: %f",
                      (int)rcode, scale );
      M0sxtrce( dbg );
#endif

      if ( (rcode == 0xAF1F) || (rcode == 0x10) ) { /* 16, 256 lvl radial img */

        M0sxtrce( "GetNexrDirs:: radial image found" );

        radp_head.code = rcode;

        if ( ! (getbuf( bp, (char *) &radp_head.first_bin,     1, 2 ) &&
                getbuf( bp, (char *) &radp_head.num_bin,       1, 2 ) &&
                getbuf( bp, (char *) &radp_head.i,             1, 2 ) &&
                getbuf( bp, (char *) &radp_head.j,             1, 2 ) &&
                getbuf( bp, (char *) &radp_head.scale,         1, 2 ) &&
                getbuf( bp, (char *) &radp_head.num_radials,   1, 2 ) ) ) {

          M0sxtrce( "Leave GetNexrDirs:: ERROR reading radial packet header" );
          return FAILURE;

        }

#if DEBUG_DIRS
        (void) sprintf( dbg,
               "GetNexrDirs:: 1st %d, nbins %d, i %d, j %d, scale %d, nrad %d",
                radp_head.first_bin, radp_head.num_bin, radp_head.i,
                radp_head.j, radp_head.scale, radp_head.num_radials );
        M0sxtrce( dbg );
#endif

        prod_info.size = 2 * (int) (radp_head.num_bin * scale + 0.5);
        hedsiz += 12;

      } else if ( (rcode == 0xBA0F) || (rcode == 0xBA07) ) { /* raster image */

        M0sxtrce( "GetNexrDirs:: raster image found" );

        rasp_head.code[0] = rcode;

        if ( ! (getbuf( bp, (char *) &rasp_head.code[1],     2, 2 ) &&
                getbuf( bp, (char *) &rasp_head.i,           1, 2 ) &&
                getbuf( bp, (char *) &rasp_head.j,           1, 2 ) &&
                getbuf( bp, (char *) &rasp_head.xscale,      1, 2 ) &&
                getbuf( bp, (char *) &rasp_head.xscalefract, 1, 2 ) &&
                getbuf( bp, (char *) &rasp_head.yscale,      1, 2 ) &&
                getbuf( bp, (char *) &rasp_head.yscalefract, 1, 2 ) &&
                getbuf( bp, (char *) &rasp_head.num_rows,    1, 2 ) &&
                getbuf( bp, (char *) &rasp_head.packing,     1, 2 )) ) {

          M0sxtrce( "GetNexrDirs:: ERROR reading raster packet header" );
          return FAILURE;

        }

        prod_info.size = rasp_head.num_rows * scale;
        hedsiz += 20;

      } else if ( (rcode == 0x11) || (rcode == 0x12) ) { /* raster on LFM grid */

        M0sxtrce( "GetNexrDirs:: LFM raster image found" );

        lfm_head.code = rcode;

        if ( ! (getbuf( bp, (char *) &lfm_head.spare,         2, 2 ) &&
                getbuf( bp, (char *) &lfm_head.num_lfm,       1, 2 ) &&
                getbuf( bp, (char *) &lfm_head.num_rows,      1, 2 ) ) ) {

          M0sxtrce( "GetNexrDirs:: ERROR reading LFM radial packet header" );
          return FAILURE;

        }

        prod_info.size = lfm_head.num_rows * scale;
        hedsiz += 8;

      } else if ( rcode == 0x1c ) { /* Generic Radial Product Format */

        M0sxtrce( "GetNexrDirs:: unsupported Generic Radial Product Format image found" );

        return FAILURE;

      } else {

        (void) sprintf( dbg, "GetNexrDirs:: unsupported product, rcode = 0x%X",
                        rcode );
        M0sxtrce( dbg );
        return FAILURE;

      }

    } else {

      M0sxtrce( "GetNexrDirs:: no product symbology block (no image data)" );
      return FAILURE;

    }

    /*
    ** Convert volume scan date (days since 700101) and time (seconds since
    ** midnight) to nominal and actual image date [CCYYDDD] and time [HHMMSS]
    ** format
    */

    cur->time = 86400 * (product.vs_date - 1) + product.vs_time;

    settime = gmtime( &cur->time );

    aradir[3] =  1000 * (1900 + settime->tm_year) + settime->tm_yday + 1;
    aradir[4] = 10000*settime->tm_hour + 100*settime->tm_min + settime->tm_sec;

    aradir[45] = aradir[3];
    aradir[46] = aradir[4];
#if 0
    /* <<<<< UPC mod 20080906 - do not truncate seconds from the time >>>>> */
    aradir[4] = 100 * ( aradir[4] / 100 );
#endif

#if DEBUG_DIRS
    (void) sprintf( dbg,
                  "GetNexrDirs:: product date: %d product time: %06d tsec: %d",
                    aradir[3], aradir[4], cur->time );
    M0sxtrce( dbg );
#endif

    /*
    ** Number of image lines and elements (Level III products are square)
    */

    aradir[8] = prod_info.size;
    aradir[9] = 4 * ((prod_info.size+3) / 4 );    /* # elements must be even */

    /*
    ** Use the product creation time as the image creation date [YYDDD]
    ** and time [HHMMSS]
    */

    datetime = 86400 * (product.prod_date - 1) + product.prod_time;

    settime = gmtime( &datetime );

    aradir[16] =  1000*settime->tm_year + settime->tm_yday + 1;
    aradir[17] = 10000*settime->tm_hour + 100*settime->tm_min + settime->tm_sec;

#if DEBUG_DIRS
    (void) sprintf( dbg,
                    "GetNexrDirs:: generation data: %d generation time: %06d",
                    aradir[16], aradir[17] );
    M0sxtrce( dbg );
#endif

    /*
    ** Product code, radar operation mode, number of data levels
    */

    aradir[21] = product.code;
    aradir[22] = product.mode;
    aradir[23] = prod_info.levels;

    /*
    ** AUX block size and byte offset to data block
    */

    auxsiz     = 43 + prod_info.levels;     /* word length of AUX block      */
    aradir[60] = 4*auxsiz;                  /* AUX block length in bytes     */
    aradir[33] = aradir[59] + aradir[60];   /* byte offset to data block     */

#if DEBUG_DIRS
    (void) sprintf( dbg, "GetNexrDirs:: id: %s  code: %d  mode: %d",
                    cur->id, aradir[21], aradir[22] );
    M0sxtrce( dbg );
    (void) sprintf( dbg, "GetNexrDirs:: date: %d  time: %06d  nlevel: %d band: %d",
                    aradir[3], aradir[4], aradir[23], aradir[18] );
    M0sxtrce( dbg );
#endif

    /*
    ** Save the start location of the data
    */

    cur->doff = hedsiz;

    /*************************************************************************/
    /*                      Navigation Codicil                               */
    /*************************************************************************/

    dlat = (double) product.latitude / 1000.0;
    dlon = (double) product.longitude / 1000.0;

    navcod[0] =  lit_( "RADR", 4 );         /* RADR navigation module        */
    navcod[1] =  aradir[8] / 2;             /* row containing radar site     */
    navcod[2] =  aradir[9] / 2;             /* col containing radar site     */
    navcod[3] =  mfd_(&dlat);               /* latitude         [DDDMMSS]    */
    navcod[4] = -mfd_(&dlon);               /* longitude        [DDDMMSS]    */
    navcod[5] =  prod_info.raster_res*1000; /* pixel resolution [m]          */
    navcod[6] =  0;                         /* rotation of north from vert.  */
    navcod[7] =  0;                         /* pixel resolution [m] in long
                                             *   direction                   */

#if DEBUG_DIRS
    (void) sprintf( dbg,
                    "GetNexrDirs:: row: %d  col: %d  lat: %d  lon: %d  res: %d",
                    navcod[1], navcod[2], navcod[3], navcod[4], navcod[5] );
    M0sxtrce( dbg );
#endif

    /*************************************************************************/
    /*                         Auxiliary Block                               */
    /*************************************************************************/

    auxblk[ 0] = 0x04030201;                /* magic number                  */
    auxblk[ 1] = aradir[60];                /* AUX block size in bytes       */
    auxblk[ 2] = 4;                         /* length of entry name in bytes */
    auxblk[ 3] = 4*(auxsiz-5);              /* length of entry bytes         */
    auxblk[ 4] = lit_("INFO",4);            /* AUX block name fixed as 'INFO'*/
    (void) memset( &auxblk[5], ' ', 40 );   /* memo field                    */
    (void) memcpy( &auxblk[5], ctitle, strlen(ctitle) );
    auxblk[15] = aradir[21];                /* NEXRAD product code (BREF=19) */
    auxblk[16] = 8;                         /* # bytes in radar textual ID   */
    (void) memset( &auxblk[17], ' ', 8 );   /* radar ID                      */
    (void) memcpy( &auxblk[17], &aradir[20], 4 );
    auxblk[19] = navcod[3];                 /* radar latitude                */
    auxblk[20] = navcod[4];                 /* radar longitude               */
    auxblk[21] = product.height;            /* radar elevation               */
    auxblk[22] = aradir[3];                 /* volume scan DAY [YYYDDD]      */
    auxblk[23] = aradir[4];                 /* volume scan TIME [HHMMSS]     */
    auxblk[24] = product.mode;              /* radar mode                    */
    auxblk[26] = prod_info.data_res;        /* resolution                    */
    auxblk[25] = prod_info.size*prod_info.data_res/2; /* range <<<<< UPC mod 20101116 >>>>> */
    auxblk[27] = update_rate[aradir[22]];   /* update rate [10|6|5]          */
    (void) memset( &auxblk[28], ' ', 8 );   /* tilt angle or layer           */
    (void) memcpy( &auxblk[28], ctilt, strlen(ctilt) );

    auxblk[33] = auxblk[22];                /* start DAY [YYYDDD]            */
    auxblk[34] = auxblk[23];                /* start TIME [HHMMSS]           */
    auxblk[35] = auxblk[22];                /* end DAY [YYYDDD]              */
    auxblk[36] = auxblk[23];                /* end TIME [HHMMSS]             */
    auxblk[37] = 8;                         /* # bytes in unit               */
    (void) memset( &auxblk[38], ' ', 8 );   /* data unit                     */
    (void) memcpy( &auxblk[38], cunit, strlen(cunit) );
    auxblk[40] = prod_info.levels;          /* # data levels                 */
    auxblk[41] = 0;                         /* data offset                   */

    /*
    ** Calibrated data values
    */

#if DEBUG_DIRS
    (void) sprintf( dbg, "GetNexrDirs:: #levels: %d", prod_info.levels );
    M0sxtrce( dbg );
#endif

    if ( prod_info.levels <= 16 ) {

#if DEBUG_DIRS
      (void) sprintf( dbg, "GetNexrDirs:: threshold: 0x%X", product.threshold[0] );
      M0sxtrce( dbg );
      (void) sprintf( dbg, "GetNexrDirs:: product min, max: %f %f",
                           prod_info.min, prod_info.max );
      M0sxtrce( dbg );
#endif

      /* <<<<< UPC mod 20100827 - add *100 scaling; change min/max/scale reporting >>>>> */
      auxblk[42] = 1;                       /* data scale             */
      ival = product.threshold[0];
      if ( (ival & 0x00004000) != 0 ) auxblk[42] = 100;
      if ( (ival & 0x00002000) != 0 ) auxblk[42] = 20;
      if ( (ival & 0x00001000) != 0 ) auxblk[42] = 10;
      auxblk[30] = auxblk[42];              /* min/max scale factor   */
      /* <<<<< UPC mod 20101116 - return unscaled min/max >>>>> */
      auxblk[31] = prod_info.min;           /* min data value */
      auxblk[32] = prod_info.max;           /* max data value */
#if 0
      /* Code removed 20101116 scaled data; appears to be incorrect reading of
         ICD document's meaning of how data is scaled in the file
      */
      auxblk[31] = prod_info.min*auxblk[30]/prod_info.scale; /* min data value */
      auxblk[32] = prod_info.max*auxblk[30]/prod_info.scale; /* max data value */
#endif

      for ( i = 0; i < auxblk[40]; i++ ) {
        ival = product.threshold[i];
#if DEBUG_DIRS
        (void) sprintf( dbg, "GetNexrDirs:: product.threshold[%0d] %d, 0x%X",
                        i, ival, ival );
        M0sxtrce( dbg );
#endif
        if ( (ival & 0x00008000) == 0 ) {
          isign = -1;
          if ( (ival & 0x00000100) == 0 ) isign = 1;
          auxblk[43+i] = isign * ( ival & 0x000000FF );
        } else {
          auxblk[43+i] = -9999 + ( ival & 0x000000FF );
        }
      }

      /*
      ** Adjust beginning and ending day/time for Storm Total Precip product.
      */

      if ( product.code == 80 ) {
        rc = Mcsectodaytime( prod_info.start_time, &auxblk[33], &auxblk[34] );
        rc = Mcsectodaytime( prod_info.end_time,   &auxblk[35], &auxblk[36] );
      }

    } else {

      switch ( product.code ) {
       
      case 32:        /* Digital Hybrid Scan Reflect                  */
      case 94:        /* Base Reflectivity Data Array DR              */
      case 153:       /* Base Reflectivity 256LVL 153/SDR             */ 
      case 180:       /* Base Reflectivity from TDWR - TRx            */ 
      case 181:       /* Base Reflectivity from TDWR - TRx            */ 
      case 186:       /* Base Reflectivity from TDWR - TZL            */
      case 187:       /* Base Reflectivity from TDWR - TZx            */
                      /* ICD Fig 3.6 Note 1                           */
                      /* halfword   variable          meaning         */
                      /*    31      product.threshold[0] min dBZ*10   */
                      /*    32      product.threshold[1] inc dBZ*10   */
                      /*    33      product.threshold[2] #levels      */

        auxblk[42] = 10;
        auxblk[30] = auxblk[42];            /* min/max scale factor   */
        auxblk[31] = prod_info.min*10;      /* min data value         */
        auxblk[32] = prod_info.max*10;      /* max data value         */

        auxblk[43] = -9998;                 /* THreshold              */
        auxblk[44] = -9999;                 /* Missing (ND)           */
        for ( i = 2; i < prod_info.levels; i++ ) {
          auxblk[43+i] = product.threshold[0] + (i-2)*product.threshold[1];
        }

        break;

      case 81:        /* Digital Precipitation Array                  */
                      /* ICD Fig 3.6 Note 1                           */
                      /* halfword   variable          meaning         */
                      /*    31      product.threshold[0] min dBA*10   */
                      /*    32      product.threshold[1] inc dBA*1000 */
                      /*    33      product.threshold[2] #levels      */

        auxblk[42] = 1000;
        auxblk[30] = auxblk[42];            /* min/max scale factor   */
        auxblk[32] = prod_info.max*100;     /* max data value         */

        auxblk[43] = -9998;                 /* No accumulation (TH)   */
        for ( i = 1; i < prod_info.levels-1; i++ ) {
          auxblk[43+i] = 100*product.threshold[0] + i*product.threshold[1];
        }
        auxblk[43+prod_info.levels-1] = -9999; /* Outside of coverage */

        break;

      case 93:        /* ITWS Digital Base Velocity                   */
      case 99:        /* Base Velocity Data Array DV                  */
      case 154:       /* Super Resolution Digital Base Velocity       */
      case 182:       /* Base Velocity from TDWR - TVx 256 level      */
      case 183:       /* Base Velocity from TDWR - TVx  16 level      */
                      /* ICD Fig 3.6 Note 1                           */
                      /* halfword   variable          meaning         */
                      /*    31      product.threshold[0] min m/s*10   */
                      /*    32      product.threshold[1] inc m/s*10   */
                      /*    33      product.threshold[2] #levels      */

        /* <<<<< UPC mod 20100709 - scale min/max values by 10  >>>>> */
        auxblk[42] = 10;
        auxblk[30] = auxblk[42];            /* min/max scale factor   */
        auxblk[31] = prod_info.min*10;      /* min data value         */
        auxblk[32] = prod_info.max*10;      /* max data value         */

        auxblk[43] = -9998;                 /* THreshold              */
        auxblk[44] = -9996;                 /* Range Folded           */
        for ( i = 2; i < prod_info.levels; i++ ) {
          auxblk[43+i] = product.threshold[0] + (i-2)*product.threshold[1];
          auxblk[43+i] *= 1.94384449;       /* convert m/s to knots   */
        }
        
        break;

      case 134: /* High Resolution Vert Int Liquid H2O                */
                /* ICD Fig 3.6 Note 1 Build 9+                        */
                /* halfword variable             meaning              */
                /*    31    product.threshold[0] linear scale  0x59AB */
                /*    32    product.threshold[1] linear offset 0x4400 */
                /*    33    product.threshold[2] log start val 20     */
                /*    34    product.threshold[3] log scale     0x54DC */
                /*    35    product.threshold[4] log offset    0x593E */

        /* Linear scale, offset */
        lscale = GetFloat16( product.threshold[0] );
        lofset = GetFloat16( product.threshold[1] );

        /* Log scale, offset */
        gscale = GetFloat16( product.threshold[3] );
        gofset = GetFloat16( product.threshold[4] );

        /* Calibrated values */
        aradir[58] = 10;
        auxblk[42] = 1000;
        auxblk[30] = 1;                     /* min/max scale factor   */
        auxblk[31] = 0;                     /* min data value         */
        /* <<<<< UPC mod 20101115 - change calculation for max value. This
           assumes that the max value needs to be scaled like the other
           data values in the DVL product, but the need to do this is not
           documented anywhere in the Build 11.0 documentation!
        auxblk[32] = prod_info.max;
        */
        if ( prod_info.max < product.threshold[2] ) { /* max data value */
          auxblk[32] = (prod_info.max - lofset)/lscale;
        } else {
          auxblk[32] = exp((prod_info.max - gofset)/gscale);
        }

        auxblk[43] = -9998;                 /* Below Threshold        */
        auxblk[44] = -9995;                 /* Flagged Data           */
        for ( i = 2; i < 255; i++ ) {
          if ( i < product.threshold[2] ) {
            auxblk[43+i] =  auxblk[42] * (i - lofset)/lscale;
          } else {
            auxblk[43+i] =  auxblk[42] * exp((i - gofset)/gscale);
          }
        }
        auxblk[43+255] = -9999;                   /* Reserved         */

        break;

      case 135:       /* High Resolution Enhanced Echo Tops           */

        auxblk[42] = 1;
        auxblk[30] = auxblk[42]*10;         /* min/max scale factor   */
        auxblk[31] = prod_info.min*10;      /* min data value         */
        auxblk[32] = prod_info.max*10;      /* max data value         */

        auxblk[43] = -9998;                 /* Below Threshold        */
        auxblk[44] = -9995;                 /* Bad data               */

        for ( i = 2; i < prod_info.levels; i++ ) {
          auxblk[43+i] = i - 2;
        }

        break;

      case 138:       /* Digital Storm Total Precip                   */

        auxblk[42] = 100;                   /* data scale             */
        auxblk[30] = auxblk[42];            /* min/max scale factor   */
        auxblk[31] = product.threshold[0];  /* min data value         */
        auxblk[32] = prod_info.max;         /* max data value         */

        auxblk[43] = product.threshold[0];
        for ( i = 1; i < prod_info.levels; i++ ) {
          /* <<<<< UPC mod 20100820 - include increment in calculation >>>>> */
          auxblk[43+i] = auxblk[43] + i * product.threshold[1];
        }

        break;

      case 136:       /* SuperOb                                      */
      case 149:       /* Digital Mesocyclone Detection                */
      case 152:       /* Archive III Status Product                   */

        break;

      case 155:       /* Super Resolution Digital Base Spectrum Width */
                      /* ICD Fig 3.6 Note 1                           */
                      /* halfword   variable          meaning         */
                      /*    31      product.threshold[0] min m/s*10   */
                      /*    32      product.threshold[1] inc m/s*10   */
                      /*    33      product.threshold[2] #levels      */

        auxblk[43] = -9998;                 /* below THreshold        */
        auxblk[44] = -9996;                 /* Range Folded           */
        for ( i = 2; i < product.threshold[2]; i++ ) {
          if ( i < 129 || i > 149 ) {
            auxblk[43+i] = 0;
          } else {
            auxblk[43+i] = product.threshold[0] + (i-2)*product.threshold[1];
          }
        }

        break;

      /*
       * The following was extracted from:
       *
       * INTERFACE CONTROL DOCUMENT FOR THE RPG TO CLASS 1 USER
       * Document Number 2620001R
       * Code Identification 0WY55
       * WSR-88D ROC
       * 08 October 2010
       * Open Build 12.1
       *
       * For products 159, 161, 163, 170, 172, 173, 174, 175 and 176 data levels
       * that are not used as leading or trailing flag values relate to the data
       * in physical units via a linear relationship through the equation:
       *
       * F = (N - OFFSET) / SCALE
       *
       * F      - floating point data value
       * N      - integer data value
       * SCALE  - halfwords 31, 32
       *          ANSI/IEEE Standard 754-1985 floating point value
       * OFFSET - halfwords 33, 34
       *          ANSI/IEEE Standard 754-1985 floating point value
       */

      case 159:         /* DZD - Digital Differential Reflectivity        */

        auxblk[30] = prod_info.scale;           /* min/max scale          */
        auxblk[31] = prod_info.min;             /* min data value         */
        auxblk[32] = prod_info.max;             /* max data value         */
        auxblk[42] = prod_info.scale;           /* data scale             */

        lscale = GetFloat32( product.threshold[0], product.threshold[1] );
        lofset = GetFloat32( product.threshold[2], product.threshold[3] );
        
        /*
        ** Sanity check on scale factor (too close to zero will make routine
        ** blow-up
        */
        if ( fabs((double)lscale) <  0.0001 ) {
          lscale = prod_info.scale;
          lofset = 128.0;
          (void) sprintf( dbg,
              "GetNexrDirs:: ZERO product %d scale, using default of %f",
              product.code, lscale );
          M0sxtrce( dbg );
        }

        switch ( product.threshold[6] ) {
          case 2:
            auxblk[43] = -9998;                 /* below THreshold        */
            auxblk[44] = -9996;                 /* Range Folded           */
            break;
          case 1:
            auxblk[43] = -9999;                 /* No Data                */
            break;
        }

        for ( i = product.threshold[6]; i < product.threshold[5]+1; i++ ) {
          auxblk[43+i] =  auxblk[42] * (i - lofset)/lscale;
        }

        for ( i = product.threshold[5]+1; i < 256; i++ ) {
          auxblk[43+i] = -9999;                 /* No Data                */
        }

        break;

      case 161:         /* DCC - Digital Correlation Coefficient          */

        auxblk[30] = prod_info.scale;           /* min/max scale          */
        auxblk[31] = prod_info.min;             /* min data value         */
        auxblk[32] = prod_info.max;             /* max data value         */
        auxblk[42] = prod_info.scale;           /* data scale             */

        lscale = GetFloat32( product.threshold[0], product.threshold[1] );
        lofset = GetFloat32( product.threshold[2], product.threshold[3] );
        
        /*
        ** Sanity check on scale factor (too close to zero will make routine
        ** blow-up
        */
        if ( fabs((double)lscale) <  0.0001 ) {
          lscale = prod_info.scale;
          lofset = -60.5;
          (void) sprintf( dbg,
              "GetNexrDirs:: ZERO product %d scale, using default of %f",
              product.code, lscale );
          M0sxtrce( dbg );
        }

        switch ( product.threshold[6] ) {
          case 2:
            auxblk[43] = -9998;                 /* below THreshold        */
            auxblk[44] = -9996;                 /* Range Folded           */
            break;
          case 1:
            auxblk[43] = -9999;                 /* No Data                */
            break;
        }

        for ( i = product.threshold[6]; i < product.threshold[5]+1; i++ ) {
          auxblk[43+i] =  auxblk[42] * (i - lofset)/lscale;
        }

        for ( i = product.threshold[5]+1; i < 256; i++ ) {
          auxblk[43+i] = -9999;                 /* No Data                */
        }

        break;

      case 163:         /* DCC - Digital Correlation Coefficient          */

        auxblk[30] = prod_info.scale;           /* min/max scale          */
        auxblk[31] = prod_info.min;             /* min data value         */
        auxblk[32] = prod_info.max;             /* max data value         */
        auxblk[42] = 10;                        /* data scale             */

        lscale = GetFloat32( product.threshold[0], product.threshold[1] );
        lofset = GetFloat32( product.threshold[2], product.threshold[3] );
        
        /*
        ** Sanity check on scale factor (too close to zero will make routine
        ** blow-up
        */
        if ( fabs((double)lscale) <  0.0001 ) {
          lscale = prod_info.scale;
          lofset = 43.0;
          (void) sprintf( dbg,
              "GetNexrDirs:: ZERO product %d scale, using default of %f",
              product.code, lscale );
          M0sxtrce( dbg );
        }

        switch ( product.threshold[6] ) {
          case 2:
            auxblk[43] = -9998;                 /* below THreshold        */
            auxblk[44] = -9996;                 /* Range Folded           */
            break;
          case 1:
            auxblk[43] = -9999;                 /* No Data                */
            break;
        }

        for ( i = product.threshold[6]; i < product.threshold[5]+1; i++ ) {
          auxblk[43+i] =  auxblk[42] * (i - lofset)/lscale;
        }

        for ( i = product.threshold[5]+1; i < 256; i++ ) {
          auxblk[43+i] = -9999;                 /* No Data                */
        }

        break;

      case 170:   /* 170/DAA: Digital Accumumulation Array                */
      case 172:   /* 172/DTA: Digital Storm Total Accumulation            */
      case 173:   /* 173/DUA: User Selectable (3/24 hr) Accumulation      */
      case 174:   /* 174/DOD: Digital One-Hour Difference Accumulation    */
      case 175:   /* 175/DSD: Digital Storm Total Difference Accumulation */

        auxblk[30] = 10;                        /* min/max scale          */
        auxblk[32] = prod_info.max;             /* max data value         */
        auxblk[42] = prod_info.scale;           /* data scale             */

        lscale = GetFloat32( product.threshold[0], product.threshold[1] );
        lofset = GetFloat32( product.threshold[2], product.threshold[3] );
        
#if DEBUG_DIRS
        (void) sprintf( dbg, "GetNexrDirs:: lscale, lofset: %f, %f",
                          lscale, lofset );
        M0sxtrce( dbg );
#endif

        /*
        ** Sanity check on scale factor (too close to zero will make routine
        ** blow-up
        */
        if ( fabs((double)lscale) <  0.0001 ) {

          if (product.code == 170 || product.code == 172 || product.code == 173 ) {
            lscale = 2.0;
          } else {
            lscale = 1.4481;
          }
          (void) sprintf( dbg,
              "GetNexrDirs:: ZERO product %d scale, using default of %f",
              product.code, lscale );
          M0sxtrce( dbg );

        }

        switch ( product.threshold[6] ) {
          case 2:
            auxblk[43] = -9998;                 /* below THreshold        */
            auxblk[44] = -9996;                 /* Range Folded           */
            break;
          case 1:
            auxblk[43] = -9999;                 /* No Data                */
            break;
        }

        for ( i = product.threshold[6]; i < product.threshold[5]+1; i++ ) {
          auxblk[43+i] = (i - lofset) / lscale;
        }

        for ( i = product.threshold[5]+1; i < 256; i++ ) {
          auxblk[43+i] = -9999;                 /* No Data                */
        }

        /*
        ** Min, max data values
        */
        auxblk[31] = prod_info.min;
        auxblk[32] = prod_info.max;

        (void) sprintf( dbg, "GetNexrDirs:: product min, max: = %6.2f, %6.2f",
                        prod_info.min/prod_info.scale,
                        prod_info.max/prod_info.scale );
        M0sxtrce( dbg );

        break;

      case 176:         /* DPR - Digital Instantaneous Precipitation Rate */
                        /* !!!!! Not yet implemented !!!!! */

        auxblk[42] = 10;                        /* data scale             */

        lscale = GetFloat32( product.threshold[0], product.threshold[1] );
        lofset = GetFloat32( product.threshold[2], product.threshold[3] );
        
        switch ( product.threshold[6] ) {
          case 2:
            auxblk[43] = -9998;                 /* below THreshold        */
            auxblk[44] = -9996;                 /* Range Folded           */
            break;
          case 1:
            auxblk[43] = -9999;                 /* No Data                */
            break;
        }

        for ( i = product.threshold[6]; i < product.threshold[5]+1; i++ ) {
          auxblk[43+i] =  auxblk[42] * (i - lofset)/lscale;
        }

        for ( i = product.threshold[5]+1; i < 256; i++ ) {
          auxblk[43+i] = -9999;                 /* No Data                */
        }

        break;

      case 165:           /* Digital Hydrometeor Classification (DHC)     */
      case 177:           /* Hybrid Scan Hydrometeor Classification (HHC) */

        auxblk[30] = 1;                         /* min/max scale          */
        auxblk[42] = prod_info.scale;           /* data scale             */

        for ( i = 0; i < prod_info.levels-9; i++ ) {
          auxblk[43+i] =  100 * (i/10);
        }
        for ( i = prod_info.levels-9; i <= prod_info.levels; i++ ) {
          auxblk[43+i] =  0;
        }

        break;

      case 167:           /* Future Dual Pol Base product                 */
      case 168:           /* Future Dual Pol Base product                 */

        break;

      case 178:           /* Future QPE product                           */
      case 179:           /* Future QPE product                           */

        break;

      }

    }

#if DEBUG_DIRS
    for ( i = 0; i < auxsiz; i++ ) {
      (void) sprintf( dbg, "GetNexrDirs:: auxblk[%3d] = 0x%08X %2d",
                      i, auxblk[i], auxblk[i] );
      M0sxtrce( dbg );
    }
#endif

    /*
    ** Return
    */

#if DEBUG_DIRS || DEBUG_CALL
    M0sxtrce( "Leave GetNexrDirs" );
#endif

    return SUCCESS;

}

/******************************* GetNexrLine *********************************/

int
GetNexrLine( FILELIST *cur, READPARM *read, int band, char *buf, char *err )

/*
** Name:    GetNexrLine
**
** Purpose: Unpack encoded reflectivity data into equivalent McIDAS AREA line
**
** Note:    This routine reads image data from a NOAAPORT or WSI Level III
**          NEXRAD formatted file.
**
** Parameters:
**          cur  - FILELIST structure containing current image information
**                 name  - full qualified name of the file
**                 pos   - position to add to list
**                 hoff  - byte offset to beginning of header information
**                 doff  - byte offset to beginning of image data
**                 size  - file size in bytes
**                 type  - file type
**                 id    - NEXRAD Level III station ID
**                 time  - file data time [sec since 1970]
**          read - READPARM struct containing read specs
**          band - band number of elements to read
**          buf  - buffer containing image data
**          err  - error string to return
**
** Returns:
**          SUCCESS == 1
**          FAILURE == 0
**
** History: 19980514: Adapted from a Unidata WSI NOWrad GetNOWradLine module
**
*/

{

    static char lastname[MAX_NAME_LEN];     /* filename last time in func.   */

    static unsigned char *imagedata = NULL; /* all packed image data in file */
    static unsigned char *dataptr;          /* pointer to data in imagedata  */

    BUFREAD              *bp=NULL;          /* image data structure pointer  */

    int                   i;                /* loop variable                 */

    Fint4                 size;             /* #rows/cols in square image    */

#if DEBUG_LINE || DEBUG_CALL
    M0sxtrce( "In GetNexrLine" );
#endif

    /*
    ** Save the product #rows/cols (image is square) in local variable
    */

    size = prod_info.size;

#if DEBUG_LINE
    (void) sprintf( dbg, "GetNexrLine:: prod_info.size: %d ", prod_info.size );
    M0sxtrce( dbg );
#endif

    /*
    ** If have not already done so, create an array containing rasterized
    ** image data.
    */

    if ( strcmp( lastname, cur->name ) != 0 ) {


      /*
      ** Allocate space for all of the packed image data in the file
      */

      int datalen = size * size;
      if ( imagedata != NULL ) free( imagedata );

      imagedata = (unsigned char *) calloc( datalen, sizeof(unsigned char) );

      if ( imagedata == (unsigned char *) NULL ) {
        (void) strcpy( err, "GetNexrLine:: Unable to allocate memory" );
        return FAILURE;
      }

      /*
      ** Remember this file as having been read already
      */

      (void) strcpy( lastname, cur->name );

      /*
      ** Open and seek to the beginning of data
      */

#if DEBUG_DIRS
     (void) sprintf( dbg, "GetNexrLine:: call ReadNexrInfo: NEXR_FILE_READ = %d",
                     NEXR_FILE_READ );
     M0sxtrce( dbg );
#endif

      bp = ReadNexrInfo( cur, (int) NEXR_FILE_READ );
      if ( bp == (BUFREAD *) NULL ) {
        (void) sprintf( err, "GetNexrLine:: ERROR reading image in: %s",
                        cur->name );
        M0sxtrce( err );
        return ERR_STAT_BLANK_IMG;
      }
      bp->_ptr += cur->doff;

#if DEBUG_LINE
      (void) sprintf( dbg, "GetNexrLine:: cur->hoff: %d  cur->doff: %d",
                      cur->hoff, cur->doff );
      M0sxtrce( dbg );
#endif

      /*
      ** Create the image
      */

#if DEBUG_LINE
      (void) sprintf( dbg, "radp_head.code: 0x%x", (int)radp_head.code );
      M0sxtrce( dbg );
      (void) sprintf( dbg, "rasp_head.code: 0x%x", (int)rasp_head.code );
      M0sxtrce( dbg );
      (void) sprintf( dbg, "lfm_head.code: 0x%x", (int)lfm_head.code );
      M0sxtrce( dbg );
#endif

      if ( (radp_head.code == 0xAF1F) || (radp_head.code == 0x10) ) {

        if ( !radial_image( bp, imagedata, datalen ) ) {

          M0sxtrce( "GetNexrLine:: ERROR converting radial product to image" );
          return SUCCESS;

        }

      } else if ( (lfm_head.code == 0x11) || (lfm_head.code == 0x12) ) {

        if ( !lfmras_image( bp, imagedata, datalen ) ) {

          M0sxtrce( "GetNexrLine:: ERROR converting LFM raster product to image" );
          return SUCCESS;

        }

      } else {

        if ( !raster_image( bp, imagedata, datalen ) ) {

          M0sxtrce( "GetNexrLine:: ERROR converting raster product to image" );
          return SUCCESS;

        }

      }

    }

    /*
    ** Return the next line of image data.
    **
    ** NOTE: AREA scans are from top-to-bottom; NEXRAD Level III products
    ** are from bottom-to-top so have to index backwards through image array
    **
    ** NOTE: the line counter, read->beglin, is 0-based; the indexing into
    ** the image, imagedata, is 1-based.  This is accounted for.
    */

#if DEBUG_LINE
      (void) sprintf( dbg, "GetNexrLine:: read->beglin = %d, size = %d",
                      read->beglin, size );
      M0sxtrce( dbg );
#endif
    if ( read->beglin < size ) {

      dataptr = imagedata + size * (size - read->beglin - 1);
      (void) memcpy( buf, dataptr+read->begele, read->numele );
      (void) memset( buf+read->numele, 0, MAX_LINE_LEN-read->numele);

    } else {                                /* shouldn't happen but...       */

      M0sxtrce( "ReadNexrLine:: Attempt to read image past EOF, return blanks)" );
      (void) memset( buf, 0, MAX_LINE_LEN );

    }

#if DEBUG_LINE || DEBUG_CALL
    M0sxtrce( "Leave GetNexrLine" );
#endif

    return SUCCESS;

}

/******************************* GetNexrTime *********************************/

int
GetNexrTime( FILELIST *cur, int *date, int *time, int *ss, int *hedsiz, int *pcode )

/*
** Name:    GetNexrTime
**
** Purpose: Get the time of the current NEXRAD Level III product
**
** Note:    This routine reads image data from a NOAAPORT or WSI Level III
**          NEXRAD product.
**
** Parameters:
**             cur    - FILELIST structure containing current image information
**                      name  - full qualified name of the file
**                      pos   - position to add to list
**                      hoff  - byte offset to beginning of header information
**                      doff  - byte offset to beginning of image data
**                      size  - file size in bytes
**                      type  - file type
**                      id    - NEXRAD Level III station ID
**                      time  - file data time [sec since 1970]
**             date   - message date [CCYYDDD] (used as quick date indicator)
**             time   - message time [HHMMSS]  (used as quick time indicator)
**             ss     - image sensor source (=7)
**             hedsiz - header size including product symbology header
**                        (<<<<< UPC mod - added to return list 20120515 >>>>>)
**
** Returns:
**          SUCCESS == 1
**          FAILURE == 0
**
*/

{
    int             rc;                     /* function return status        */

    struct stat     file_stat;              /* 'stat' information            */
    struct tm      *settime;                /* time structure                */

    static BUFREAD *bp;                     /* image data structure          */

#if DEBUG_TIME || DEBUG_CALL
    M0sxtrce( "In GetNexrTime" );
#endif

    /*
    ** Read NEXRAD Level III file
    */

#if DEBUG_DIRS
    (void) sprintf( dbg, "GetNexrTime:: call ReadNexrInfo: NEXR_FILE_READ = %d",
                    NEXR_FILE_READ );
    M0sxtrce( dbg );
#endif

#if 0
    bp = ReadNexrInfo( cur, (int) NEXR_DIR_READ );
#endif
    bp = ReadNexrInfo( cur, (int) NEXR_FILE_READ );
    if ( bp == (BUFREAD *) NULL ) {
      (void) sprintf( dbg, "GetNexrTime:: ERROR reading file: %s", cur->name );
      M0sxtrce( dbg );
      return FAILURE;
    }

    /*
    ** Get product message header (for length of image in product)
    */

#if DEBUG_TIME
    (void) sprintf( dbg, "GetNexrTime:: cur->hoff: %d", cur->hoff );
    M0sxtrce( dbg );

#if 0
    bp->_ptr += cur->hoff + 18;
#endif
#endif

    /* <<<<< UPC mod 20120515 - read product message header >>>>> */
    bp->_ptr += cur->hoff;
    bp->_cnt -= cur->hoff;
    *hedsiz   = cur->hoff;

    /*
    ** Get product message header
    **
    ** Returns:
    **
    **        Bytes Element          Meaning        Range
    **          2   header.code    - code        -131 - -16,  0 - 211
    **          2   header.date    - Julian day     1 - 32767
    **          4   header.time    - time [s]       0 - 86399
    **          4   header.length  - length        18 - 1329270
    **          2   header.source  - source ID      0 - 999
    **          2   header.dest    - dest ID        0 - 999
    **          2   header.blocks  - # blocks       1 - 51  (includes this one)
    **         ----------------------------------------------------------------
    **         18
    */

    rc = read_msghead( bp, &header );
    if ( rc == FAILURE ) {
      M0sxtrce( "GetNexrTime:: ERROR reading msghead" );
      return FAILURE;
    }

    *hedsiz += 18;

#if DEBUG_TIME
    (void) sprintf( dbg,
        "GetNexrTime:: header code %d, date %d, time %d, length %d, blocks %d",
        header.code, header.date, header.time, header.length, header.blocks );
    M0sxtrce( dbg );
#endif

    /*
    ** Read product description block
    **
    ** Returns:
    **
    **        Bytes Element              Meaning        Range
    ** 10       2   product.divider     - divider         -1
    ** 11-12    4   product.latitude    - latitude
    ** 13-14    4   product.longitude   - longitude
    ** 15       2   product.height      - height of radar
    ** 16       2   product.code        - product code
    ** 17       2   product.mode        - radar operational mode
    ** 18       2   product.coverage    - radar coverage pattern
    ** 19       2   product.seqnum      - sequence #
    ** 20       2   product.vs_num      - vol scan #
    ** 21       2   product.vs_date     - vol scan date
    ** 22-23    4   product.vs_time     - vol scan Time
    ** 24       2   product.prod_date   - generation date
    ** 25-26    4   product.prod_time   - generation time
    ** 27       2   product.p1          - product dependent
    ** 28       2   product.p2          - product dependent
    ** 29       2   product.elev_num    - elevation number
    ** 30       2   product.p3          - product dependent
    ** 31-46 16*2   product.threshold   - threshold values
    ** 47       2   product.p4          - product dependent
    ** 48       2   product.p5          - product dependent
    ** 49       2   product.p6          - product dependent
    ** 50       2   product.p7          - product dependent
    ** 51       2   product.p8          - product dependent
    ** 52       2   product.p9          - product dependent
    ** 53       2   product.p10         - product dependent
    ** 54       2   product.num_maps    - Version, Spot Blank
    ** 55-56    4   product.off_symbol  - offset to symbology block
    ** 57-58    4   product.off_graphic - offset to graphic alphanumeric block
    ** 59-60    4   product.off_tabular - offset to tabular alphanumeric block
    **        ----------------------------------------------------------------
    **        102
    */

    rc = read_proddesc( bp, &product );

    if ( rc == FAILURE || (product.divider != -1) ) {
      M0sxtrce( "GetNexrTime:: ERROR reading NEXRAD product description" );
      return FAILURE;
    }

    /* <<<<< UPC add 20120522 - need product code in TestNexrImages >>>>> */

    *pcode = product.code;
    *hedsiz += 102;

#if DEBUG_TIME
    (void) sprintf( dbg, "GetNexrTime:: product vs_date/vs_time:  %d/%d",
                    product.vs_date, product.vs_time );
    M0sxtrce( dbg );
    (void) sprintf( dbg, "GetNexrTime:: product.code:  %d",
                    product.code );
    M0sxtrce( dbg );
    (void) sprintf( dbg, "GetNexrTime:: product.mode:  %d",
                    product.mode );
    M0sxtrce( dbg );
    (void) sprintf( dbg, "GetNexrTime:: product.seqnum:  %d",
                    product.seqnum );
    M0sxtrce( dbg );
    (void) sprintf( dbg, "GetNexrTime:: product.num_maps:  0x%x",
                    product.num_maps );
    M0sxtrce( dbg );
    (void) sprintf( dbg, "GetNexrTime:: product.off_symbol:  %d",
                    product.off_symbol );
    M0sxtrce( dbg );
    (void) sprintf( dbg, "GetNexrTime:: product.off_graphic:  %d",
                    product.off_graphic );
    M0sxtrce( dbg );
    (void) sprintf( dbg, "GetNexrTime:: product.off_tabular:  %d",
                    product.off_tabular );
    M0sxtrce( dbg );
#endif

    /*
    ** Add to hedsiz the number of bytes for product symbology header
    */

    *hedsiz += 18;

    /*
    ** Check to see if we need to figure out the station ID from Lat,Lon
    */

    if ( !strcmp( cur->id, "???" ) )
      rc = StnIdFromLatLon( product.latitude, product.longitude, cur->id );

    /*
    ** Convert volume scan date (days since 700101) and time (seconds since
    ** midnight) to McIDAS date [CCYYDDD] and time [HHMMSS] format.
    */

    cur->time = 86400 * (product.vs_date - 1) + product.vs_time;

    settime = gmtime( &cur->time );

    *date =  1000 * (1900 + settime->tm_year) + settime->tm_yday + 1;
    /* <<<<< UPC mod 20090128 - include seconds portion of time >>>>> */
    *time = 10000 * settime->tm_hour + 100 * settime->tm_min +settime->tm_sec;

#if DEBUG_TIME
    (void) sprintf( dbg, "GetNexrTime:: date: %d time: %06d tsec: %d",
                    *date, *time, cur->time );
    M0sxtrce( dbg );
#endif

    /*
    ** Sensor source hard coded to be 7
    */

    *ss = 7;

    /*
    ** Return
    */

#if DEBUG_TIME
    (void) sprintf( dbg, "GetNexrTime:: date %d, time %d, ss %d",
                    *date, *time, *ss );
    M0sxtrce( dbg );
#endif

#if DEBUG_TIME || DEBUG_CALL
    M0sxtrce( "Leave GetNexrTime" );
#endif

    return SUCCESS;

}

/***************************** StnIdFromLatLon *******************************/

int
StnIdFromLatLon( int lat, int lon, char *id )

/*
** Name:    StnIdFromLatLon
**
** Purpose: Get the NEXRAD station ID from its latitude and longitude
**
** Method:  The station databases (STNDB.[USER|SITE|CORE] are searched for
**          one station of type NEXRAD that fall within a 30 km radius
**          of the station's location as defined in the product.
**
** Parameters:
**             prod - Proddesc structure
**                    lat   - latitude
**                    lon   - longitude
**                    id    - station ID
** Returns:
**          SUCCESS == 1
**          FAILURE == 0
**
*/

{
    int              q_handle;              /* STNDB query handle            */
    int              rc;                    /* function return status        */

    double           delta;
    double           dlat, dlatmin, dlatmax;
    double           dlon, dlonmin, dlonmax;

    double           d2r = 0.017453292;
    STATION         *stn;
    STATION_QRY      stn_query;

    LATLONID        *curid=NULL;
    LATLONID        *newid=NULL;

    static LATLONID *idlist=NULL;

#if DEBUG_STNID || DEBUG_CALL
    M0sxtrce( "In StnIdFromLatLon" );
#endif

    /*
    ** Check existing list of lat,lon,ids to see if this one matches
    */
    curid = idlist;
    while ( curid != (LATLONID *) NULL ) {
      if ( (lat == idlist->lat) && (lon == idlist->lon) ) {
        strcpy( id, idlist->id );
#if DEBUG_STNID
        (void) sprintf( dbg, "StnIdFromLatLon:: returning ID = %s", id );
        M0sxtrce( dbg );
#endif
        return SUCCESS;
      }
      curid = curid->next;
    }

    /*
    ** Setup default query arguments
    */
    rc = McSTNSetQueryDefaults( &stn_query );
    if ( rc < 0 ) {
      M0sxtrce( "StnIdFromLatLon:: Error setting default STNDB query" );
      return( FAILURE );
    }

    /*
    ** Convert input lat and lon to HHMMSS format
    */
    dlat =   (double) lat / 1000.0;
    dlon = - (double) lon / 1000.0;

    /*
    ** Calculate the number of degrees of latitude that corresponds to 30 km
    **   NB: 111.1 km -> 1 degree of longitude
    */
    delta   = cos(dlat * d2r) * (30.0 / 111.1);
    dlatmin = dlat - delta;
    dlatmax = dlat + delta;

    /*
    ** Calculate the number of degrees of longitude that corresponds to 30 km
    **   NB: 111.1 km -> 1 degree of longitude
    */
    delta   = (30.0 / 111.1);
    dlonmin = dlon - delta;
    dlonmax = dlon + delta;

    /*
    ** We want on match for 1 station of type "NEXRAD" in the lat,lon bounds
    ** previously defined
    */
    stn_query.n_datatypes  = 1;
    stn_query.typenames    = (char **) malloc( 7 );
    stn_query.typenames[0] = "NEXRAD";

    stn_query.minlat = mfd_( &dlatmin );
    stn_query.maxlat = mfd_( &dlatmax );
    stn_query.minlon = mfd_( &dlonmin );
    stn_query.maxlon = mfd_( &dlonmax );

#if DEBUG_STNID
    (void) sprintf( dbg,"StnIdFromLatLon:: query minlat: %d",stn_query.minlat );
    M0sxtrce( dbg );
    (void) sprintf( dbg,"StnIdFromLatLon:: query maxlat: %d",stn_query.maxlat );
    M0sxtrce( dbg );
    (void) sprintf( dbg,"StnIdFromLatLon:: query minlon: %d",stn_query.minlon );
    M0sxtrce( dbg );
    (void) sprintf( dbg,"StnIdFromLatLon:: query maxlon: %d",stn_query.maxlon );
    M0sxtrce( dbg );
#endif

    /*
    ** Open the query
    */
    q_handle = McSTNOpenQuery( &stn_query );
    if ( q_handle < 0 ) {
      M0sxtrce( "StnIdFromLatLon:: Error querying STNDB" );
      return( FAILURE );
    }

    /*
    ** Get the number of stations that match the query
    */
    if ( McSTNGetNumStations(q_handle) == 1 ) {

      stn = McSTNGetNextStation( q_handle );
      (void) strcpy( id, stn->id );

      newid = (LATLONID *) calloc( (size_t) 1, sizeof(LATLONID) );
      if ( newid != (LATLONID *) NULL ) {

#if DEBUG_STNID
        (void) sprintf( dbg, "StnIdFromLatLon:: creating new ID node = ", id );
        M0sxtrce( dbg );
#endif
        newid->lat = lat;
        newid->lon = lon;
        strcpy( newid->id, id );

        if ( idlist != (LATLONID *) NULL ) {
          curid = idlist;
          while ( curid->next != (LATLONID *) NULL ) {
            curid = curid->next;
          }
          curid->next = newid;

        } else {
          idlist = newid;

        }

        rc = SUCCESS;

      } else {

        M0sxtrce( "StnIdFromLatLon:: Error creating new ID node" );
        rc = FAILURE;
      }

    } else {

      rc = FAILURE;

    }

    free( stn_query.typenames );
    rc = McSTNCloseQuery( q_handle );

#if DEBUG_STNID || DEBUG_CALL
    M0sxtrce( "Leave StnIdFromLatLon" );
#endif

    return( rc );
}

/******************************* ReadNexrInfo ********************************/

BUFREAD *
ReadNexrInfo( FILELIST *cur, int len )

/*
** Name:    ReadNexrInfo
**
** Purpose: Read bytes from a NEXRAD Level III product into a buffer
**
** Parameters:
**          cur  - FILELIST structure containing current image information
**                 name  - full qualified name of the file
**                 pos   - position to add to list
**                 hoff  - byte offset to beginning of header information
**                 doff  - byte offset to beginning of image data
**                 size  - file size in bytes
**                 type  - file type
**                 id    - NEXRAD Level III station ID
**                 time  - file data time [sec since 1970]
**          len  - number of bytes to read/was read from file
**
** Returns:
**          p    -> pointer to uncompressed image
**          NULL -> failure
**
*/

{
    char           *l;                      /* generic pointer               */

    static unsigned char *data=NULL;        /* pointer to allocated memory   */
    static unsigned char *dest=NULL;        /* pointer to allocated memory   */

    unsigned short  bziped=0;               /* indicates bzip2 compression   */
    int             encrypt=0;              /* indicates encrypted images    */
    int             zlibed=0;               /* indicates zlib compression    */
    int             destLen=0;              /* unbunzip2 image buffer size   */
    int             rc;                     /* generic function return value */

    size_t          nbytes;                 /* number of bytes to malloc     */
    static size_t   lbytes=0;               /* last # of bytes malloced      */

    static BUFREAD *bp=NULL;                /* image data structure          */
    static BUFREAD  wp={NULL};              /* image data structure          */

    FILE           *fd;                     /* file descriptor for 'fread'   */

#if DEBUG_INFO || DEBUG_CALL
    M0sxtrce( "In ReadNexrInfo" );
#endif

#if DEBUG_INFO
    (void) sprintf( dbg, "ReadNexrInfo:: file size: %d, file name: %s, len = %d",
                    cur->size, cur->name, len );
    M0sxtrce( dbg );
#endif

    if ( cur->size == 0 ) {                 /* Find out how big the file is  */
      struct stat     f_stat;               /* 'stat' information            */

      if ( stat( cur->name, &f_stat ) < 0 ) {
        (void) sprintf( dbg, "ReadNexrInfo:: can not stat file: %s",
                        cur->name );
        M0sxtrce( dbg );
        return (BUFREAD *) NULL;
      }

#if DEBUG_INFO
      (void) sprintf( dbg, "ReadNexrInfo:: stat file size: %d",
                      f_stat.st_size );
      M0sxtrce( dbg );
#endif

      cur->size = f_stat.st_size;
    }

    /*
    ** malloc/realloc enough space to do the requested read
    */

    nbytes = ( len == -1 ) ? cur->size : len;

    if ( data == (unsigned char *) NULL ) {
#if DEBUG_INFO
      (void) sprintf( dbg, "ReadNexrInfo:: malloc %d bytes in data", nbytes );
      M0sxtrce( dbg );
#endif
      data = (unsigned char *) malloc( nbytes );
    } else {
#if DEBUG_INFO
      (void) sprintf( dbg, "ReadNexrInfo:: realloc %d bytes in data", nbytes );
      M0sxtrce( dbg );
#endif
      data = (unsigned char *) realloc( data, nbytes );
    }

    if ( data == (unsigned char *) NULL ) {
      (void) sprintf( dbg, "ReadNexrInfo:: ERROR in (re)malloc of %d bytes",
                      nbytes );
      M0sxtrce( dbg );
      return (BUFREAD *) NULL;
    }

    if ( nbytes > lbytes ) lbytes = nbytes;

#if DEBUG_INFO
    (void) sprintf( dbg, "ReadNexrInfo:: %d bytes to read from NEXRAD file",
                    nbytes );
    M0sxtrce( dbg );
#endif

    /*
    ** Open and read the file if first time through
    */

    (void) sprintf( dbg, "ReadNexrInfo:: open file %s", cur->name );
    M0sxtrce( dbg );
    fd = fopen( cur->name, "rb" );
    if ( fd == (FILE *) NULL ) {
      (void) sprintf( dbg, "ReadNexrInfo:: ERROR opening: %s", cur->name );
      M0sxtrce( dbg );
      return (BUFREAD *) NULL;
    }

    (void) sprintf( dbg, "ReadNexrInfo:: read %d bytes from file", nbytes );
    M0sxtrce( dbg );
    rc = fread( data, (size_t) 1, nbytes, fd );
    if ( rc != nbytes )  {
      (void) sprintf( dbg, "ReadNexrInfo:: read %d bytes out of %d",
                      rc, nbytes );
      M0sxtrce( dbg );
      if ( ferror( fd ) != 0 ) {
        (void) sprintf( dbg, "ReadNexrInfo:: ERROR reading file %s", cur->name );
        M0sxtrce( dbg );
        (void) fclose( fd );
        return (BUFREAD *) NULL;
      } else {
        nbytes = rc;
      }
    }

    if ( strlen( cur->id ) == 0 ) {

      char           *p;                    /* generic pointer               */
      char            buf[101]={0};         /* working buffer                */

      /*
      ** Copy the first 100 bytes to the working buffer and null terminate it
      ** so it can be treated like a string.
      */

      (void) memcpy( buf, data, (size_t)100 );
      *(buf+100) = 0;

      /*
      ** Look for "lines" of information in buf.  A "line" of information
      ** is terminated by a sequence of 'CR CR NL' (\r\r\n).  The number
      ** of lines of information will indicate the NEXRAD product origin:
      **
      **        # lines  type  Origin
      **        -------  ----  ------------------------
      **           0       0   WSI
      **           1       1   NOAAPORT like: PIL
      **           2       2   NOAAPORT like: WMO + PIL
      **           3       3   NOAAPORT like: seq + WMO + PIL
      **           4       4   NOAAPORT LDM/SDI: SOH + seq + WMO + PIL
      **          >4      -1   unknown
      **
      ** For WSI NIDS images, the station ID will be found in a separate
      ** record at the end of the file.
      **
      ** In any of the NOAAPORT cases, the station ID will be assumed to be
      ** 6 bytes before the product header.  The offset from the beginning
      ** of the file to the product header is saved in the 'hoff' element
      ** of the 'cur' structure.
      */

      cur->type = 0;
      l = buf;
      while ( (p=strstr((const char *) l, "\r\r\n")) != (char *) NULL ) {
        l = p + 3;
        cur->type++;
      }
      cur->hoff = l - buf;

    } else {

      l = data + cur->hoff;

    }

#if DEBUG_INFO
    (void) sprintf( dbg, "ReadNexrInfo:: %d header bytes in file type %d",
                    cur->hoff, cur->type );
    M0sxtrce( dbg );
#endif

    /*
    ** Test to see if the image looks like it is zlib-compressed.
    ** This first time is designed to find blocks that were
    ** compressed for NOAAPort broadcast
    */

    zlibed = IsZlibed( (unsigned char *) l );

    if ( zlibed ) {

      bp = GetZlibedNexr( data, nbytes, cur->hoff );

      if ( bp == (BUFREAD *) NULL ) {

        M0sxtrce( "ReadNexrInfo:: back from GetZlibedNexr: ERROR uncompressing image" );
        return (BUFREAD *) NULL;

      } else {

        /* Update nbytes with the size of the zlib-uncompressed image */
        nbytes = bp->_cnt;

#if DEBUG_INFO
        (void) sprintf( dbg, "ReadNexrInfo:: realloc %d bytes for data", nbytes );
        M0sxtrce( dbg );
#endif

        /*
        ** realloc data to account for the image new size
        */

        data = (unsigned char *) realloc( data, nbytes );
        if ( data == (unsigned char *) NULL ) {
          (void) sprintf( dbg, "ReadNexrInfo:: ERROR in data remalloc of %d bytes",
                          nbytes );
           M0sxtrce( dbg );
           return (BUFREAD *) NULL;
        }


        /*
        ** Copy the zlib-uncompressed buffer to data; fix up the 'l'
        ** pointer; and set the zlib-compressed flag to zero
        */
        (void *) memcpy( data, bp->_base, nbytes );
        l = data + cur->hoff;
        zlibed = 0;
      }

    }

    /*
    ** Test to see if the image portion looks like it is zlib-compressed,
    ** bzip2 compressed, or if it is encrypted.
    */

    if ( !zlibed ) {

      bziped = IsBziped( (unsigned char *)(l+100), &destLen );

      if ( !bziped ) {

        encrypt = IsEncrypt( l );

      }

    }

    /*
    ** Fill in the file type and station ID
    */

    /* Set the station ID in cur->id <<<<< UPC moved 20220403 >>>>> */
    if ( strlen( cur->id ) == 0 ) {
      (void) memcpy( cur->id, l-6, 3 );
      *(cur->id + 3) = 0;
    }

#if DEBUG_INFO
    (void) sprintf( dbg, "ReadNexrInfo:: cur->type = %d, cur->hoff = %d, cur->size = %d, cur->id = %s",
                    cur->type, cur->hoff, cur->size, cur->id );
    M0sxtrce( dbg );
#endif

    switch ( cur->type ) {

      case 0:
        if ( zlibed || fseek(fd, (long) -4, SEEK_END) ) {
          (void) sprintf( dbg, "ReadNexrInfo:: Unable to seek to ID in: %s",
                          cur->name );
          M0sxtrce( dbg );
          bp = (BUFREAD *) NULL;

        } else {

          /* <<<<< UPC mod 20021216 - Begin >>>>> */
          if ( strlen( cur->id ) == 0 ) {
            int  i;
            char ch;

            M0sxtrce( "ReadNexrInfo:: reading type 0 station ID" );

            rc = fread( cur->id, (size_t) 3, (size_t) 1, fd );

            *(cur->id + 3) = '\0';
            for ( i=0; i<3; i++ ) {
              ch = *(cur->id+i);
              if ( ch < 'A' || ch > 'Z' ) {
                strcpy( cur->id, "???" );
                break;
              }
            }

          }
          /* <<<<< UPC mod 20021216 - End >>>>> */

          wp._ptr  = data;
          wp._base = data;
          wp._cnt  = nbytes;
          bp = &wp;
        }
        break;

      case 1:
      case 2:
      case 3:
      case 4:
      
        if ( zlibed ) {
          bp = GetZlibedNexr( data, nbytes, cur->hoff );
          if ( bp == (BUFREAD *) NULL ) {
            M0sxtrce( "ReadNexrInfo:: back from GetZlibedNexr: ERROR uncompressing image" );
          } else {
            (void) sprintf( dbg, "ReadNexrInfo:: back from GetZlibedNexr, cur->type, = %d, cur->hoff = %d, cur->size = %d, cur->id = %s",
                            cur-> type, cur->hoff, cur->size, cur->id );
            M0sxtrce( dbg );
          }

        } else if ( bziped ) {
          dest = (unsigned char *) malloc( destLen+cur->hoff+120 );
          if ( dest == (unsigned char *)NULL ) {
            M0sxtrce( "ReadNexrInfo:: bziped: ERROR mallocing bzip2 buffer dest" );
            return FAILURE;
          }

          (void) memcpy( dest, data, cur->hoff+120 );

          rc = BZ2_bzBuffToBuffDecompress( dest+cur->hoff+120, &destLen,
                                           data+cur->hoff+120, nbytes-cur->hoff-120,
                                           0, 4 );

          if ( rc ) {
            (void) sprintf( dbg, "ReadNexrInfo:: bziped: ERROR unbzip2ing buffer, rc = %d",
                            rc );
            M0sxtrce( dbg );
            return FAILURE;
          }

          wp._ptr  = dest;
          wp._base = dest;
          wp._cnt  = destLen + cur->hoff + 120;
          bp = &wp;

        } else if ( encrypt ) {
          (void) sprintf( dbg, "ReadNexrInfo:: apparent encrypted image" );
          M0sxtrce( dbg );
          bp = (BUFREAD *) NULL;

        } else {
          wp._ptr  = data;
          wp._base = data;
          wp._cnt  = nbytes;
          bp = &wp;
        }

        break;

      default:
        bp = (BUFREAD *) NULL;
        break;

    }

    (void) fclose( fd );

    /*
    ** Return
    */

#if DEBUG_INFO || DEBUG_CALL
    M0sxtrce( "Leave ReadNexrInfo" );
#endif

    return bp;

}

/****************************** GetZlibedNexr ********************************/

BUFREAD *
GetZlibedNexr( unsigned char *buf, int buflen, int hoff )

/*
** Name:    GetZlibedNexr
**
** Purpose: Read bytes from a NEXRAD Level III product into a buffer
**
** Parameters:
**          buf     - buffer containing Zlib compressed image data
**          buflen  - number of bytes in buf
**          hoff    - offset in buf to start of Zlib compressed data
**
** Returns:
**          bp      -> pointer to uncompressed image
**          NULL    -> failure
**
*/

{
    Bytef           uncompr[ZLIB_BUF_LEN];  /* decompression buffer          */
    uLong           uncomprLen=ZLIB_BUF_LEN;/* length of decompress space    */

    char           *in;                     /* pointer into input buffer     */

    unsigned char  *p;                      /* pointer into outbuf buffer    */
    static unsigned char *outbuf=NULL;      /* buffer (uncompressed image)   */

    int             doff;                   /* # bytes offset to image       */
    int             i;                      /* generic counters              */
    int             numin=0;                /* # input bytes processed       */
    int             rc;                     /* function return               */

    float           ratio;                  /* expansion ratio               */

    uLong           ctot=0;                 /* # compressed bytes processed  */

    size_t          nbytes;                 /* number of bytes to malloc     */
    static size_t   lbytes=0;               /* last # of bytes malloced      */

    z_stream        d_stream;               /* decompression stream          */

    static BUFREAD *bp=NULL;                /* image data structure          */

/*
** z_stream structure
** typedef struct z_stream_s {
**    z_const Bytef           *next_in; // next input byte
**    uInt                    avail_in; // number of bytes available at next_in
**    uLong                   total_in; // total number of input bytes read so far
**
**    Bytef                  *next_out; // next output byte should be put there
**    uInt                   avail_out; // remaining free space at next_out
**    uLong                  total_out; // total number of bytes output so far
**
**    z_const char                *msg; // last error message, NULL if no error
**    struct internal_state FAR *state; // not visible by applications
**
**    alloc_func                zalloc; // used to allocate the internal state
**    free_func                  zfree; // used to free the internal state
**    voidpf                    opaque; // private data object passed to zalloc and zfree
**
**    int                    data_type; // best guess about the data type: binary or text
**    uLong                      adler; // adler32 value of the uncompressed data
**    uLong                   reserved; // reserved for future use
**} z_stream;
*/

#if DEBUG_ZLIBD || DEBUG_CALL
    M0sxtrce( "In GetZlibedNexr" );
#endif

    /*
    ** Create BUFREAD structure
    */

    if ( bp == (BUFREAD *) NULL ) {
      bp = (BUFREAD *) malloc( sizeof(BUFREAD) );
      if ( bp == (BUFREAD *) NULL ) {
        M0sxtrce( "GetZlibedNexr:: ERROR creating BUFREAD structure" );
        return bp;
      }
    }
    bp->_cnt = 0;

    /*
    ** Uncompress first portion of the image.  This should include:
    **
    **     SHO\r\r\n             <--+
    **     SEQ#\r\r\n               |  hoff bytes long
    **     WMO header\r\r\n         |
    **     PIL\r\r\n             <--+
    **
    **  -> CCB
    **     WMO header
    **     PIL
    **     portion of the image
    **
    */

    in    = (char *) (buf + hoff);
    numin = buflen - hoff;

#if DEBUG_ZLIBD
    (void) sprintf( dbg, "GetZlibedNexr:: process %d zlib bytes after %d header bytes",
                   numin, hoff );
    M0sxtrce( dbg );
#endif

    d_stream.zalloc    = (alloc_func) 0;
    d_stream.zfree     = (free_func) 0;
    d_stream.opaque    = (voidpf) 0;

    d_stream.next_in   = (Bytef *) in;
    d_stream.avail_in  = (uInt) numin;
    d_stream.next_out  = uncompr;
    d_stream.avail_out = (uInt) uncomprLen;

    rc = inflateInit( &d_stream );
    if ( rc != Z_OK ) {
      M0sxtrce( "GetZlibedNexr:: ERROR initializing zlib decompression" );
      return (BUFREAD *) NULL;
    }

    rc = inflate( &d_stream, Z_NO_FLUSH );
    if ( (rc == Z_STREAM_END) || (rc == Z_OK) ) {

      unsigned char   b1, b2;

      /*
      ** Find out how long CCB is.  This is done by using the lower order
      ** 6 bits from the first uncompressed byte and all 8 bits of the
      ** second uncompressed byte.
      */

      b1 = uncompr[0];
      b2 = uncompr[1];
      doff  = 2 * (((b1 & 63) << 8) + b2);

#if DEBUG_ZLIBD
      (void) sprintf( dbg, "GetZlibedNexr:: uncompr[0] = 0x%02X, uncompr[1] = 0x%02X, CCB length: %d",
                      uncompr[0], uncompr[1], doff );
      M0sxtrce( dbg );
#endif

      for ( i = 0; i < 2; i++ ) {                         /* eat WMO and PIL in the uncompr buffer */
        while ( (doff < d_stream.total_out) && (uncompr[doff] != '\n') ) doff++;
        doff++;
      }

    } else {

      if ( inflateEnd( &d_stream ) != Z_OK ) {
        M0sxtrce( "GetZlibedNexr:: ERROR at inflateEnd" );
      }
      return (BUFREAD *) NULL;

    }

    ctot = d_stream.total_in;
    bp->_cnt = d_stream.total_out;

#if DEBUG_ZLIBD
    (void) sprintf( dbg, "GetZlibedNexr:: ctot: %d  bp->cnt: %d",
                    (int) ctot, (int) bp->_cnt );
    M0sxtrce( dbg );
#endif

    /*
    ** malloc/realloc enough space to for the uncompressed data.  Estimate
    ** the size of the output buffer based on the compression ratio
    ** found in the first (header) block.  We always add a full uncompressed
    ** block "just to be sure".
    */

    ratio  = ((float) buflen) / ctot;
    nbytes = (ratio + 1.0) * ZLIB_BUF_LEN;

    if ( outbuf == (unsigned char *) NULL ) {
      outbuf = (unsigned char *) malloc( nbytes );

    } else if ( nbytes > lbytes ) {
      outbuf = (unsigned char *) realloc( outbuf, nbytes );

    }

    if ( outbuf == (unsigned char *) NULL ) {
      (void) sprintf( dbg, "GetZlibedNexr:: ERROR in (re)malloc of %d bytes",
                      nbytes );
      M0sxtrce( dbg );
      if ( inflateEnd( &d_stream ) != Z_OK ) {
        M0sxtrce( "GetZlibedNexr:: ERROR at inflateEnd" );
      }
      return (BUFREAD *) NULL;
    }

    bp->_ptr  = outbuf;
    bp->_base = outbuf;

    if ( nbytes > lbytes ) lbytes = nbytes;

#if DEBUG_ZLIBD
    (void) sprintf( dbg, "GetZlibedNexr:: %d bytes for decompression", nbytes );
    M0sxtrce( dbg );
#endif

    /*
    ** Copy header bytes to decompression buffer.  The objective is to
    ** create an output buffer that looks like an uncompressed NOAAPORT
    ** NEXRAD product:
    **
    **   Section               Product               Example             End
    **            +--------------------------------+
    **            |                                |
    **      1     |        start of product        | CTRL-A              \r\r\n
    **            |                                |
    **            +--------------------------------+
    **            |                                |
    **      2     |        sequence number         | 237                 \r\r\n
    **            |                                |
    **            +--------------------------------+
    **            |                                |
    **      3     |          WMO header            | SDUS53 KARX 062213  \r\r\n
    **            |                                |
    **            +--------------------------------+
    **            |                                |
    **      4     |             PIL                | N0RARX              \r\r\n
    **            |                                |
    **            +--------------------------------+
    **            |                                |
    **      5     |                                | AAO130006R2 CH-1
    **            |                                | Interface Control
    **            |             CCB                | Document (ICD)
    **            |                                | for the NWS NWSTG
    **            |                                | Figure 7-1 p 38
    **            |                                |
    **            +--------------------------------+
    **            |                                |
    **      6     |          WMO header            | SDUS53 KARX 062213  \r\r\n
    **            |                                |
    **            +--------------------------------+
    **            |                                |
    **      7     |             PIL                | N0RARX              \r\r\n
    **            |                                |
    **            +--------------------------------+
    **            |                                |
    **            |                                |
    **            |                                |
    **            |                                |
    **      8     |            image               |
    **            |                                |
    **            |                                |
    **            |                                |
    **            |                                |
    **            +--------------------------------+
    **            |                                |
    **      9     |            trailer             | \r\r\nETX
    **            |                                |
    **            +--------------------------------+
    **            |                                |
    **     10     |     Unidata floater trailer    | \0\0
    **            |                                |
    **            +--------------------------------+
    **
    ** Sections 5-8 are zlib compressed.  They must be uncompressed and
    ** read to find out where the image begins.  When this is done, sections
    ** 5-7 are thrown away and 8 is returned immediately following 4.
    ** Section 9 and, if it is there, section 10 are also thrown away.
    **
    */

    /* Copy the header from buf to outbuf */
    (void *) memcpy( outbuf, buf, hoff );
    p = outbuf + hoff;

    /* Copy the bytes just uncompressed to outbuf */
    (void *) memcpy( p, uncompr+doff, d_stream.total_out-doff );
    p += d_stream.total_out - doff;

    /*
    ** Update the input pointer, number of input bytes left, and
    ** the number of bytes processed
    */
    in       += d_stream.total_in;
    numin    -= d_stream.total_in;
    bp->_cnt += hoff - doff;

    /*
    ** Check to see if there are more bytes to uncompress.
    */

    if ( numin <= 0 ) {                       /* No more bytes to process */
      (void) sprintf( dbg, "GetZlibedNexr:: returning %d bytes",
                      (int) bp->_cnt );
      M0sxtrce( dbg );
      if ( inflateEnd( &d_stream ) != Z_OK ) {
        M0sxtrce( "GetZlibedNexr:: ERROR at inflateEnd" );
      }
      return bp;
    }

#if DEBUG_ZLIBD
    (void) sprintf( dbg, "GetZlibedNexr:: %d bytes remaining to process",
                    numin );
    M0sxtrce( dbg );
#endif

    /*
    ** Uncompress the rest of the image.
    **
    **   Section               Product               Example             End
    **            +--------------------------------+
    **            |                                |
    **            |                                |
    **            |                                |
    **            |                                |
    **      8     |            image               |
    **            |                                |
    **            |                                |
    **            |                                |
    **            |                                |
    **            +--------------------------------+
    **            |                                |
    **      9     |            trailer             | \r\r\nETX
    **            |                                |
    **            +--------------------------------+
    **            |                                |
    **     10     |     Unidata floater trailer    | \0\0
    **            |                                |
    **            +--------------------------------+
    */

    while ( numin > 0 ) {

      /*
      ** If last decompress action was at the end of compressed block,
      ** test the next two bytes to see if the image portion looks like
      ** it is zlib-compressed.  This logic was extracted from the
      ** 'inflate.c' routine delivered with the zlib package.
      */

      if ( rc == Z_STREAM_END ) {

        if ( !IsZlibed( (unsigned char *) in ) ) {
          /* The rest should not be compressed, so copy the remaining
          ** bytes to data
          */
          M0sxtrce( "GetZlibedNexr:: !zlibed, copy data to p" );
          (void *) memcpy( p, in, numin );
          bp->_cnt += numin;
          bp->_ptr  = outbuf;
          bp->_base = outbuf;
          break;
        }

        if ( inflateEnd( &d_stream ) != Z_OK ) {
          M0sxtrce( "GetZlibedNexr:: ERROR at inflateEnd" );
        }

      }

      d_stream.next_in   = (Bytef *) in;
      d_stream.avail_in  = (uInt) numin;
      d_stream.next_out  = (Bytef *) p;
      d_stream.avail_out = (uInt) (nbytes - bp->_cnt);

      if ( inflateInit(&d_stream) != Z_OK ) {
        M0sxtrce( "GetZlibedNexr:: ERROR initializing zlib decompression" );
        return (BUFREAD *) NULL;
      }

      rc = inflate( &d_stream, Z_NO_FLUSH );
      if ( (rc == Z_STREAM_END) || (rc == Z_OK) ) {

        p        += d_stream.total_out;
        in       += d_stream.total_in;
        numin    -= d_stream.total_in;
        ctot     += d_stream.total_in;
        bp->_cnt += d_stream.total_out;

#if DEBUG_ZLIBD
        (void)sprintf(dbg,"GetZlibedNexr:: inflate: %d  ctot: %d  bp->cnt: %d",
                      rc, (int)ctot, (int)bp->_cnt);
        M0sxtrce( dbg );
#endif

      } else {

        (void)sprintf(dbg, "GetZlibedNexr:: ERROR %d uncompressing zlib buffer",
                      rc );
        M0sxtrce( dbg );
        return (BUFREAD *) NULL;

      }

#if DEBUG_ZLIBD
      (void) sprintf( dbg, "GetZlibedNexr:: %d input bytes remaining", numin );
      M0sxtrce( dbg );
#endif

      }

      /*
      ** Debugging - write uncompressed data to a disk file for examination
      */

#if DEBUG_ZLIBD
      {
         /* <<<<< UPC add 20220401 - debug output file >>>>> */
         char  oname[] = "/tmp/nexrad3_nexrutil_InGetZlibedNexr_decompressed_buffer";
         FILE *ofile;

         (void) sprintf( dbg, "GetZlibedNexr:: open debug file %s", oname );
         M0sxtrce( dbg );
         ofile = fopen( oname, "w" );
         if( ofile != (FILE *)NULL ) {
            (void) sprintf( dbg, "GetZlibedNexr:: write %d bytes to ofile", bp->_cnt );
            M0sxtrce( dbg );
            fwrite( outbuf, 1, bp->_cnt, ofile );
            fclose( ofile );
         }
      }
#endif

      /*
      ** Image portion decompressed
      */

#if DEBUG_ZLIBD
      M0sxtrce( "Leave GetZlibedNexr" );
#endif
      return bp;

}

/******************************** IsBziped ***********************************/

int
IsBziped( unsigned char *buf, int *len )

/*
** Name:       IsBziped
**
** Purpose:    Check a particular header word to see if the image is
**             bzip2 compressed.  If it is, return the length of the
**             uncompressed image.
**
** Parameters:
**             buf     - buffer containing at least 102 bytes
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
*/

{

    unsigned short p[3];

    int            three=3;

#if DEBUG_ISBZP || DEBUG_CALL
    M0sxtrce( "In IsBziped" );
#endif

    /*
    ** Check product specific dword 'p8' (product description block)
    */

    (void) memcpy( p, buf, 6 );
    (void) swbyt2_( p, &three );

    if ( p[0] == 1 ) {

      *len = 65536 * p[1] + p[2];

#if DEBUG_ISBZP || DEBUG_CALL
    M0sxtrce( "Leave IsBziped:: bziped" );
#endif

      return SUCCESS;

    }

    *len = 0;

#if DEBUG_ISBZP || DEBUG_CALL
    M0sxtrce( "Leave IsBziped:: !bziped" );
#endif

    return FAILURE;
}

/******************************** IsEncrypt **********************************/

int
IsEncrypt( char *buf )

/*
** Name:       IsEncrypt
**
** Purpose:    Check a two-byte sequence to see if it indicates the start of
**             an encrypted image.
**
** Parameters:
**             buf     - buffer containing at least two bytes
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
*/

{

#if DEBUG_ISENC || DEBUG_CALL
    M0sxtrce( "In IsEncrypt" );
#endif

    /*
    ** These tests were deduced from inspection from encrypted NOAAPORT files.
    */

    if ( !strncmp( buf, "R3", 2 ) ) {
      return SUCCESS;
    }

    return FAILURE;
}

/******************************** IsZlibed ***********************************/

int
IsZlibed( unsigned char *buf )

/*
** Name:       IsZlibed
**
** Purpose:    Check a two-byte sequence to see if it indicates the start of
**             a zlib-compressed buffer
**
** Parameters:
**             buf     - buffer containing at least two bytes
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
*/

{

#if DEBUG_ISZLB || DEBUG_CALL
    M0sxtrce( "In IsZlibed" );
#endif

    /*
    ** These tests were extracted from the 'inflate.c' routine that is
    ** included in the zlib distribution.
    */

    if ( (*buf & 0xf) == Z_DEFLATED ) {
      if ( (*buf >> 4) + 8 <= DEF_WBITS ) {
        if ( !(((*buf << 8) + *(buf+1)) % 31) ) {
#if DEBUG_ISZLB || DEBUG_CALL
          M0sxtrce( "Leave IsZlibed:: zlibed" );
#endif
          return SUCCESS;

        }
      }
    }

#if DEBUG_ISZLB || DEBUG_CALL
    M0sxtrce( "Leave IsZlibed:: !zlibed" );
#endif

    return FAILURE;
}

/******************************** GetFloat16 *********************************/

float
GetFloat16( unsigned short buf )

/*
** Name:       GetFloat16
**
** Purpose:    unpack 2-byte half word into 16-bit float
**
** Parameters:
**             buf     - buffer containing at least two bytes
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
*/

{
    int            isign;                   /* sign of 16-bit float          */

    float          expon;                   /* exponent of 16-bit float      */
    float          fract;                   /* fraction of 16-bit float      */
    float          value;                   /* extracted floating point val  */

#if DEBUG_FLT16 || DEBUG_CALL
    M0sxtrce( "In GetFloat16" );
#endif

    /*
    ** The meaning of bits is defined in ICD for WSR-88D ROC RPG Build 11.0
    ** Table 3.6, Note 1
    ** +----------------------------------------------------------------+
    ** | S | E | E | E | E | E | F | F | F | F | F | F | F | F | F | F  |
    ** |___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|____|
    ** | 0 | 1 |   |   |   | 5 | 6 |   |   |   |   |   |   |   |   | 15 |
    ** |___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|____|
    ** |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   |    |
    ** |___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|____|
    **
    **     E <   0, value = (-1)**S * 2         * (0 + (F/2**10))
    ** 0 < E < 255, value = (-1)**S * 2**(E-16) * (1 + (F/2**10))
    **
    ** 2**10 == 1024
    */

    isign = (buf & 0x8000) ? -1 : 1;
    expon = (buf & 0x7C00) >> 10;
    fract = (buf & 0x03FF);
    if ( !expon ) {
      value = isign * 2.0 * (fract/1024.0);
    } else {
      value = isign * pow(2.0,expon-16.0) * (1 + fract/1024.0);
    }

#if DEBUG_FLT16
    (void) sprintf( dbg, "GetFloat16:: linear scale = 0x%X, %8.4f", buf, value );
    M0sxtrce( dbg );
#endif

#if DEBUG_FLT16 || DEBUG_CALL
    M0sxtrce( "Leave GetFloat16" );
#endif

    return value;
}

/******************************** GetFloat32 *********************************/

float
GetFloat32( unsigned short buf1, unsigned short buf2 )

/*
** Name:       GetFloat32
**
** Purpose:    convert 2 2-byte half words into 32-bit float
**
** Parameters:
**             buf1    - buffer containing at least two bytes
**             buf2    - buffer containing at least two bytes
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
*/

{
    Fint  one=1;
    float value;

    union {
      unsigned short ival[2];
      unsigned int   ival4;
               float val;
    } v;

    v.ival[0] = buf1;
    v.ival[1] = buf2;

#if DEBUG_FLT32 || DEBUG_CALL
    M0sxtrce( "In GetFloat32" );
#endif

#if DEBUG_FLT32
    (void) sprintf( dbg,
      "GetFloat32 raw:: buf1, buf2, ival4 = 0x%04X, 0x%04X, 0x%08X",
      v.ival[0], v.ival[1], v.ival4 );
    M0sxtrce( dbg );
#endif

    (void) swbyt2_(&(v.ival[0]), &one);
    (void) swbyt2_(&(v.ival[1]), &one);
    (void) swbyt4_(&(v.ival4),   &one);

    value = v.val;

#if DEBUG_FLT32
    (void) sprintf( dbg,
      "GetFloat32 new:: buf1, buf2, ival4, value = 0x%04X, 0x%04X, 0x%08X, %8.4f",
      v.ival[0], v.ival[1], v.ival4, value );
    M0sxtrce( dbg );
#endif

#if DEBUG_FLT32 || DEBUG_CALL
    M0sxtrce( "Leave GetFloat32" );
#endif

    return value;
}

/***************************** SelectNexrImages ******************************/

int
SelectNexrImages( CRITERIA *request, FILELIST **satisfy, char *err)

/*
** Name:       SelectNexrImages
**
** Purpose:    Return a list of NEXRAD Level III products
**
** Parameters:
**             cur     - FILELIST structure containing current image information
**                       name  - full qualified name of the file
**                       pos   - position to add to list
**                       hoff  - byte offset to beginning of header information
**                       doff  - byte offset to beginning of image data
**                       size  - file size in bytes
**                       type  - file type
**                       id    - NEXRAD Level III station ID
**                       time  - file data time [sec since 1970]
**             dir     - directory under which files will be found
**             bpos    - beginning data set position
**             epos    - ending data set position
**             request - request criteria
**             satisfy - list of files that satisfy criteria
**             err     - textual error message
**
** Returns:
**             SUCCESS 1
**
*/

{

    int         ALL=lit_( "ALL ", 4 );      /* select all images in dir      */
    int         curpos=0;                   /* initial position number       */
    int         rc;                         /* function return status        */

    time_t      dtime=0;                    /* data time [sec since 1970]    */

    FILELIST    list={0};                   /* local FILELIST object         */

    FILELIST   *head=NULL;                  /* list of files in directory    */
    FILELIST   *cur=NULL;                   /* list of files in directory    */

    FILELIST   *rawlist=NULL;               /* list of files in directory    */

    FILELIST   *test=NULL;                  /* list of files to test         */
    FILELIST   *timlist=NULL;               /* list of time ordered files    */

#if DEBUG_SELIM || DEBUG_CALL
    M0sxtrce( "In SelectNexrImages" );
#endif

    /*
    ** The integer representation of 'ALL ' means return all the images in dir
    */

    M0swbyt4( &ALL, 1 );

    /*
    ** Generate a list of files whose names match 'request->filemask' in
    ** all directories under 'dir'.
    */

    rc = GetFileList( request, &rawlist );

    if ( rc == FAILURE ) {

      (void)sprintf( err, "No images meet the selection criteria " );
      M0sxtrce( "Leave SelectNexrImages:: ERROR GetFileList" );
      return FAILURE;

    }

    /*
    ** Test images found to see if they match the patterns in 'request'.
    ** Add those that match to the renumbered list 'head'.
    */

    rc = TestNexrImages( rawlist, &head, request );

    if ( rc == FAILURE ) {
      (void) strcpy( err, "SelectNexrImages:: TestNexrImages ERROR" );
      M0sxtrce( "Leave SelectNexrImages:: ERROR TestNexrImages" );
      return FAILURE;
    }

    /*
    ** Create the return list of files
    */

    if ( request->bpos == ALL ) {           /* ALL files requested           */

      M0sxtrce( "SelectNexrImages:: ALL files requested" );

      *satisfy = head;

    } else if ( request->bpos > 0 ) {       /* absolute positions requested  */

      M0sxtrce( "SelectNexrImages:: absolute file positions" );

      while ( head != NULL ) {

        /*
        ** Pop a file off of the stack and check position
        */

        rc = PopFile( &list, &head );

        if (list.pos < request->bpos || list.pos > request->epos ) continue;

        /*
        ** Push file that passes test onto new 'test' stack
        */

        rc = PushFileByName( &list, &test );

        if ( rc == FAILURE ) {
          (void) strcpy( err, "Unable to allocate memory" );
          M0sxtrce( "Leave SelectNexrImages:: ERROR PushFileByName" );
          return FAILURE;
        }

      }

      *satisfy = test;

    } else if ( request->bpos <= 0 ) {      /* time relative file positions  */

      M0sxtrce( "SelectNexrImages:: relative file positions" );

      cur = head;
      while ( cur != NULL ) {               /* Reorder by time               */

        rc = PushFileByTime( cur, &test );
        cur = cur->next;

      }

      /*
      ** Reorder file list by time
      */

      curpos = 1;

      while ( test != NULL ) {

        /*
        ** Pop a file off of the stack and check position
        */

        curpos--;

        rc = PopFile( &list, &test );
        if (curpos > request->bpos || curpos < request->epos) continue;

        /*
        ** Push file that passes test onto new 'timlist' stack
        */

        rc = PushFileByTime( &list, &timlist );

        if ( rc == FAILURE ) {
          (void) strcpy( err, "Unable to allocate memory" );

#if DEBUG_SELIM || DEBUG_CALL
          M0sxtrce( "Leave SelectNexrImages:: ERROR PushFileByTime" );
#endif

          return FAILURE;
        }

      }

      *satisfy = timlist;

    }

#if DEBUG_SELIM || DEBUG_CALL
    M0sxtrce( "Leave SelectNexrImages:: SUCCESS" );
#endif

    return SUCCESS;

}

/****************************** TestNexrImages *******************************/

int
TestNexrImages( FILELIST *oldlist, FILELIST **newlist, CRITERIA *request )

/*
** Name:       TestNexrImages
**
** Purpose:    Test an image against the specification criteria detailed
**             in struct request
**
** Parameters:
**             oldlist - FILELIST list containing current image information
**                       name  - full qualified name of the file
**                       pos   - position to add to list
**                       hoff  - byte offset to beginning of header information
**                       doff  - byte offset to beginning of image data
**                       size  - file size in bytes
**                       type  - file type
**                       id    - NEXRAD Level III station ID
**                       time  - file data time [sec since 1970]
**             newlist - FILELIST list of images satisfing 'request' criteria
**             request - CRITERIA to test name
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
** NOTES:
**             No tests will be performed for CRITERIA values of -1
**
*/

{

    int         curpos=0;
    int         idislist;

    int         pcode;                      /* product code                  */
    int         hedsiz;                     /* product header length         */
    int         imgday;                     /* image day [CCYYDDD]           */
    int         imgsiz;                     /* size of image portion of file */
    int         imgss;                      /* satellite id for image        */
    int         imgtim;                     /* image time [HHMMSS]           */
    int         rc;                         /* function status               */

    FILELIST   *cur;                        /* current element of 'oldlist'  */
    FILELIST   *old;                        /* previous element of 'oldlist' */

    FILELIST   *last;                       /* previous element of 'newlist' */
    FILELIST   *next;                       /* next element of 'newlist'     */
    FILELIST   *this;                       /* current element of 'newlist'  */

#if DEBUG_TEST || DEBUG_CALL
    M0sxtrce( "In TestNexrImages" );
#endif

    /*
    ** Save information that the requested ID is "LIST"
    */

    idislist = !strcmp( request->id, "LIST" );

    /*
    ** Loop through input list of images and create an ouput list
    ** that has only ones that meet 'request' criteria
    */

    cur = oldlist;

    while ( cur != (FILELIST *) NULL ) {

      curpos++;

      /*
      ** Get McIDAS date, time, and sensor source information for this image
      */

      /* <<<<< UPC mod 20120514 - added hedsize to list of passed parms >>>>> */
      rc = GetNexrTime( cur, &imgday, &imgtim, &imgss, &hedsiz, &pcode );

      if ( rc == FAILURE ) {
        M0sxtrce( "TestNexrImages:: ERROR return from GetNexrTime" );
        old = cur;
        cur = cur->next;
        free( old );
        continue;
      }

      /*
      ** Test product size (needed because of short Dual Pol products)
      */

#if DEBUG_TEST
      (void)sprintf( dbg, "TestNexrImages:: cur->size %d hedsiz %d",
                     cur->size, hedsiz );
      M0sxtrce( dbg );
#endif
      
      /*
      ** <<<<< UPC add 20120514 - hack for bad (too small) Dual Pol images >>>>>
      ** Note:
      **      The test for (cur->size-hedsiz) being less than 600 is arbitrary.
      **      The "real" way to tell if the image is too small is to crack the
      **      image down to the point where its shape is known and then use
      **      that to come up with a heuristic for what is unacceptably
      **      small. I did not do this because it would slow things down
      **      too much.
      **
      **      Typical (cur->size-hedsiz) for small Dual Pol products seen are:
      **      prod type size description
      **      ----+----+----+------------------------------------------------
      **      169  OHA    60 QPE ONE HOUR ACCUMULATION 16 LEVELS
      **      170  DAA    60 QPE ONE HOUR DIGITAL ACCUMULATION ARRAY
      **      174  DOD   550 QPE-PPS ONE HOUR ACCUMULATION DIFFERENCE ID 175
      **      175  DSD   166 QPE-PPS STORM TOTAL ACCUMULATION DIFFERENCE
      */

      if ( (pcode==169 || pcode==170 || pcode==174 || pcode==175) &&
           (cur->size-hedsiz) < 600 ) {
        (void)sprintf( dbg,
        "TestNexrImages:: Size test exclusion: cur->size: %d, hedsiz %d, pcode %d",
        cur->size, hedsiz, pcode );
        M0sxtrce( dbg );
        old = cur;
        cur = cur->next;
        free( old );
        continue;
      }

      /*
      ** Test radar ID
      */

      if ( request->id[0] != '*' && !idislist ) {

        if ( strcmp(request->id, cur->id) != 0 ) {
          (void)sprintf( dbg,
                "TestNexrImages:: ID test exclusion: request ID: %s  image ID: %s",
                request->id, cur->id);
          M0sxtrce( dbg );
          old = cur;
          cur = cur->next;
          free( old );
          continue;
        }

      }

      /*
      ** Test beginning and ending day
      */

      if ( request->begday >= 0 ) {

        if (imgday < request->begday || imgday > request->endday) {
          M0sxtrce( "TestNexrImages:: Begin/end day test exclusion" );
          old = cur;
          cur = cur->next;
          free( old );
          continue;
        }

      }

      /*
      ** Test beginning and ending time
      */

      if ( request->begtim >= 0 ) {

        /* <<<<< UPC mod 20090205 - limit comparison to HHMM >>>>> */
        int reqbtim = request->begtim / 100;
        int reqetim = request->endtim / 100;
        int imghhmm = imgtim / 100;

        if (imghhmm < reqbtim || imghhmm > reqetim) {
          M0sxtrce( "TestNexrImages:: Begin/end time test exclusion" );
          old = cur;
          cur = cur->next;
          free( old );
          continue;
        }

      }

      /*
      ** Test beginning and ending McIDAS ss number
      */

      if ( request->begss >= 0 ) {

        if (imgss < request->begss || imgss > request->endss) {
          M0sxtrce( "TestNexrImages:: Begin/end SS test exclusion" );
          old = cur;
          cur = cur->next;
          free( old );
          continue;
        }

      }

      /*
      ** At this point, we have an image that meets all 'request'
      ** criteria.  If the 'request' station ID is "LIST" we
      ** need to compare the time of the current image with
      ** images from the same station in the ouput list and keep
      ** the newest image.
      */

      cur->pos = curpos;

      if ( *newlist == (FILELIST *) NULL ) {         /* output list is empty */

        *newlist   = cur;
        last       = cur;
        cur        = cur->next;
        last->next = (FILELIST *) NULL;
        continue;

      }

      if ( !idislist ) {                           /* request ID is not LIST */

        last->next = cur;
        last       = cur;
        cur        = cur->next;
        last->next = (FILELIST *) NULL;

      } else {                                         /* request ID is LIST */

        /*
        ** Search for a 'newlist' entry that has ID that matches 'cur->id'
        */

        this = *newlist;

        while ( this != (FILELIST *) NULL && strcmp(this->id,cur->id) ) {
          this = this->next;
        }

        /*
        ** We either have found a matching ID or are at the end of 'newlist'
        */

        if ( this == (FILELIST *) NULL ) {     /* did not find a matching ID */

          last->next = cur;              /* save new image info in 'newlist' */
          last       = cur;
          cur        = cur->next;
          last->next = (FILELIST *) NULL;

        } else {                                      /* found a matching ID */

          if ( cur->time > this->time ) {   /* save newer image in 'newlist' */
            next       = this->next;
            *this      = *cur;
            this->next = next;
          }

          old = cur;
          cur = cur->next;
          free( old );

        }

      }

    }

    /*
    ** Done
    */

#if DEBUG_TEST || DEBUG_CALL
    M0sxtrce( "Leave TestNexrImages:: SUCCESS" );
#endif

    return SUCCESS;

}

/******************************** code_lookup ********************************/

static void
code_lookup( int code, Prodinfo *prod_info )

/*
** Name:       code_lookup
**
** Purpose:    Codify some invariant metadata
**
** Parameters:
**             code      - product code
**             prod_info - product information structure
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
** History: 1996???? - Developed by Harry Edmon of U. Washington
**          19980511 - Adapted from ldm-mcidas distribution for McIDAS-X 7.3
**          20100115 - Extend definitions up to product type 189
**          20150608 - Fix definitions for products type 180, 183, 186, 187
*/

{

/*
** Inteface Control Document for the RPG to Class 1 User
**
** WSR-88D  Radar Operations Center
** 1313 Halley Circle
** Norman, OK 73069
**
** Document Number 2620001R
** Code Identification 0WY55
** WSR-88D ROC
** 08 October 2010
** Open Build 12.1
**
** and
**
** TDWR Interface Control and Specifications Documentation
** for the NWS Supplemental Product Generator
**
** National Oceanic and Atmospheric Administration
** National Weather Service
** Office of Science &Technology Systems Engineering Center
** August 3, 2005
** Version 4.3
**
** NB: data resolutions represent conversions from nmi to km
**
*/
    static float data_res[200] = {
/*    0     1     2     3     4     5       6     7     8     0                
  ----|-----|-----|-----|-----|-----|-------|-----|-----|-----|----------------*/
   0.00, 0.00, 0.00, 0.00, 0.00, 0.00,   0.00, 0.00, 0.00, 0.00,    /*   0-  9 */
   0.00, 0.00, 0.00, 0.00, 0.00, 0.00,   1.00, 2.00, 4.00, 1.00,    /*  10- 19 */
   2.00, 4.00, 0.25, 0.50, 1.00, 0.25,   0.50, 1.00, 0.25, 0.50,    /*  20- 29 */
   1.00, 2.00, 1.00, 1.00, 1.00, 1.00,   4.00, 1.00, 4.00, 0.00,    /*  30- 39 */
   0.00, 4.00, 0.00, 0.00, 0.00, 0.00,   0.00, 0.00, 0.00, 0.00,    /*  40- 49 */
   1.00, 1.00, 1.00, 1.00, 1.00, 0.50,   1.00, 4.00, 0.00, 0.00,    /*  50- 59 */
   0.00, 0.00, 0.00, 4.00, 4.00, 4.00,   4.00, 4.00, 0.00, 0.00,    /*  60- 69 */
   0.00, 0.00, 0.00, 0.00, 0.00, 0.00,   0.00, 0.00, 2.00, 2.00,    /*  70- 79 */
   2.00, 4.00, 0.00, 0.00, 0.00, 1.00,   1.00, 0.00, 0.00, 4.00,    /*  80- 89 */
   4.00, 0.00, 0.00, 1.00, 1.00, 1.00,   4.00, 1.00, 4.00, 0.25,    /*  90- 99 */
   0.00, 0.00, 0.00, 0.00, 0.00, 0.00,   0.00, 0.00, 0.00, 0.00,    /* 100-109 */
   0.00, 0.00, 0.00, 0.00, 0.00, 0.00,   0.00, 0.00, 0.00, 0.00,    /* 110-119 */
   0.00, 0.00, 0.00, 0.00, 0.00, 0.00,   0.00, 0.00, 0.00, 0.00,    /* 120-129 */
   0.00, 0.00, 1.00, 1.00, 1.00, 1.00,   1.00, 1.00, 2.00, 0.00,    /* 130-139 */
   0.00, 0.00, 0.00, 0.00, 1.00, 1.00,   1.00, 1.00, 0.00, 0.00,    /* 140-149 */
   1.00, 1.00, 0.00, 0.25, 0.25, 0.25,   2.00, 2.00, 0.00, 0.25,    /* 150-159 */
   0.00, 0.25, 0.00, 0.25, 0.25, 0.25,   0.00, 0.00, 0.00, 2.00,    /* 160-169 */
   0.25, 2.00, 0.25, 0.25, 0.25, 0.25,   0.25, 0.25, 0.00, 0.00,    /* 170-179 */
   0.15, 0.15, 0.15, 0.15, 0.00, 0.15,   0.30, 0.30, 0.00, 0.00,    /* 180-189 */
   0.00, 0.00, 0.00, 0.00, 1.00, 0.00,   0.00, 0.00, 0.00, 0.50,    /* 190-199 */
/*----|-----|-----|-----|-----|-----|-------|-----|-----|-----|----------------
      0     1     2     3     4     5       6     7     8     0                */
    };

    static int levels[200] = {
/*    0     1     2     3     4     5       6     7     8     0                
  ----|-----|-----|-----|-----|-----|-------|-----|-----|-----|----------------*/
      0,    0,    0,    0,    0,    0,      0,    0,    0,    0,    /*   0-  9 */
      0,    0,    0,    0,    0,    0,      8,    8,    8,   16,    /*  10- 19 */
     16,   16,    8,    8,    8,   16,     16,   16,    8,    8,    /*  20- 29 */
      8,   16,  256,   16,    8,    8,      8,   16,   16,    0,    /*  30- 39 */
      0,   16,    0,    0,    0,    0,      0,    0,    0,   16,    /*  40- 49 */
     16,   16,    0,    0,    0,   16,     16,   16,    0,    0,    /*  50- 59 */
      0,    0,    0,    8,    8,    8,      8,    8,    0,    0,    /*  60- 69 */
      0,    0,    0,    0,    9,    0,      0,    0,   16,   16,    /*  70- 79 */
     16,  256,    0,    0,    0,    8,      8,    0,    0,    8,    /*  80- 89 */
      8,    0,    0,  256,  256,    8,      8,   16,   16,  256,    /*  90- 99 */
      0,    0,    0,    0,    0,    0,      0,    0,    0,    0,    /* 100-109 */
      0,    0,    0,    0,    0,    0,      0,    0,    0,    0,    /* 110-119 */
      0,    0,    0,    0,    0,    0,      0,    0,    0,    0,    /* 120-129 */
      0,    0,   11,   12,  256,   72,      0,   16,  256,    0,    /* 130-139 */
      0,    0,    0,    0,   16,   16,     16,   16,    0,    0,    /* 140-149 */
     16,   16,    0,  256,  256,  256,     64,    8,   16,  256,    /* 150-159 */
     16,  256,   16,  256,   16,  160,      0,    0,    0,   16,    /* 160-169 */
    256,   16,  256,  256,  256,  256,  65536,  160,    0,    0,    /* 170-179 */
    256,   16,  256,   16,    0,    8,    256,   16,    0,    0,    /* 180-189 */
      0,    0,    0,    0,  256,    0,      0,    0,    0,  256,    /* 190-199 */
/*----|-----|-----|-----|-----|-----|-------|-----|-----|-----|----------------
      0     1     2     3     4     5       6     7     8     0                */
    };

#if DEBUG_CODLU || DEBUG_CALL
    M0sxtrce( "In code_lookup" );
#endif

    prod_info->code = code;

    if ( code < 0 || code > 199 ) {
      prod_info->data_res = 0;
      prod_info->levels   = 0;
    } else {
      prod_info->data_res = data_res[code];
      prod_info->levels   = levels[code];
    }

#if DEBUG_CODLU || DEBUG_CALL
    M0sxtrce( "Leave code_lookup" );
#endif
    return;

}

/*********************************** getbuf **********************************/

int
getbuf( BUFREAD *bp, char *buf, int nelem, int eltsize )

/*
** Name:       getbuf
**
** Purpose:    Get n bytes from a file swapping bytes for little endian
**             systems
**
** Parameters:
**             bp      - image data buffer structure
**             buf     - return buffer
**             nelem   - number of elements to process
**             eltsize - element size
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
** History: 1996???? - Developed by Harry Edmon of U. Washington
**          19980511 - Adapted from ldm-mcidas distribution for McIDAS-X 7.3
**          20001208 - Modified to read from buffer structure
**
*/

{
    int         nleft;                      /* number of bytes remaining     */
    int         ncopy;                      /* number of bytes to copy       */

#if DEBUG_GTBUF || DEBUG_CALL
    M0sxtrce( "In getbuf" );
#endif

    /*
    ** Request sanity check
    */

    ncopy = nelem * eltsize;

    if ( ncopy <= 0 ) return FAILURE;

    /*
    ** Copy nbytes from 'bp->_ptr' to buf and swap bytes on little
    ** endian systems.
    */

    nleft = bp->_cnt - (bp->_ptr - bp->_base);

#if DEBUG_GTBUF
    (void) sprintf( dbg, "getbuf:: bp->_cnt: %d, bp->ptr: %ld, bp->_base: %ld",
                    bp->_cnt, bp->_ptr, bp->_base );
    M0sxtrce( dbg );
    (void) sprintf( dbg, "getbuf:: ncopy: %d  nleft: %d", ncopy, nleft );
    M0sxtrce( dbg );
#endif

    if ( nleft >= ncopy ) {
      (void) memcpy( buf, bp->_ptr, ncopy );
      bp->_ptr += ncopy;
    } else {
      (void) sprintf( dbg, "getbuf:: ncopy = %d, nleft = %d", ncopy, nleft );
      M0sxtrce( dbg );
      return FAILURE;
    }

    /*
    ** Swap bytes on little-endian systems
    */

    switch ( eltsize ) {

      case 2:
          (void) swbyt2_( buf, &nelem );
          break;

      case 4:
          (void) M0swbyt4( buf, nelem );
          break;

    }

#if DEBUG_GTBUF || DEBUG_CALL
    M0sxtrce( "Leave getbuf" );
#endif

    return SUCCESS;

}

/******************************* read_dividlen *******************************/

int
read_dividlen( BUFREAD *bp, product_symbology_header *buf )

/*
** Name:       read_dividlen
**
** Purpose:    Read divider ID header from NEXRAD Level III product
**
** Parameters:
**             bp      - input data buffer structure
**             buf     - divider ID structure
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
** History: 1996???? - Developed by Harry Edmon of U. Washington
**          19980511 - Adapted from ldm-mcidas distribution for McIDAS-X 7.3
**          20001128 - Recast to read from buffer, not file
**
*/

{

#if DEBUG_DVDLN || DEBUG_CALL
    M0sxtrce( "In read_dividlen" );
#endif

    return getbuf( bp, (char *)&buf->divider, 1, 2 ) &&
           getbuf( bp, (char *)&buf->id,      1, 2 ) &&
           getbuf( bp, (char *)&buf->length,  1, 4 ) &&
           getbuf( bp, (char *)&buf->layers,  1, 2 );
}

/******************************** read_msghead *******************************/

int
read_msghead( BUFREAD *bp, Msghead *buf )

/*
** Name:       read_msghead
**
** Purpose:    Read message header from NEXRAD Level III product
**
** Parameters:
**             info    - buffer containing image file data
**             buf     - message header structure
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
** Document # 2620001M
** Code Identification 0WY55
** WSR-88D ROC
** 03 March 2009
** RPG Build 11.0
**
** Figure 3-3. Message Header
**
** HWord  Fieldname  Type  Units      Range     Precision    Remarks
** 01     Message    INT*2 N/A   -131 to -16    N/A        Message code
**        Code                      0 to 211               from Table II
** 02     Date of    INT*2 Julian   1 to 32767   1         Modified
**        Message          Day                             Julian Date at time
**                                                         of transmission
**                                                         (number of days
**                                                         since 1 January
**                                                         1970, where 1=1
**                                                         January 1970). To
**                                                         obtain actual Julian
**                                                         Date, add
**                                                         2,440,586.5 to the
**                                                         modified date
** 03-04  Time of    INT*4 Seconds    0-86399    1         Number of seconds
**        Message                                          after midnight,
**                                                         Greenwich Mean
**                                                         Time (GMT).
** 05-06  Length of  INT*4  N/A     18-1329270   1         Number of bytes in
**        Message                                          message including
**                                                         header
** 07     Source ID  INT*2  N/A      0-999       1         Source (originators')
**                                                         ID of the sender
** 08     Dest ID    INT*2  N/A      0-999       1         Destination ID
**                                                         (receivers') for
**                                                         message
**                                                         transmission
** 09     # Blocks   INT*2  N/A      1-51        1         Header Block plus
**                                                         the Product
**                                                         Description Blocks
**                                                         in message
**
** History: 1996???? - Developed by Harry Edmon of U. Washington
**          19980511 - Adapted from ldm-mcidas distribution for McIDAS-X 7.3
**          20001128 - Recast to read from buffer, not file
**
*/

{

#if DEBUG_MSGHD || DEBUG_CALL
    M0sxtrce( "In read_msghead" );
#endif

    return getbuf( bp, (char *)&buf->code,   1, 2 ) &&
           getbuf( bp, (char *)&buf->date,   1, 2 ) &&
           getbuf( bp, (char *)&buf->time,   1, 4 ) &&
           getbuf( bp, (char *)&buf->length, 1, 4 ) &&
           getbuf( bp, (char *)&buf->source, 3, 2 );

}

/******************************** read_proddesc ******************************/

int
read_proddesc( BUFREAD *bp, Proddesc *buf )

/*
** Name:       read_proddesc
**
** Purpose:    Read product description header from NEXRAD Level III product
**
** Parameters:
**             info    - buffer containing image file data
**             buf     - product description structure
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
** History: 1996???? - Developed by Harry Edmon of U. Washington
**          19980511 - Adapted from ldm-mcidas distribution for McIDAS-X 7.3
**          20001128 - Recast to read from buffer, not file
**
*/

{

#if DEBUG_PRDES || DEBUG_CALL
    M0sxtrce( "In read_proddesc" );
#endif

    return getbuf( bp, (char *)&buf->divider,     1, 2 ) &&
           getbuf( bp, (char *)&buf->latitude,    1, 4 ) &&
           getbuf( bp, (char *)&buf->longitude,   1, 4 ) &&
           getbuf( bp, (char *)&buf->height,      1, 2 ) &&
           getbuf( bp, (char *)&buf->code,        1, 2 ) &&
           getbuf( bp, (char *)&buf->mode,        1, 2 ) &&
           getbuf( bp, (char *)&buf->coverage,    1, 2 ) &&
           getbuf( bp, (char *)&buf->seqnum,      1, 2 ) &&
           getbuf( bp, (char *)&buf->vs_num,      1, 2 ) &&
           getbuf( bp, (char *)&buf->vs_date,     1, 2 ) &&
           getbuf( bp, (char *)&buf->vs_time,     1, 4 ) &&
           getbuf( bp, (char *)&buf->prod_date,   1, 2 ) &&
           getbuf( bp, (char *)&buf->prod_time,   1, 4 ) &&
           getbuf( bp, (char *)&buf->p1,          1, 2 ) &&
           getbuf( bp, (char *)&buf->p2,          1, 2 ) &&
           getbuf( bp, (char *)&buf->elev_num,    1, 2 ) &&
           getbuf( bp, (char *)&buf->p3,          1, 2 ) &&
           getbuf( bp, (char *)&buf->threshold,  16, 2 ) &&
           getbuf( bp, (char *)&buf->p4,          1, 2 ) &&
           getbuf( bp, (char *)&buf->p5,          1, 2 ) &&
           getbuf( bp, (char *)&buf->p6,          1, 2 ) &&
           getbuf( bp, (char *)&buf->p7,          1, 2 ) &&
           getbuf( bp, (char *)&buf->p8,          1, 2 ) &&
           getbuf( bp, (char *)&buf->p9,          1, 2 ) &&
           getbuf( bp, (char *)&buf->p10,         1, 2 ) &&
           getbuf( bp, (char *)&buf->num_maps,    1, 2 ) &&
           getbuf( bp, (char *)&buf->off_symbol,  1, 4 ) &&
           getbuf( bp, (char *)&buf->off_graphic, 1, 4 ) &&
           getbuf( bp, (char *)&buf->off_tabular, 1, 4 );

}

/******************************** radial_image *******************************/

int
radial_image( BUFREAD *bp, unsigned char *raster, int datalen )

/*
** Name:       radial_image
**
** Purpose:    Convert a radial NEXRAD Level III product into a raster image
**
** Parameters:
**             buf     - buffer containing image file data
**             raster  - pointer to array with raster data
**             datalen - size of 'raster' in bytes
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
** History: 19980514 - Adapted from ldm-mcidas distribution for McIDAS-X 7.3
**          20001128 - Recast to read from buffer, not file
**          20100207 - Updated to handle type 0x10 radial images
**
*/

{
    int                   radial;
    int                   run;
    int                   i;
    int                   nbeam;
    int                   size=2;

    radial_header         r_head={0};

    static unsigned char *data = NULL;
    static unsigned char *beam = NULL;

#if DEBUG_RADIM || DEBUG_CALL
    M0sxtrce( "In radial_image" );
#endif

    if ( beam == NULL )
      beam = (unsigned char *) malloc( radp_head.num_bin + 2 );
    else
      beam = (unsigned char *) realloc( beam, radp_head.num_bin + 2 );

    /*
     * For rasterization overflows.
     */

    beam[radp_head.num_bin] = FILL;
    beam[radp_head.num_bin + 1] = FILL;

    (void) memset( raster, FILL, datalen );

#if DEBUG_RADIM
    (void) sprintf( dbg,
           "radial_image:: radp_head.num_radials: %d", radp_head.num_radials );
    M0sxtrce( dbg );
#endif

    /*
    ** Sanity check for number of radials <<<<< UPC add 20030217 >>>>>
    ** 20220128 - check removed because super res images have more than
    **            365 radials
    */

    if ( radp_head.num_radials > 365 ) {
      (void) sprintf( dbg, "radial_image:: NOTE: #radials: %d > 365, Super Resolution image?",
                      radp_head.num_radials );
      M0sxtrce( dbg );
#if 0
      radp_head.num_radials = 365;
#endif
    }

    if ( radp_head.code == 0x10 ) size = 1;

    for ( radial = 0; radial < radp_head.num_radials; radial++ ) {

      if ( !getbuf( bp, (char *) &r_head, 3, 2 ) ) {
        M0sxtrce( "radial_image:: ERROR getting r_head" );

#if DEBUG_RADIM || DEBUG_CALL
       M0sxtrce( "Leave radial_image: FAILURE" );
#endif

        return FAILURE;
      }

#if DEBUG_RADIM
      (void) sprintf( dbg,
           "radial_image:: radial %d, length %d, start %d, delta %d",
           radial, r_head.length, r_head.start, r_head.delta );
      M0sxtrce( dbg );
#endif

      if ( data == NULL )
        data = (unsigned char *) malloc( (size_t) (size*r_head.length));
      else
        data = (unsigned char *) realloc(data, (size_t) (size*r_head.length));

      if ( !getbuf( bp, (char *) data, size*r_head.length, 1 ) ) {
        (void) sprintf( dbg, "radial_image:: ERROR reading r_head.length" );
        M0sxtrce( dbg );

#if DEBUG_RADIM || DEBUG_CALL
       M0sxtrce( "Leave radial_image: FAILURE" );
#endif

        return FAILURE;
      }

      nbeam = 0;

      if ( radp_head.code == 0xAF1F ) {

        for ( run = 0; run < size*r_head.length && data[run]; ++run ) {

          unsigned int drun  = data[run] >> 4;
          unsigned int dcode = data[run] & 0xf;

          for ( i = 0; i < drun; ++i ) {
            beam[nbeam++] = dcode;
          }

        }

      } else {

        for ( run = 0; run < r_head.length; ++run ) {
          beam[nbeam++] = data[run];
        }

      }

      if ( abs(nbeam-radp_head.num_bin) > 1 ) {

        (void) sprintf( dbg,
             "radial_image:: Bad size radial %d, run %d, radp_head.num_bin %d",
             radial, nbeam, radp_head.num_bin);
        M0sxtrce( dbg );

#if DEBUG_RADIM || DEBUG_CALL
       M0sxtrce( "Leave radial_image: FAILURE" );
#endif

        return FAILURE;

      }

      rasterize( raster, beam, (int) radp_head.num_bin,
                 (double) r_head.start / 10., (double) r_head.delta / 10. );

    }

#if DEBUG_RADIM || DEBUG_CALL
    M0sxtrce( "Leave radial_image: SUCCESS" );
#endif

    return SUCCESS;

}

/******************************** raster_image *******************************/

int
raster_image( BUFREAD *bp, unsigned char *raster, int datalen )

/*
** Name:       raster_image
**
** Purpose:    Convert a raster NEXRAD Level III product into a raster image
**
** Parameters:
**             bp      - data buffer structure
**             raster  - pointer to array with raster data
**             datalen - size of 'raster' in bytes
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
** History: 19980514 - Adapted from ldm-mcidas distribution for McIDAS-X 7.3
**
*/

{

    unsigned char *data = NULL, *p;
    int            run;
    int            i, nrow, ncol, num_cols, top, bottom;
    short          length;
    float          scale;

#if DEBUG_RASIM || DEBUG_CALL
    M0sxtrce( "In raster_image" );
#endif

    scale = prod_info.data_res / prod_info.raster_res;

#if DEBUG_RASIM
      (void) sprintf( dbg,
             "rasp_head.code[0] 0x%x, rasp_head.code[1] 0x%x, rasp_head.code[2] 0x%x",
             rasp_head.code[0], rasp_head.code[1], rasp_head.code[2] );
      M0sxtrce( dbg );
#endif

    if ( (rasp_head.code[0] & 0xfff7) != 0xba07  ||
          rasp_head.code[1]           != 0x8000  ||
          rasp_head.code[2]           != 0x00c0     ) {

      (void) sprintf( dbg, "raster_image:: incorrect packet code for raster image");
      M0sxtrce( dbg );

#if DEBUG_RASIM || DEBUG_CALL
        M0sxtrce( "Leave raster_image: FAILURE" );
#endif

      return FAILURE;

    }

    for  ( nrow = 0; nrow < rasp_head.num_rows; ++nrow ) {

      if ( !getbuf( bp, (char *) &length, 1, 2 ) )

#if DEBUG_RASIM || DEBUG_CALL
        M0sxtrce( "Leave raster_image: FAILURE" );
#endif

        return FAILURE;

      if ( data == (unsigned char *) NULL )
        data = (unsigned char *) malloc(length);
      else
        data = (unsigned char *) realloc(data, length);

      if ( !getbuf( bp, (char *) data, length, 1 ) )

#if DEBUG_RASIM || DEBUG_CALL
        M0sxtrce( "Leave raster_image: FAILURE" );
#endif

        return FAILURE;

      if ( nrow == 0 ) {

        num_cols = 0;
        for ( run = 0; run < length && data[run]; ++run ) {
          unsigned int         drun = data[run] >> 4;
          num_cols += drun;
        }

        (void) sprintf( dbg, "raster_image:: first %d/%d, scale %d/%d, nrows %d, ncols %d",
               rasp_head.i, rasp_head.j, rasp_head.xscale, rasp_head.yscale,
               rasp_head.num_rows, num_cols);
        M0sxtrce( dbg );

        if ( rasp_head.num_rows != num_cols ) {
          (void) sprintf( dbg, "raster_image:: Rows != Columns: %d/%d",
                          rasp_head.num_rows, num_cols);
          M0sxtrce( dbg );
        }

      }

      bottom = scale * (rasp_head.num_rows - nrow - 1);
      top    = scale * (rasp_head.num_rows - nrow);
      p      = raster + prod_info.size * bottom;
      ncol   = 0;

#if DEBUG_RASIM
      (void) sprintf( dbg, "raster_image::    Row %d, length %d", nrow, length);
      M0sxtrce( dbg );
#endif

      for ( run = 0; run < length && data[run]; ++run ) {

        unsigned int  drun  = data[run] >> 4;
        unsigned int  dcode = data[run] & 0xf;
        int           iend  = (int)(scale*(ncol+drun)) - (int)(scale*ncol);

        ncol += drun;
        for ( i = 0; i < iend; ++i ) {
          *p++ = dcode;
        }

      }

      if (p - raster != prod_info.size * (bottom + 1) || ncol != num_cols) {

        (void) sprintf( dbg, "raster_image:: ***Bad size raster %d, run %d", nrow, ncol);
        M0sxtrce( dbg );

#if DEBUG_RASIM || DEBUG_CALL
        M0sxtrce( "Leave raster_image: FAILURE" );
#endif

        return FAILURE;

      }

      for (i = bottom + 1; i < top; ++i)
        (void) memcpy( raster + prod_info.size * i,
                       raster + prod_info.size * (i - 1), prod_info.size );

    }

    /*
    ** Return success
    */

#if DEBUG_RASIM || DEBUG_CALL
    M0sxtrce( "Leave raster_image: SUCCESS" );
#endif

    return SUCCESS;

}

/******************************** lfmras_image *******************************/

int
lfmras_image( BUFREAD *bp, unsigned char *raster, int datalen )

/*
** Name:       lfmras_image
**
** Purpose:    Convert a raster NEXRAD Level III product into a raster image
**
** Parameters:
**             bp      - data buffer structure
**             raster  - pointer to array with raster data
**             datalen - size of 'raster' in bytes
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
** History: 20200223 - Adapted from raster_image for support of DPA products
**
*/

{

    unsigned char *data = NULL, *p;
    unsigned int   drun, dcode;
    int            iend, run;
    int            i, nrow, ncol, num_cols, top, bottom;
    short          nbytes;
    float          scale;

#if DEBUG_LFMRS || DEBUG_CALL
    M0sxtrce( "In lfmras_image" );
#endif

    scale = prod_info.data_res / prod_info.raster_res;

#if DEBUG_LFMRS
      (void) sprintf( dbg,
             "lfm_head.code=0x%X, lfm_head.num_lfm=%d, lfm_head.num_row=%d",
             lfm_head.code, lfm_head.num_lfm, lfm_head.num_rows );
      M0sxtrce( dbg );
#endif

    num_cols = lfm_head.num_lfm;

    for  ( nrow = 0; nrow < lfm_head.num_rows; ++nrow ) {

      if ( !getbuf( bp, (char *) &nbytes, 1, 2 ) ) {

#if DEBUG_LFMRS || DEBUG_CALL
        M0sxtrce( "Leave lfmras_image: FAILURE" );
#endif

        return FAILURE;
      }

      if ( data == (unsigned char *) NULL )
        data = (unsigned char *) malloc(nbytes);
      else
        data = (unsigned char *) realloc(data, nbytes);

      if ( !getbuf( bp, (char *) data, nbytes, 1 ) ) {

#if DEBUG_LFMRS || DEBUG_CALL
        M0sxtrce( "Leave lfmras_image: FAILURE" );
#endif

        return FAILURE;
      }

#if DEBUG_LFMRS
      (void) sprintf( dbg, "lfmras_image:: data[0] = %d, data[1] = %d",
                      data[0], data[1] );
      M0sxtrce( dbg );
#endif

      bottom = scale * (lfm_head.num_rows - nrow - 1);
      top    = scale * (lfm_head.num_rows - nrow);
      p      = raster + prod_info.size * bottom;
      ncol   = 0;

#if DEBUG_LFMRS
      (void) sprintf( dbg, "lfmras_image:: Row %d, nbytes %d", nrow, nbytes );
      M0sxtrce( dbg );
#endif

      for ( run = 0; run < nbytes; run+=2 ) {

        drun  = data[run];
        dcode = data[run+1];
        iend  = (int)(scale*(ncol+drun)) - (int)(scale*ncol);

        if ( dcode == 0xFF ) dcode = 0; /* set values outside of area to 0 */

        ncol += drun;
        for ( i = 0; i < iend; ++i ) {
          *p++ = dcode;
        }

      }

      if (p - raster != prod_info.size * (bottom + 1) || ncol != num_cols) {

        (void) sprintf( dbg, "raster_image:: ***Bad size raster %d, run %d",
               nrow, ncol);
        M0sxtrce( dbg );

#if DEBUG_LFMRS || DEBUG_CALL
        M0sxtrce( "Leave lfmras_image: FAILURE" );
#endif

        return FAILURE;

      }

      for (i = bottom + 1; i < top; ++i)
        (void) memcpy( raster + prod_info.size * i,
                       raster + prod_info.size * (i - 1), prod_info.size );

    }

    /*
    ** Return success
    */

#if DEBUG_LFMRS || DEBUG_CALL
    M0sxtrce( "Leave lfmras_image: SUCCESS" );
#endif

    return SUCCESS;

}

/********************************* rasterize *********************************/

int
rasterize( unsigned char *raster, unsigned char *radial, int num_radial,
           double start, double delta )

/*
** Name:       rasterize
**
** Purpose:    Add a radial to a raster array
**
** Parameters:
**             raster     - pointer to array with raster data
**             radial     - pointer to array with radial data
**             num_radial - number of point in radial array
**             start      - start angle for this radial
**             delta      - angle delta for this radial
**
** Returns:
**             SUCCESS 1
**             FAILURE 0
**
** History: 19980514 - Adapted from ldm-mcidas distribution for McIDAS-X 7.3
**
*/

{

    float          az,          /* Angle of mid-point of radial              */
                   az1,         /* Angle of left/bottom of radial            */
                   az2,         /* Angle of right/top of radial              */
                   pgs,         /* raster points per radial point            */
                   tmp;         /* temporary storage for angles              */
    double         sin1, cos1,  /* sine and cosine of az1                    */
                   sin2, cos2,  /* sine and cosine of az2                    */
                   inc1,        /* Tangent/cotangent of az1                  */
                   inc2,        /* Tangent/cotangent of az2                  */
                   rowinc,      /* Change in radial index per row            */
                   colinc,      /* Change in radial index per column         */
                   x0,          /* Raster x value of end of radial           */
                   y0,          /* Raster y value of end of radial           */
                   gate,        /* Radial index of start of row/column       */
                   gate_col,    /* Radial index of current column            */
                   gate_row,    /* Radial index of current row               */
                   left, right, /* Raster index limits for column            */
                   bottom, top; /* Raster index limits for row               */
    unsigned char *dp;          /* Pointer into raster array                 */
    int            col, row,    /* Current column/row                        */
                   vertical,    /* Is this a vertical radial (315-45,135-225)*/
                   flip = 0;    /* Was az1 and za2 flipped?                  */

#if DEBUG_RADRS || DEBUG_CALL
    M0sxtrce( "In rasterize" );
#endif

    /*
     * Set up radial angles az, az1 and az2.
     */

    az1 = start;
    if ( (az = start + 0.5 * delta) >= 360.0 )
      az -= 360;
    if ( (az2 = start + delta) >= 360.0 )
      az2 -= 360;

    /*
     * Figure out the orientation of the radial.  Then, if need be, swap the
     * two angles to correspond to the direction in which we will be
     * rasterizing (i.e. az1 is always left of az2 in a vertical radial, and
     * az1 is always below az2 in a horizontal radial).
     */

    vertical = (az <= 45.0 || (az >= 135.0 && az <= 225.0) || az >= 315.0);

    if ( (vertical && az > 90.0 && az < 270.0) || (!vertical && az < 180.0) ) {
      tmp  = az1;
      az1  = az2;
      az2  = tmp;
      ++flip;
    }

    sin1 = sin( DegToRad(az1) );
    cos1 = cos( DegToRad(az1) );
    sin2 = sin( DegToRad(az2) );
    cos2 = cos( DegToRad(az2) );
    pgs  = prod_info.raster_res / prod_info.data_res;

    /*
     * For vertical radials (315-45, 135-225).
     */

    if ( vertical ) {

      inc1   = sin1 / cos1;
      inc2   = sin2 / cos2;
      rowinc = (sin1 < sin2) ? pgs / cos1 : pgs / cos2;
      colinc = pgs * sin( DegToRad(az) );

      /*
       * If in right side of circle, use az2 (right) side of beam in
       * calculation of y0 (and vice-versa).
       */

      if ( az < 180. )
        y0 = prod_info.size * 0.5 + prod_info.size * 0.5 * cos2;
      else
        y0 = prod_info.size * 0.5 + prod_info.size * 0.5 * cos1;

      /*
       * flip => bottom quadrant of circle, start rasterization from end of
       * radial (edge).
       */

      if ( flip ) {

        gate   = num_radial - .00001;
        left   = prod_info.size * 0.5 + (y0 - prod_info.size * 0.5) * inc1;
        right  = prod_info.size * 0.5 + (y0 - prod_info.size * 0.5) * inc2;
        bottom = y0;
        top    = prod_info.size * 0.5;

      } else {      /* start rasterization from beginning of radial (center) */

        left   = prod_info.size * 0.5;
        right  = prod_info.size * 0.5;
        bottom = prod_info.size * 0.5;
        top    = y0;
        gate   = 0;

      }

      /*
       * Correct the starting index of the radial for the right side of the
       * circle since we will be rasterizing towards the edge.
       */

      if (az < 180.)
        gate -= colinc * (right - left + 1);

      /*
       * Set up pointer for raster array.
       */

      dp = raster + ((int) bottom) * prod_info.size;

      /*
       * Outer loop is by row.
       */

      if ( (int) bottom < 0 || (int) top >= prod_info.size ) {
        (void) sprintf( dbg, "rasterize::  vTB %6.1f %6.3f %6.3f", az, bottom, top);
        M0sxtrce( dbg );
      }

      for ( row = (int) bottom; row < (int) top; row++ ) {

        gate_col = gate;

        /*
         * Inner loop is by column.
         */

        if ( (int) left < 0 || (int) right >= prod_info.size ) {
          (void) sprintf( dbg, "rasterize::  vLR %6.1f %6.3f %6.3f", az, left, right);
          M0sxtrce( dbg );
        }

        for ( col = (int) left; col <= (int) right; col++ ) {

          if ((int) gate_col < 0 || (int) gate_col >= num_radial) {
            (void) sprintf( dbg, "rasterize::    vBAD %6.1f %d %d %6.3f %6.3f",
                   az, row, col, gate, gate_col);
            M0sxtrce( dbg );
          }
          dp[col]   = radial[(int) gate_col];
          gate_col += colinc;

        }

        /*
         * Set up for next row.
         */

        gate  += rowinc;
        left  += inc1;
        right += inc2;
        dp    += prod_info.size;

      }

    } else {                     /* For horizontal radials (45-135, 225-315) */

      inc1   = cos1 / sin1;
      inc2   = cos2 / sin2;
      rowinc = pgs * cos( DegToRad(az) );
      colinc = (cos1 < cos2) ? pgs / sin1 : pgs / sin2;

      /*
       * If in bottom of circle, use az1 (bottom) side of beam in
       * calculation of x0 (and vice-versa).
       */

      if ( az > 90. && az < 270. )
        x0 = prod_info.size * 0.5 + prod_info.size * 0.5 * sin1;
      else
        x0 = prod_info.size * 0.5 + prod_info.size * 0.5 * sin2;

      /*
       * not flip => left quadrant of circle, start rasterization from end
       * of radial (edge).
       */

      if ( !flip ) {

        gate   = num_radial - .00001;
        bottom = prod_info.size * 0.5 + (x0 - prod_info.size * 0.5) * inc1;
        top    = prod_info.size * 0.5 + (x0 - prod_info.size * 0.5) * inc2;
        left   = x0;
        right  = prod_info.size * 0.5;

      } else {      /* start rasterization from beginning of radial (center) */

        bottom = prod_info.size * 0.5;
        top    = prod_info.size * 0.5;
        left   = prod_info.size * 0.5;
        right  = x0;
        gate   = 0;

      }

      /*
       * Correct the starting index of the radial for the top of the circle
       * since we will be rasterizing towards the edge.
       */

      if ( az < 90. || az > 270. )
        gate -= rowinc * (top - bottom + 1);

      /*
       * Outer loop is by column.
       */

      if ( (int) left < 0 || (int) right >= prod_info.size ) {
        (void) sprintf( dbg, "rasterize::  hLR %6.1f %6.3f %6.3f", az, left, right);
        M0sxtrce( dbg );
      }

      for ( col = (int) left; col < (int) right; col++ ) {

        dp = raster + ((int) bottom) * prod_info.size + col;
        gate_row = gate;

        /*
         * Inner loop is by row.
         */

        if ((int) bottom < 0 || (int) top >= prod_info.size) {
          (void) sprintf( dbg, "rasterize::  hTB %6.1f %6.3f %6.3f", az, bottom, top);
          M0sxtrce( dbg );
        }

        for (row = (int) bottom; row <= (int) top; row++) {

          if ((int) gate_row < 0 || (int) gate_row >= num_radial) {
            (void) sprintf( dbg, "rasterize::    hBAD %6.1f %d %d %6.3f %6.3f",
                   az, row, col, gate, gate_row);
            M0sxtrce( dbg );
          }

          *dp       = radial[(int) gate_row];
          gate_row += rowinc;
          dp       += prod_info.size;

        }

        /*
         * Set up for next column.
         */

        gate   += colinc;
        bottom += inc1;
        top    += inc2;

      }

    }

#if DEBUG_RADRS || DEBUG_CALL
    M0sxtrce( "Leave rasterize: SUCCESS" );
#endif

    return SUCCESS;
}
