/**** $Id: servutil.h,v 1.11 2022/02/04 16:05:31 tyoksas Tst $ ****/

/*
**  servutil.h
**  This is the main include file for the Unidata ADDE servers
**
*/

/*
** Include files
*/

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>
#include "m0arg.h"                 /* McIDAS arg fetcher include file        */
#include "mcidas.h"                /* McIDAS include file                    */
#include "mcidasp.h"               /* McIDAS routine prototypes              */

/*
** Return status
*/

#define DEBUG                 0    /* debug flag                             */
#define FAILURE               0    /* failure                                */
#define SUCCESS               1    /* success                                */

/*
** Missing data value
*/

#define MISS         0x80808080    /* hex 80808080                           */

/*
** Max lengths
*/

#define AUX_BLK_LEN         768    /* word length of supplemental block      */
#define IMG_DIR_LEN          64    /* word length of AREA directory          */
#define NAV_COD_LEN         128    /* word length of navigation codicil      */
#define CAL_COD_LEN         128    /* word length of calibration codicil     */
#define MAX_CAL_LEN         512    /* max byte length of calibration codicil */
#define MAX_CARD_LEN         80    /* len of a single comment card           */
#define MAX_DIR_LEN        1024    /* len of directory where data lives      */
#define MAX_ERRMSG_LEN       72    /* byte len of err msg back to client     */
#define MAX_ERR_LEN         512    /* len of error string                    */
#define MAX_HEAD_LEN        300    /* length of NOWrad (tm) header           */
#define MAX_ID_LEN           32    /* max. len of a station id               */
#define MAX_LINE_LEN      20000    /* max. len of a line in the image file   */
#define MAX_MASK_LEN        256    /* max. number of mcidas comment cards    */
#define MAX_NAME_LEN       1024    /* max. len of a image path name          */
#define MAX_NAV_LEN         640    /* max. word length of navigation codicil */
#define MAX_NUM_BAND         32    /* max. number of bands in an image       */
#define MAX_NUM_CARDS       512    /* max. number of mcidas comment cards    */
#define MAX_TOK_LEN          12    /* max. token length                      */

/*
** Comment card strings
*/

#define CEN_LAT_CARD       "Center latitude = %f"
#define CEN_LON_CARD       "Center longitude = %f"
#define CRES_X_CARD        "Computed Longitude resolution (km) = %f"
#define CRES_Y_CARD        "Computed Latitude resolution (km) = %f"
#define RES_X_CARD         "Longitude resolution (km) = %d"
#define RES_Y_CARD         "Latitude resolution (km) = %d"
#define VALID_UNIT_CARD    "Valid calibration unit for band %d = %s \"%s\""

/*
** BUFREAD is a structure used to read memory buffers like files
*/

typedef struct {
  unsigned char   *_ptr;            /* next character from/to here in buffer */
  int             _cnt;             /* # of available characters in buffer */
  unsigned char   *_base;           /* the buffer */
} BUFREAD;

/*
** CONFIG holds the information from the server configuration file
*/

typedef struct CONFIG_ {
  char  dataloc[MAX_NAME_LEN];      /* directory location of data files     */
  char  infoname[MAX_NAME_LEN];     /* pathname of auxiliary info. file     */
  char  filemask[MAX_NAME_LEN];     /* mask for data file names             */
  char  ipmask[MAX_NAME_LEN];       /* mask of acceptable IP addresses      */
} CONFIG;

/*
** CRITERIA holds the search parameters for image selection
*/

typedef struct CRITERIA_ {
  int  bpos;                       /* beginning dataset position             */
  int  epos;                       /* ending dataset position                */
  int  begband;                    /* beginning band number                  */
  int  endband;                    /* ending band number                     */
  int  begday;                     /* beginning data day                     */
  int  endday;                     /* ending data day                        */
  int  begtim;                     /* beginning data time                    */
  int  endtim;                     /* ending data time                       */
  int  begss;                      /* beginning sensor source number         */
  int  endss;                      /* ending sensor source number            */
  char ptype[MAX_ID_LEN];          /* dataset descriptor                     */
  char id[MAX_ID_LEN];             /* station ID                             */
  char filemask[MAX_NAME_LEN];     /* data file name mask                    */
  char dirmask[MAX_NAME_LEN];      /* directory name mask                    */
  char infoname[MAX_NAME_LEN];     /* auxiliary information file name        */
} CRITERIA;

/*
**  FILELIST holds a list of files to be compared to parameters in CRITERIA
*/

typedef struct FILELIST_ {
  char   name[MAX_NAME_LEN];       /* fully qualified file name              */
  int    pos;                      /* ADDE dataset position number           */
  int    hoff;                     /* byte offset to start of header         */
  int    doff;                     /* byte offset to start of data           */
  int    size;                     /* file size in bytes (should be off_t?)  */
  int    type;                     /* file type (dataset dependent)          */
  char   id[MAX_ID_LEN];           /* for NEXRAD Level III station ID        */
  time_t time;                     /* image time [sec since 1970]            */
  struct FILELIST_ *next;
} FILELIST;

/*
**  NAMEMASK holds a list of regular expressions to use for file selection
*/

typedef struct NAMEMASK_ {
  char   mask[MAX_NAME_LEN];       /* regular expression to match            */
  struct NAMEMASK *next;
} NAMEMASK;

/*
**  READPARM holds the specifications needed to read a line of image data
*/

typedef struct READPARM_ {
  char src_type[4];
  char des_unit[4];
  char src_unit[4];

  int  begele;
  int  beglin;
  int  bufsiz;
  int  des_len;
  int  elem_res;
  int  line_res;
  int  maxele;
  int  maxlin;
  int  minele;
  int  minlin;
  int  numband;
  int  numele;
  int  numlin;
  int  src_len;
  int  ul_elem;
  int  ul_line;
} READPARM;

/*
** Unidata server utility interface prototypes
*/

int UPCAddCard( char *, int, int *, int );
int AllowedAccess( CONFIG cfg, servacct *record );
int UPCCheckImgBounds( READPARM *, int, int );
int GetConfigInfo( const char *, const char *, CONFIG *, char * );
int GetFileList( CRITERIA *, FILELIST ** );
int GetIDList( CRITERIA *, FILELIST ** );
int GetImgCards( FILELIST *, int *, int *, int *, int *, char * );
int GetImgRes( FILELIST *, int *, float, float, float *, float *, char * );
int GetImgUnits( int *, int *, int *, int *, char * );
int GetUnitsAndScale( char *, char *, int * );
int NavEarthToImg( FILELIST *, int *, float, float, float *, float *, char * );
int NavImgToEarth( FILELIST *, int *, float, float, float *, float *, char * );
int NumToken( char * );
int PopFile( FILELIST *, FILELIST ** );
int PushFile( FILELIST *, FILELIST ** );
int PushFileByName( FILELIST *, FILELIST ** );
int PushFileByTime( FILELIST *, FILELIST ** );
int ReplaceToken( char *, char *, char * );
int UPCSendCards( int *, int * );
int UPCSendAux( int *, int );
int UPCSendCod( int *, int );
int UPCSendDir( int * );
int UPCSendLine( char *, int );
int UPCSendZeros( char *, int );
